/////////////////////////////////////////////////////////////////////////////////////
// File        : api2218.h
// Function    : Application include file for Sensoray Model 2218
// Author      : Jim Lamberson
// Copyright   : Copyright (C) 2019-2023 Sensoray
/////////////////////////////////////////////////////////////////////////////////////

#ifndef _INC_API2218_H_
#define	_INC_API2218_H_

// Constants ----------------------------------------
#define S2218_MAX_DEVICES           16      // Maximum number of 2218s connected to a host (limited by device selector rotary switch).
#define S2218_NUM_CHANS             8       // Number of sensor channels.
#define S2218_NUM_REFS              9       // Number of internal reference standards.
#define S2218_NUM_GPIOS             8       // Model 2218 has 8 GPIOs.
#define S2218_INFINITE_WAIT         (~0)    // Infinite wait time for blocking functions.
// Error codes ---------------------------------------
#define S2218_ERR_OK                0       // Success
// Device errors
#define S2218_ERR_OPCODE            1    // Command not recognized.
#define S2218_ERR_ARG               2       // Illegal argument value.
#define S2218_ERR_UNPOWERED         3       // Analog circuitry unpowered.
#define S2218_ERR_OFFLINE           4       // Offline (smad.state != SS_RUNNING)
#define S2218_ERR_CRMT              5       // Logged timeout waiting for CRMT.
#define S2218_ERR_DAV               6       // Logged timeout waiting for DAV.
#define S2218_ERR_NO_EFFECT         7       // Command has no effect (e.g., will not cause state change or send associated notification)
#define S2218_ERR_LOCKOUT           8       // Command prohibited due to in-progress diagnostic
#define S2218_ERR_SCANNING          9       // Command prohibited during analog scanning (e.g., SetSensorType)
// API errors
#define S2218_ERR_TIMEOUT           20      // Timed out waiting for device. If relevant, use more specific LOCK or SIG timeout
#define S2218_ERR_FAULT             21      // Couldn't execute command due to fault condition (determine fault(s) by checking FaultFlags)
#define S2218_ERR_REOPEN            22      // Attempted to open smad when it's already open (i.e., smad.state != SS_POWER_OFF).
#define S2218_ERR_CREATE_RESOURCE   23      // Failed to create mutex/semaphore.
#define S2218_ERR_RESOURCE_CLOSED   24      // Resource is closed or unavailable.
#define S2218_ERR_DEVICE_CLOSED     25      // Device is closed.
#define S2218_ERR_DEVICE_OPEN       26      // Device is already open.
#define S2218_ERR_WAS_SHUTDOWN      27      // USB connection was shutdown.
#define S2218_ERR_READ              28      // USB read error.
#define S2218_ERR_WRITE             29      // USB write error.
#define S2218_ERR_OPENHIDLIB        30      // Failed to open HIDLIB.
#define S2218_ERR_DEVICE_NOT_FOUND  31      // Specified device is not detected.
#define S2218_ERR_COMMAND_MISMATCH  32      // Response didn't match last issued command.
#define S2218_ERR_API_OPEN          33      // API is already open.
#define S2218_ERR_API_CLOSED        34      // API is closed.
#define S2218_ERR_LOCK_TIMEOUT      35      // Failed to get resource lock
#define S2218_ERR_SIG_TIMEOUT       36      // Timed out waiting for message from device.
#define S2218_ERR_NOT_ALLOWED       37      // Prohibited by current settings.
#define S2218_ERR_DUPLICATE_ADDRESS 38      // Two modules have the same address.
#define S2218_ERR_CANT_OPEN_DEVICE  39      // Failed to open device.
#define S2218_ERR_MALLOC            40      // Memory allocation failed.
#define S2218_ERR_SYSTEM            41      // Unknown system error.

// Primitive types --------------------------------

typedef unsigned int    uint;
typedef int             boole;

#include <limits.h>

#if (UCHAR_MAX == 0xff)
typedef unsigned char   u8;
#else
#error "UNSUPPORTED 8-BIT SIZE"
#endif

#if (USHRT_MAX == 0xffff)
typedef signed short    s16;
typedef unsigned short  u16;
#else
#error "UNSUPPORTED 16-BIT SIZE"
#endif

#if (UINT_MAX >= 0xffffffff)
typedef unsigned int    u32;
#elif (ULONG_MAX >= 0xffffffffUL)
typedef unsigned long   u32;
#else
#error "UNSUPPORTED 32-BIT SIZE"
#endif

// Enumerated types ------------------------------------------

typedef enum SENSOR_TYPE {              // API sensor type codes:
    SDC_DISABLED                = 0,    //  Disabled channel
    SDC_VOLTS_5V                = 1,    //  5 V range
    SDC_VOLTS_500MV             = 2,    //  500 mV range
    SDC_VOLTS_100MV             = 3,    //  100 mV range
    SDC_OHMS_400                = 4,    //  400 ohms
    SDC_OHMS_4K                 = 5,    //  4 Kohms
    SDC_OHMS_600K               = 6,    //  600 Kohms
    SDC_TEMP_THERMOCOUPLE_B     = 7,    //  B thermocouple
    SDC_TEMP_THERMOCOUPLE_C     = 8,    //  C
    SDC_TEMP_THERMOCOUPLE_E     = 9,    //  E
    SDC_TEMP_THERMOCOUPLE_J     = 10,   //  J
    SDC_TEMP_THERMOCOUPLE_K     = 11,   //  K
    SDC_TEMP_THERMOCOUPLE_N     = 12,   //  N
    SDC_TEMP_THERMOCOUPLE_T     = 13,   //  T
    SDC_TEMP_THERMOCOUPLE_S     = 14,   //  S
    SDC_TEMP_THERMOCOUPLE_R     = 15,   //  R
    SDC_TEMP_THERMISTOR         = 16,   //  10K thermistor
    SDC_TEMP_RTD_PT100_385_400  = 17,   //  RTD Pt 385 400C
    SDC_TEMP_RTD_PT100_385_800  = 18,   //  Pt 385 800C
    SDC_TEMP_RTD_PT100_392_400  = 19,   //  Pt 392 400C
    SDC_TEMP_RTD_PT100_392_800  = 20,   //  Pt 392 800C
    SDC_TEMP_RTD_NI200          = 21,   //  Ni 200
    SDC_TEMP_RTD_NI1000         = 22,   //  Ni 1000
    SDC_TEMP_RTD_CU10           = 23,   //  Cu 10
    S2218_NUM_SENSOR_TYPES      = 24
} SENSOR_TYPE;

// Composite types ----------------------------------

typedef struct S2218_GPIO_PAIR {                    // GPIO configuration pair:
    u8                  Rising;                     //   Value or flags associated with rising edges.
    u8                  Falling;                    //   Value or flags associated with falling edges.
} S2218_GPIO_PAIR;

typedef struct S2218_GPIO_CONFIG {                  // GPIO configurations:
    S2218_GPIO_PAIR     DbTime[S2218_NUM_GPIOS];    //   Debounce interval[gpionum][0=rising edge;1=falling edge]: 0=disabled; else=milliseconds.
    u8                  PinDirections;              //   Pin direction flags: 1=input; 0=output.
    u8                  BiasEnables;                //   Pull-up/down enable flags: 1=enable; 0=disable.
    u8                  BiasPolarities;             //   Pull-up/down type flags: 1=up; 0=down.
    u8                  spare;
} S2218_GPIO_CONFIG;

typedef struct S2218_GPIO_SETTINGS {                // GPIO control states:
    S2218_GPIO_CONFIG   Config;                     //   GPIO configurations.
    S2218_GPIO_PAIR     CapEnables;                 //   Rising/falling edge capture enables.
    u8                  OutputStates;               //   Pin output states (valid only for pins with direction=0).
    u8                  spare;
} S2218_GPIO_SETTINGS;

typedef struct S2218_ALARM_CONFIG {                 // Sensor alarm configuration for one sensor channel:
    double              LoLimit;                    //   Low sensor data threshold.
    double              HiLimit;                    //   High sensor data threshold.
    u8                  Flags;                      //   Configuration/status flags: (x, x, x, x, LIMITENABLE_HI, LIMITENABLE_LO, ALARMENABLE_HI, ALARMENABLE_LO).
} S2218_ALARM_CONFIG;

// S2218_ALARM_CONFIG Flags bits:
#define LIMITENABLE_HI  (1 << 3)        // Enable high threshold checking
#define LIMITENABLE_LO  (1 << 2)        // Enable low threshold checking
#define ALARMENABLE_HI  (1 << 1)        // Enable high threshold alarms
#define ALARMENABLE_LO  (1 << 0)        // Enable high threshold alarms

typedef struct S2218_ANALOG_CONFIG {                // Analog system configuration:
    SENSOR_TYPE         SensorType[S2218_NUM_CHANS]; //   Sensor types for all channels.
    u8                  SmadState;                  //   Smad state (PowerEnabled is indicated by state != SS_POWER_OFF).
    u8                  TcCondEnables;              //   Thermocouple conditioner enables.
    u8                  OpenFlags;                  //   Open-sensor data value flags for all sensor channels: 1 = max positive, 0 = max negative.
} S2218_ANALOG_CONFIG;

typedef struct S2218_TRIG_CONFIG {                  // DataReport trigger configuration:
    u16                 Rate;                       //   Periodic report rate (scans per report).
    S2218_GPIO_PAIR     GpioFlags;                  //   GPIO edge selector flags.
} S2218_TRIG_CONFIG;

typedef struct S2218_TIMESTAMP {                    // Timestamp (elapsed time since last boot):
    u32                 s;                          //   Seconds.
    u32                 ms;                         //   Milliseconds.
} S2218_TIMESTAMP;

typedef struct S2218_GPIOREPORT {                   // GPIO edge detection report:
    S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
    int                 nDropped;                   //   Number of dropped reports.
    S2218_GPIO_PAIR     Flags;                      //   EdgeDetected flags (MSB = rising edges; LSB = falling edges).
    u8                  PinStates;                  //   GPIO pin states.
} S2218_GPIOREPORT;

typedef struct S2218_ALARMREPORT {                  // Limit alarm report:
    S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
    int                 nDropped;                   //   Number of dropped reports.
    u16                 Flags;                      //   Alarm flags (MSB = high alarms; LSB = low alarms).
} S2218_ALARMREPORT;

typedef struct S2218_DATAREPORT {                   // Data report:
    double              Samples[S2218_NUM_CHANS];   //   Sensor data.
    S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
    int                 nDropped;                   //   Number of dropped reports.
    u8                  HiLimitVios;                //   Limit violations.
    u8                  LoLimitVios;
} S2218_DATAREPORT;

typedef struct S2218_VERSION {                      // Version number:
    u8                  Major;                      //   Major version.
    u8                  Minor;                      //   Minor version.
    u16                 Build;                      //   Build number.
} S2218_VERSION;

typedef struct S2218_VERSION_INFO {                 // Version info:
    S2218_VERSION       DevFw;                      //   Device firmware version number.
    u16                 SmadFw;                     //   Analog system firmware version number (x100).
    char                PwbRev;                     //   Device hardware revision letter.
} S2218_VERSION_INFO;

typedef struct S2218_STATUSREPORT {                 // Status report:
    int                 SoftTrig;                   //   TRUE if report was triggered by a soft notification command (vs. smad system state change).
    int                 nDropped;                   //   Number of reports dropped from notification queue (at the moment the queue is read by api client).
    s16                 Meter_mV[4];                //   Power supply voltages in mV.
    u8                  PowerEnables;               //   Power supply enable flags: (x/5, D, P, N).
    u8                  PowerGood;                  //   All power supplies at nominal normal operating voltage.
    u8                  DPowerState;                //   Digital power state.
    u8                  PPowerState;                //   +Analog power state.
    u8                  NPowerState;                //   -Analog power state.
    u8                  SmadState;                  //   Smad system state.
    u8                  ComState;                   //   Smad communication state.
} S2218_STATUSREPORT;

typedef struct S2218_TASKINFO {                     // Task statistics:
    char                Name[4];                    //   Task ID.
    u16                 StackMargin;                //   Task stack headroom.
    u16                 ExecTime;                   //   Worst-case task completion time (us).
    u16                 LoopTime;                   //   Worst-case task loop time (us).
} S2218_TASKINFO;

typedef struct S2218_VFINFO {
    u32     acc;        // Accumulated 16MHz ticks
    u16     iters;      // VF cycle count
    double  freq;       // VF frequency
} S2218_VFINFO;

typedef struct S2218_INT_SNAPSHOT {   // All reference standards
    S2218_VFINFO  vfinfo[S2218_NUM_REFS];
} S2218_INT_SNAPSHOT;

typedef struct S2218_EXT_SNAPSHOT {   // One external chan ----------
    u8              ctrlreg;        // Control register image
    u8              acc0;           // "Null counts" flag
    u16             tsettle;        // Front-end settling time
    S2218_VFINFO    vfinfo;         // VF info
} S2218_EXT_SNAPSHOT;

typedef struct S2218_EEPROM_DATA {
    double V5;       // LG ref
    double MV500;    // HG ref
    double R0;       // R ref (low resistance range)
    double R1;       // R ref (mid resistance range)
    double R2;       // R ref (high resistance range)
} S2218_EEPROM_DATA;

// API functions -----------------------------------

// Admin
int s2218_OpenApi(int *devflags);
int s2218_CloseApi(void);
int s2218_OpenDevice(int devaddr, int *online);
int s2218_ShutdownDevice(int devaddr);
int s2218_CloseDevice(int devaddr);
const char *s2218_ErrString(int errcode);
int s2218_GetApiVersion(S2218_VERSION *ver);
int s2218_GetDeviceVersions(int devaddr, S2218_VERSION_INFO *ver);
// Analog configuration
int s2218_SetSensorType(int devaddr, int chan, SENSOR_TYPE sdc);
int s2218_SetOpenSensorValues(int devaddr, u8 failHighFlags);
int s2218_GetAnalogConfig(int devaddr, S2218_ANALOG_CONFIG *cfg);
// Threshold limits and alarms
int s2218_GetAlarmConfig(int devaddr, u8 chan, S2218_ALARM_CONFIG *cfg);
int s2218_SetLoThreshold(int devaddr, u8 chan, double limval);
int s2218_SetHiThreshold(int devaddr, u8 chan, double limval);
int s2218_SetLoLimitEnable(int devaddr, u8 chan, short enab);
int s2218_SetHiLimitEnable(int devaddr, u8 chan, short enab);
int s2218_SetLoAlarmEnable(int devaddr, u8 chan, short enab);
int s2218_SetHiAlarmEnable(int devaddr, u8 chan, short enab);
int s2218_WaitForLimitAlarm(int devaddr, S2218_ALARMREPORT *report, int tmax);
// Data acquisition
int s2218_SetTrigConfig(int devaddr, S2218_TRIG_CONFIG *cfg);
int s2218_GetTrigConfig(int devaddr, S2218_TRIG_CONFIG *cfg);
int s2218_SetTrigEnable(int devaddr, u8 enable);
int s2218_GetTrigEnable(int devaddr, u8 *enable);
int s2218_WaitForDataReport(int devaddr, S2218_DATAREPORT *report, int tmax);
int s2218_ReadDeviceTemp(int devaddr, double *degC);
// GPIO
int s2218_SetGpioOutputs(int devaddr, u8 states);
int s2218_GetGpioInputs(int devaddr, u8 *states);
int s2218_SetGpioConfig(int devaddr, S2218_GPIO_CONFIG *cfg);
int s2218_GetGpioSettings(int devaddr, S2218_GPIO_SETTINGS *cfg);
int s2218_SetGpioCapEnables(int devaddr, S2218_GPIO_PAIR *enables);
int s2218_WaitForGpioEvent(int devaddr, S2218_GPIOREPORT *report, int tmax);
// Factory use only
int s2218_SetTcCondEnable(int devaddr, int chan, int enable);
int s2218_CalibrateStandard(int devaddr, int chan, int calID, double sensval, double *refval);
int s2218_CalibrateCJ(int devaddr, u8 dir, u8 stepcount);
int s2218_GrabStatus(int devaddr, S2218_STATUSREPORT *status, int tmax);
int s2218_WaitForAsyncStatusReport(int devaddr, S2218_STATUSREPORT *status, int tmax);
int s2218_ReadSensorData(int devaddr, int chan, double *data, int tmax);
int s2218_WriteReg(int devaddr, int addr, int data);
int s2218_ReadReg(int devaddr, int addr, u8 *data);
int s2218_LockReg(int devaddr, u8 lock);
int s2218_SetPowerEnables(int devaddr, u8 flags);
int s2218_GetPowerEnables(int devaddr, u8 *flags);
int s2218_AnalogReset(int devaddr);
int s2218_GetStatistics(int devaddr, int taskid, S2218_TASKINFO *info);
int s2218_HaltOnChan(int devaddr, u8 chan);
int s2218_HaltOnSetup(int devaddr, u8 setup);
int s2218_ReadEeprom(int devaddr, S2218_EEPROM_DATA *eedata);
int s2218_GetIntSnapshot(int devaddr, S2218_INT_SNAPSHOT *s);
int s2218_GetExtSnapshot(int devaddr, u8 chan, S2218_EXT_SNAPSHOT *s);
int s2218_ReadSN(int devaddr, u32 *sn);
int s2218_WriteSN(int devaddr, u32 sn);

#endif // ifdef _INC_API2218_H_
