#!/usr/bin/env python3
# Python demo for the Sensoray Model 2218
# Copyright (c) 2024 Sensoray Co., Inc.

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib

import os, ctypes, glob, threading, subprocess, time, select, string, sys, socket
import tkinter as tk
from tkinter import Tk
from tkinter import filedialog as fd

from fcntl import *
from mmap import *

try:
	import s2218
except OSError as e:
	msg = Gtk.MessageDialog(parent=None,
		modal=True,
		message_type=Gtk.MessageType.ERROR,
		buttons=Gtk.ButtonsType.CLOSE,
		text="Unable to open 'libs2218.so' shared library. Please check README file and run 'make'.")
	msg.show()
	msg.run()
	msg.destroy()
	sys.exit(0)
        
debug = False
for arg in sys.argv:
        if arg == "-d":
                debug = True

def dprint(*args):
        if debug:
                for x in args:
                        print(x,)
                print

def update_devices(demo):
        s2218.CloseApi()
        numdev = 0
        devflags = s2218.OpenApi()
        demo.vbox_device.remove(demo.hbox_device)
        hbox = Gtk.HBox(homogeneous=False, spacing=10)                        
        for i in range (0, 16):
                if (1 << i & devflags.value):
                        dprint("found device " + str(i))
                        demo.device_present[i] = 1
                        numdev += 1
                else:
                        demo.device_present[i] = 0
        for i in range (0, 16):                        
                if (demo.device_present[i]):
                        ver = s2218.devVersionStruct(0,0,0)
                        txt = "2218 #" + str(i)
                        if (demo.device_open[i]):
                                dprint("device_open")
                                rc,ver = s2218.GetDevVersions(i)
                        else:
                                s2218.OpenDevice(i)
                                demo.device_open[i] = 1
                                rc,ver = s2218.GetDevVersions(i)
                                s2218.CloseDevice(i)
                                demo.device_open[i] = 0
                        if (rc == 0):
                                txt += "\n DEV FW: " + str(ver.major) + "." + str(ver.minor) + " build " + str(ver.build)
                                txt += "\n SMAD FW: " + str(ver.smad_fw/100)
                                txt += "\n PWD REV: " + chr(ver.pwb_rev)
                                button = Gtk.Button(label=txt)
                                button.show();
                                button.index = i
                                button.connect("clicked", demo.button_device)
                                hbox.pack_start(button, False, False, 0)
                        else:
                                dprint("rc error" + str(rc))
        hbox.show()
        demo.hbox_device = hbox
        if (numdev == 0):
                label = Gtk.Label(label="Scanning for devices: None found\n\n\nPlug in 2218 device(s).")
                label.show()
                hbox.pack_start(label, False, False, 0)

        demo.vbox_device.pack_start(hbox, False, False, 0)
        demo.vbox_device.show()
        return

class device_thread(threading.Thread):
        def __init__(self, demo):
                threading.Thread.__init__(self)
                self.demo = demo
                return
        def run(self):
                self.quit = 0
                dprint("device_thread running")
                while (self.quit != 1):
                        dprint("device_thread running")
                        GLib.idle_add(update_devices, self.demo)
                        time.sleep(1)
                dprint("device_thread exit")
                return
        def stop(self):
                self.quit = 1
                return
        
class scan_thread(threading.Thread):
        def __init__(self, device, name, demo):
                threading.Thread.__init__(self)
                self.action = 'scanning'
                self.logfile = 0
                self.demo = demo
                self.name = name
                self.deviceid = device
                self.report = s2218.s2218_datareport()
                self.trig = s2218.trig_config_python(0,0,0)
                self.trig.rate = 5
                self.trig.gpio_flags_rising = 0
                self.trig.gpio_flags_falling = 0
                s2218.SetTrigConfig(device, self.trig)
                s2218.SetTrigEnable(device, 1)
                self.quit = False
                self.flash_status = 1
                if (demo.chk_log.get_active() != 0):
                        if (demo.log_filename != ""):
                                self.f = open(demo.log_filename, "w")
                                self.logfile = 1
        def run(self):
                report = s2218.s2218_datareport()
                while not self.quit:
                        rc = s2218.WaitForDataReport(self.deviceid, report, 2000)
                        if (rc == 0):
                                txt = "{:d}".format(report.timestamp_s) + "." + str(report.timestamp_ms)
                                for i in range(0,8):
                                        txt_analog = "{:.3f}".format(report.samples[i])
                                        GLib.idle_add(demo.cmb_analog_val[i].set_text, txt_analog)
                                        txt += "," + txt_analog
                        else:
                                txt = "error: " + str(rc)
                        txt += "\n"
                        if (self.logfile):
                                self.f.write(txt)
                        demo.scan_buf += txt
                        txt_time = "{:5d}".format(report.timestamp_s) + "." + "{:3d}".format(report.timestamp_ms)
                        
                        
                        GLib.idle_add(demo.label_time.set_text, txt_time)
                        GLib.idle_add(demo.textbuf.set_text, demo.scan_buf)
                        if (self.flash_status):
                                GLib.idle_add(demo.lbl_status_scan.set_text,"Scanning")
                        else:
                                GLib.idle_add(demo.lbl_status_scan.set_text,"")
                        if (self.flash_status == 0):
                                self.flash_status = 1;
                        else:
                                self.flash_status = 0;
                s2218.SetTrigEnable(self.deviceid, 0)                
                GLib.idle_add(demo.lbl_status_scan.set_text,"Idle")
        def stop(self):
                dprint("scan thread stop")
                if (self.logfile):
                        self.logfile = 0
                        self.f.close()
                self.quit = True

class gpio_thread(threading.Thread):
        def __init__(self, device, name, demo):
                threading.Thread.__init__(self)
                self.action = 'gpio_scanning'
                self.logfile = 0
                self.demo = demo
                self.name = name
                self.deviceid = device
                self.quit = False
                self.flash_status = 1
                if (demo.chk_log_gpio.get_active() != 0):
                        try:
                                if (demo.log_filename_gpio != ""):
                                        self.f = open(demo.log_filename_gpio, "w")
                                        self.logfile = 1
                        except:
                                dprint("no file selected\n")
        def run(self):
                while not self.quit:
                        rc, inputs = demo.read_gpio_inputs(self.deviceid)
                        txt = ""
                        for i in range(0, s2218.NUM_GPIOS):
                                if (i != (s2218.NUM_GPIOS - 1)):
                                        txt += "{:d},".format((inputs.value >> i) & 0x01)
                                else:
                                        txt += "{:d}\n".format((inputs.value >> i) & 0x01)
                        if (self.logfile):
                                self.f.write(txt)
                        if (self.flash_status):
                                GLib.idle_add(demo.lbl_status_gpio.set_text,"Scanning")
                        else:
                                GLib.idle_add(demo.lbl_status_gpio.set_text,"")

                        if (self.flash_status == 0):
                                self.flash_status = 1;
                        else:
                                self.flash_status = 0;
                        time.sleep(1)
        def stop(self):
                dprint("scan thread stop")
                if (self.logfile):
                        self.logfile = 0
                        self.f.close()
                self.quit = True
                
class Demo():
        device_present = [0, 0, 0, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0]
        device_open = [0, 0, 0, 0, 0, 0, 0, 0,
                       0, 0, 0, 0, 0, 0, 0, 0]        
        cur_device = -1
        analog_cfg = s2218.analog_config_struct
        cmb_gpio_pin_dir = []
        cmb_gpio_bias = []
        cmb_gpio_polarity = []
        cmb_gpio_val = []
        cmb_analog_val = []        
        def btn_read_gpio_inputs(self, widget):
                devid = self.cur_device
                rc,inputs = s2218.GetGpioInputs(devid)
                for i in range(0, s2218.NUM_GPIOS):
                        #dprint("val:" + str(inputs.value & (1 << i)))
                        if (self.cmb_gpio_val[i].get_sensitive() == False):
                                if (inputs.value & (1 << i)):
                                        GLib.idle_add(self.cmb_gpio_val[i].set_active, 1)
                                else:
                                        GLib.idle_add(self.cmb_gpio_val[i].set_active, 0)
                return rc, inputs
        def read_gpio_inputs(self, devid):
                rc,inputs = s2218.GetGpioInputs(devid)
                for i in range(0, s2218.NUM_GPIOS):
                        #dprint("val:" + str(inputs.value & (1 << i)))
                        if (inputs.value & (1 << i)):
                                GLib.idle_add(self.cmb_gpio_val[i].set_active, 1)
                        else:
                                GLib.idle_add(self.cmb_gpio_val[i].set_active, 0)
                return rc, inputs
        def get_gpio_settings(self, devid):
                cfg = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(devid, cfg)
                return
        
                
        def power_n12v_enable(self, widget, flags):
                value = int(widget.get_active())
                dprint("value: " + str(value))

                if (value == 0):
                        flags.value &= ~1
                else:
                        flags.value |= 1
                dprint("flags: " + str(flags.value))
                s2218.SetPowerEnables(self.cur_device, int(flags.value))
                #update analog status
                time.sleep(0.1)
                self.update_analog_status()
                return
        def power_15v_enable(self, widget, flags):
                value = int(widget.get_active())
                if (value == 0):
                        flags.value &= ~2
                else:
                        flags.value |= 2
                s2218.SetPowerEnables(self.cur_device, int(flags.value))
                # update status
                time.sleep(0.1)
                self.update_analog_status()
                return
        def set_sensor_type(self, widget):
                idx = widget.index
                rc = s2218.SetSensorType(self.cur_device, idx, int(widget.get_active()))
                return
        def set_gpio_falling(self, widget):
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                try:
                        val = int(widget.get_text())
                        gpio = cfg_gpio.config
                        if (val > 255):
                                val = 255
                                widget.set_text("255")
                        if (val < 0):
                                val = 0;
                                widget.set_text("0")
                        gpio.dbtime[idx].falling = val
                        
                        rc = s2218.SetGpioConfig(self.cur_device, gpio)
                        dprint("set gpio config " + str(rc))
                except:
                        dprint("invalid value")
                return
        def set_gpio_rising(self, widget):
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                try:
                        val = int(widget.get_text())
                        if (val > 255):
                                val = 255
                                widget.set_text("255")
                        if (val < 0):
                                val = 0;
                                widget.set_text("0")                        
                        gpio = cfg_gpio.config
                        gpio.dbtime[idx].rising = val
                        rc = s2218.SetGpioConfig(self.cur_device, gpio)
                        dprint("set gpio config " + str(rc))
                except:
                        dprint("set gpio fail")
                return
        def set_gpio_output(self, widget):
                dprint("set_gpio_output")
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                val = int(widget.get_active())
                states = cfg_gpio.output_states
                if (val != 0):
                        states |= (1 << idx)
                else:
                        states &= ~(1 << idx)
                rc = s2218.SetGpioOutputs(self.cur_device, states)
                dprint("set gpio outputs " + str(rc))
                return
        def gpio_dir_changed(self, idx, b_output):
                if (b_output == 1):
                        self.cmb_gpio_val[idx].set_sensitive(True)
                        cfg_gpio = s2218.gpio_settings()
                        rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                        if (rc != 0):
                                return
                        states = cfg_gpio.output_states
                        if (states & (1 << idx)):
                                self.cmb_gpio_val[idx].set_active(1)
                        else:
                                self.cmb_gpio_val[idx].set_active(0)
                else:

                        rc, states = s2218.GetGpioInputs(self.cur_device)
                        if (rc != 0):
                                return
                        if (states.value & (1 << idx)):
                                self.cmb_gpio_val[idx].set_active(1)
                        else:
                                self.cmb_gpio_val[idx].set_active(0)
                        self.cmb_gpio_val[idx].set_sensitive(False)
                return
        def set_gpio_dir(self, widget):
                dprint("set_gpio_dir")
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                val = int(widget.get_active())
                gpio = cfg_gpio.config
                # val = 0 == input. pin_direction = 1
                # val = 1 == output. pin_direction = 0
                if (val == 0):
                        gpio.pin_directions |= (1 << idx)
                else:
                        gpio.pin_directions &= ~(1 << idx)
                rc = s2218.SetGpioConfig(self.cur_device, gpio)
                dprint("set gpio dir: dir: " + str(gpio.pin_directions) + " rc: " + str(rc))
                self.gpio_dir_changed(idx, val)
                return
        def set_gpio_polarity(self, widget):
                dprint("set_gpio_polarity")
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                val = int(widget.get_active())
                gpio = cfg_gpio.config
                if (val != 0):
                        gpio.bias_polarities |= (1 << idx)
                else:
                        gpio.bias_polarities &= ~(1 << idx)
                dprint("gpio bias_polarities: " + str(gpio.bias_polarities))
                rc = s2218.SetGpioConfig(self.cur_device, gpio)
                dprint("set gpio polarity " + str(rc))
                self.read_gpio_inputs(self.cur_device)
                return
        def set_gpio_bias(self, widget):
                dprint("set_gpio_bias enable")
                idx = widget.index
                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                if (rc != 0):
                        return
                val = int(widget.get_active())
                gpio = cfg_gpio.config
                if (val != 0):
                        gpio.bias_enables |= (1 << idx)
                else:
                        gpio.bias_enables &= ~(1 << idx)
                dprint("gpio bias: " + str(gpio.bias_enables))
                rc = s2218.SetGpioConfig(self.cur_device, gpio)
                dprint("set gpio bias enable: rc: " + str(rc))
                self.read_gpio_inputs(self.cur_device)
                return
        
        def button_clear(self, widget):
                self.scan_buf = ""
                self.textbuf.set_text("")
                return
        def btn_start_scan(self, widget):
                thread = scan_thread(self.cur_device, "scan", self)
                self.btn_start.set_sensitive(False)
                if (self.chk_log.get_active() != 0):
                        if (self.log_filename == ""):
                                dprint("select filename first")
                self.threads.append(thread)
                thread.start()
                self.btn_select.set_sensitive(False)
                self.chk_log.set_sensitive(False)                
                return
        def btn_stop_scan(self, widget):
                self.stop_thread(widget)
                self.btn_select.set_sensitive(True)
                self.chk_log.set_sensitive(True)
                self.btn_start.set_sensitive(True)
                return
        def btn_select_file(self, widget):
                root = Tk()
                root.withdraw()
                filename = fd.asksaveasfilename()
                root.update()
                dprint(filename)
                self.log_filename = filename
                self.label_filename.set_text(filename)
                return
        def btn_start_gpio(self, widget):
                thread = gpio_thread(self.cur_device, "gpio", self)
                self.btn_gpio_start.set_sensitive(False)
                if (self.chk_log.get_active() != 0):
                        if (self.log_filename_gpio == ""):
                                dprint("select filename first")
                self.threads.append(thread)
                thread.start()
                self.lbl_status_gpio.set_text("Scanning")
                self.chk_log_gpio.set_sensitive(False)
                self.btn_gpio_select.set_sensitive(False)
                dprint("btn_start_gpio")
                return
        def btn_stop_gpio(self, widget):
                self.stop_thread(widget)
                self.btn_gpio_start.set_sensitive(True)
                self.lbl_status_gpio.set_text("Idle")
                self.chk_log_gpio.set_sensitive(True)
                self.btn_gpio_select.set_sensitive(True)
                dprint("btn_stop_gpio")
                return
        def btn_select_gpio(self, widget):
                root = Tk()
                root.withdraw()
                filename = fd.asksaveasfilename()
                root.update()
                dprint(filename)
                self.log_filename_gpio = filename
                self.label_filename_gpio.set_text(filename)
                dprint("btn_select_gpio")                
                return        
        def button_device(self, widget):
                board = widget
                dprint("button device: " + str(board.index))
                self.thread_device.stop()
                self.window.remove(self.window.get_child())
                if (self.cur_device != -1):
                        self.device_open[self.cur_device] = 0
                        dprint("close device " + str(self.cur_device))
                        s2218.CloseDevice(self.cur_device)
                if (self.device_open[board.index] == 0):
                        dprint("open device")
                        rc, online = s2218.OpenDevice(board.index)
                        self.device_open[board.index] = 1
                        self.cur_device = board.index
                self.window.add(self.setup_analog())
                return
        def update_analog_status(self):
                rc, stat = s2218.GrabStatus(self.cur_device, 1000)
                if (rc == 0):
                        if (stat.power_good & s2218.POWERGOOD_3R3V):
                                txt = "DIG = " + "{:.2f}".format(stat.meter_mv[1]*0.001) + "V (ok),"
                        else:
                                txt = "DIG = " + "{:.2f}".format(stat.meter_mv[1]*0.001) + "V (off),"                                
                        self.label_status0.set_text(txt)
                        if (stat.power_good & s2218.POWERGOOD_N12V):
                                txt ="ANA- = " + "{:.2f}".format(stat.meter_mv[0]*0.001) + "V (ok),"
                        else:
                                txt ="ANA- = " + "{:.2f}".format(stat.meter_mv[0]*0.001) + "V (off),"
                        self.label_status1.set_text(txt)
                        if (stat.power_good & s2218.POWERGOOD_P15V):
                                txt = "ANA+ = {:.2f}V (ok),".format(stat.meter_mv[3]*0.001)
                        else:
                                txt = "ANA+ = {:.2f}V (off),".format(stat.meter_mv[3]*0.001)
                        self.label_status2.set_text(txt)
                        if (stat.power_good & s2218.POWERGOOD_5V):        
                                txt = "VREF = {:.2f}V (ok)".format(stat.meter_mv[2]*0.001)
                        else:
                                txt = "VREF = {:.2f}V (off)".format(stat.meter_mv[2]*0.001)                                
                        self.label_status3.set_text(txt)
                        # if analog voltage correct
                        if (stat.power_good & (s2218.POWERGOOD_P15V | s2218.POWERGOOD_N12V)):
                                rc, temp = s2218.ReadDeviceTemp(self.cur_device)
                                if rc == 0:
                                        self.label_temperature.set_text("Device Temperature: " + "{:.1f}C".format(temp.value))
                                else:
                                        self.label_temperature.set_text("")
                        else:
                                self.label_temperature.set_text("")
                else:
                        self.label_status0.set_text("Get status failed with error " + str(rc))
                        self.label_status1.set_text("")
                        self.label_status2.set_text("")
                        self.label_status3.set_text("")                        
                return
        def button_status(self, widget):
                self.update_analog_status()
                return
        def button_board(self, widget):
                dprint("button board")
                if (self.cur_device != -1):
                        self.device_open[self.cur_device] = 0
                        dprint("close device " + str(self.cur_device))
                        s2218.CloseDevice(self.cur_device)
                self.window.remove(self.window.get_child())
                self.window.add(self.setup_board())
                self.window.show()
                return
        def button_gpio(self, widget):
                dprint("button gpio")
                # stop all active scanning threads
                self.window.remove(self.window.get_child())
                self.window.add(self.setup_gpio())
                return
        def button_analog(self, widget):
                dprint("button analog")
                self.window.remove(self.window.get_child())
                self.window.add(self.setup_analog())
                return
                
        def delete_event(self, widget, event, data=None):
                return False

        def destroy(self, widget, data=None):
                # search for scanning threads not stopped
                self.stop_all_threads()
                self.thread_device.stop()
                if (self.cur_device != -1):
                        if (self.device_open[self.cur_device] != 0):
                                dprint("close device")
                                s2218.CloseDevice(self.cur_device)
                                self.device_open[self.cur_device] = 0
                        self.cur_device = -1
                s2218.CloseApi()
                dprint("close api")
                Gtk.main_quit()
                dprint("main quit")

        def set_opt(self, widget, opt, value):
                ddprint("set_active %d" % (value))
                opt.set_active(value)
                opt.get_menu().get_active().activate()
        
        def set_toggle(self, widget, id):
                self.set_ctrl(widget, id, int(widget.get_active()))
                
        def message_box(self, msg):
                msg = Gtk.MessageDialog(parent=self.window,
                        modal=True,
                        message_type=Gtk.MessageType.INFO,
                        buttons=Gtk.ButtonsType.CLOSE,
                        text=msg)
                msg.show()
                msg.run()
                msg.destroy()
                
        def message_box_yes_no(self, msg):
                msg = Gtk.MessageDialog(parent=self.window,
                        modal=True,
                        message_type=Gtk.MessageType.INFO,
                        buttons=Gtk.ButtonsType.YES_NO,
                        text=msg)
                msg.show()
                result = msg.run()
                msg.destroy()
                return result == Gtk.ResponseType.YES
        
        def stop_thread(self, widget):
                for thread in self.threads:
                        if thread.deviceid == self.cur_device:
                                dprint("thread found stopping")
                                thread.stop()
                                self.threads.remove(thread)
                        else:
                                dprint("different device")
                dprint("stop thread")
        def stop_all_threads(self):
                for thread in self.threads:
                        if (thread.is_alive()):
                                thread.stop()
                        for i in range(0,10):
                                if (thread.is_alive()):
                                        time.sleep(0.1)
                        if (thread.is_alive()):
                                dprint("thread still alive!")
                        else:
                                dprint("thread stopped")
                        self.threads.remove(thread)
                
        def __init__(self):
                self.threads = []
                self.window = Gtk.Window(title = "Sensoray 2218 Demo")
                self.window.connect("delete_event", self.delete_event)
                self.window.connect("destroy", self.destroy)
                self.window.set_border_width(10)
                self.default_stream_ip = None
                self.default_stream_port = None
                self.default_folder = os.getcwd()
                self.window.add(self.setup_board())
                self.window.show()
                # create scanning thread for devices (unplug/replug)
                return
        def add_view_frame(self, vbox):
                frame = Gtk.Frame(label="View")
                frame.show()
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                btn_board = Gtk.Button(label="Change Board")
                btn_board.connect("clicked", self.button_board)
                btn_board.show()
                hbox.pack_start(btn_board, False, False, 10)
                btn_analog = Gtk.Button(label="Analog")
                btn_analog.connect("clicked", self.button_analog)
                btn_analog.show()
                hbox.pack_start(btn_analog, False, False, 10)
                btn_gpio = Gtk.Button(label="GPIO")
                btn_gpio.connect("clicked", self.button_gpio)
                btn_gpio.show()
                hbox.pack_start(btn_gpio, False, False, 10)                
                hbox.show()
                frame.add(hbox)
                vbox.pack_end(frame, False, False, 0)
        def setup_analog(self):
                #core banner
                rc, self.analog_cfg = s2218.GetAnalogConfig(self.cur_device)
                vbox = Gtk.VBox(homogeneous=False, spacing=10)
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                header = Gtk.Label(label="Analog Settings")
                header.show()
                hbox.pack_start(header, False, False, 0)
                hbox.show()
                vbox.pack_start(hbox, False, False, 0)
                
                hbox = Gtk.HBox(homogeneous=False, spacing=10)

                rc_ver, devver = s2218.GetDevVersions(self.cur_device)
                txt = "Device Index: " + str(self.cur_device) + ", Version Info: "
                if (rc == 0):
                        txt += "dev " + str(devver.major) + "." + str(devver.minor) + "." + str(devver.build) + ", "
                        txt += "smad " + str(devver.smad_fw/100) + ", pwd rev " + chr(devver.pwb_rev)
                        label=Gtk.Label(label=txt)
                else:
                        txt += "Unknown. (error: " + str(rc) + ")"
                self.label_version=Gtk.Label(label=txt)
                hbox.pack_start(self.label_version, False, False, 0)
                self.label_temperature = Gtk.Label(label="")
                self.label_temperature.show()
                hbox.pack_end(self.label_temperature, False, False, 0)
                vbox.pack_start(hbox, False, False, 0)
                self.label_version.show()                
                hbox.show()
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                label=Gtk.Label(label="Power:")
                label.show()
                hbox.pack_start(label, False, False, 0)
                hbox.show()
                vbox.pack_start(hbox, False, False, 0)
                rc_power, flags = s2218.GetPowerEnables(self.cur_device)
                dprint("flags " + str(flags))
                chk = Gtk.CheckButton(label="15V")
                chk.set_active(0)
                if (flags.value & 2):
                        chk.set_active(1)
                else:
                        chk.set_active(0)                
                hbox.pack_start(chk, False, False, 8);
                chk.show()
                chk.connect("clicked", self.power_15v_enable, flags)

                chk = Gtk.CheckButton(label="-12V")
                if (flags.value & 1):
                        chk.set_active(1)
                else:
                        chk.set_active(0)
                hbox.pack_start(chk, False, False, 8);
                chk.show()
                chk.connect("clicked", self.power_n12v_enable, flags)
                button = Gtk.Button(label="Status")
                button.connect("clicked", self.button_status)
                hbox.pack_start(button, False, False, 10);
                button.show()
                self.label_status0=Gtk.Label(label="Status unknown")
                self.label_status0.show()
                hbox.pack_start(self.label_status0, False, False, 0)
                hbox.show()
                self.label_status1=Gtk.Label(label="")
                self.label_status1.show()
                hbox.pack_start(self.label_status1, False, False, 0)
                self.label_status2=Gtk.Label(label="")
                self.label_status2.show()
                hbox.pack_start(self.label_status2, False, False, 0)
                self.label_status3=Gtk.Label(label="")
                self.label_status3.show()
                hbox.pack_start(self.label_status3, False, False, 0)
                self.button_status(button)
                hbox.show()                                                
                vboxf_out = Gtk.VBox(homogeneous=False, spacing=10)
                vboxf = Gtk.VBox(homogeneous=False, spacing=10)
                frame = Gtk.Frame(label="Sensor")
                self.cmb_analog_val.clear()
                for i in range(0, 8):
                        hbox = Gtk.HBox(homogeneous=False, spacing=10)
                        textstr = "Chan" + str(i)
                        label = Gtk.Label(label=textstr)
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        opt = Gtk.ComboBoxText()
                        menudefault = 0
                        values = []
                        for j in range(0, 24):
                                opt.append_text(s2218.STR_SENSOR_TYPE[j])
                        if (rc == 0):
                                sensor_type = list(self.analog_cfg.sensor_type)
                                opt.set_active(sensor_type[i])
                        hbox.pack_start(opt, False, False, 0);
                        opt.index = i
                        opt.connect("changed", self.set_sensor_type)
                        opt.show()
                        label = Gtk.Label(label="Value")
                        label.show()
                        self.cmb_analog_val.append(label)
                        hbox.pack_start(label, False, False, 0);
                        hbox.show()
                        vboxf.pack_start(hbox, False, False, 0);
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                label = Gtk.Label(label="Time:")
                label.show()
                hbox.pack_start(label, False, False, 0);
                self.label_time = Gtk.Label(label="")
                self.label_time.show()
                hbox.pack_start(self.label_time, False, False, 0);
                hbox.show()
                vboxf.pack_start(hbox, False, False, 0);
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                hbox.pack_start(vboxf, False, False, 0)
                label=Gtk.Label(label=",VREF = 5.00V (ok)")
                label.show()
                scrolled = Gtk.ScrolledWindow()
                scrolled.set_hexpand(True)
                scrolled.set_vexpand(True)
                textview = Gtk.TextView()
                self.textbuf = textview.get_buffer()
                demo.scan_buf = ""
                self.textbuf.set_text("")
                textview.show()
                scrolled.add(textview)
                scrolled.show()
                hbox.pack_start(scrolled, True, True, 10)
                hbox.show()
                
                vboxf_out.pack_start(hbox, False, False, 0);
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                btn = Gtk.Button(label="Clear")
                btn.connect("clicked", self.button_clear);
                btn.show()
                hbox.show()
                hbox.pack_end(btn, False, False, 10)
                vboxf_out.pack_start(hbox, False, False, 0);

                # lastly, add Scan Sensors to Sensor frame
                vboxf2 = Gtk.VBox(homogeneous=False, spacing=10)
                frame2 = Gtk.Frame(label="Scan Sensors")
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                self.btn_start = Gtk.Button(label="Start")
                self.btn_stop = Gtk.Button(label="Stop")
                self.chk_log = Gtk.CheckButton(label="LogFile")
                self.btn_select = Gtk.Button(label="Select")
                self.lbl_status_scan = Gtk.Label(label="Idle")
                lbl_status_preamble = Gtk.Label(label="Status:")
                self.label_filename = Gtk.Label(label="")
                self.label_filename.show()
                self.lbl_status_scan.show()
                self.lbl_status_scan.set_width_chars(8)
                self.lbl_status_scan.set_max_width_chars(8)                
                lbl_status_preamble.show()
                self.btn_start.connect("clicked", self.btn_start_scan)
                self.btn_stop.connect("clicked", self.btn_stop_scan)
                self.btn_select.connect("clicked", self.btn_select_file)
                hbox.pack_start(lbl_status_preamble, False, False, 10)
                hbox.pack_start(self.lbl_status_scan, False, False, 10)
                hbox.pack_start(self.btn_start, False, False, 10)
                hbox.pack_start(self.btn_stop, False, False, 10)
                hbox.pack_start(self.chk_log, False, False, 10)
                hbox.pack_start(self.btn_select, False, False, 10)
                hbox.pack_start(self.label_filename, False, False, 10)
                vboxf2.pack_start(hbox, False, False, 10)
                frame2.add(vboxf2)
                vboxf2.show()
                hbox.show()
                self.btn_start.show()
                self.btn_stop.show()
                self.chk_log.show()
                self.btn_select.show()
                frame2.show()
                vboxf_out.pack_start(frame2, False, False, 10)
                frame.add(vboxf_out)
                vboxf.show();
                vboxf_out.show();
                
                frame.show()
                vbox.pack_start(frame, False, False, 0)
                self.add_view_frame(vbox)
                vbox.show()
                return vbox
        def setup_board(self):
                #enumerate all the 2218s in the system
                vbox = Gtk.VBox(homogeneous=False, spacing=10)
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                hbox.show()
                self.hbox_device = hbox
                vbox.pack_start(hbox, False, False, 0)
                vbox.show()
                self.vbox_device = vbox
                #create device scanning thread
                self.thread_device = device_thread(self)
                self.thread_device.start()
                return self.vbox_device
        def setup_gpio(self):
                #core banner
                vbox = Gtk.VBox(homogeneous=False, spacing=10)
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                label=Gtk.Label(label="GPIO Settings")

                cfg_gpio = s2218.gpio_settings()
                rc = s2218.GetGpioSettings(self.cur_device, cfg_gpio)
                rc, states = s2218.GetGpioInputs(self.cur_device)
                hbox.pack_start(label, False, False, 0)
                vbox.pack_start(hbox, False, False, 0)
                label.show()                
                hbox.show()
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                rc, devver = s2218.GetDevVersions(self.cur_device)
                txt = "Device Index: " + str(self.cur_device) + ", Version Info: "
                if (rc == 0):
                        txt += "dev " + str(devver.major) + "." + str(devver.minor) + "." + str(devver.build) + ", "
                        txt += "smad " + str(devver.smad_fw/100) + ", pwd rev " + chr(devver.pwb_rev)
                        label=Gtk.Label(label=txt)
                else:
                        txt = "Unknown. (error: " + str(rc) + ")"
                        label=Gtk.Label(label=txt)
                label.show()
                hbox.pack_start(label, False, False, 0)
                hbox.show()
                vbox.pack_start(hbox, False, False, 0)

                vboxf_gpio = Gtk.VBox(homogeneous=False, spacing=10)
                vboxf_gpio.show()
                frame_gpio = Gtk.Frame(label="GPIO")
                frame_dir = Gtk.Frame(label="Direction")
                frame_debounce = Gtk.Frame(label="Debounce(0-255ms)")
                frame_bias = Gtk.Frame(label="Pull-up,Pull-down bias")
                frame_value = Gtk.Frame(label="Value")
                frame_gpio.show()
                frame_dir.show()
                frame_debounce.show()
                frame_bias.show()
                frame_value.show()                

                vboxf_dir = Gtk.VBox(homogeneous=False, spacing=10)
                for i in range(0, 8):
                        
                        hbox = Gtk.HBox(homogeneous=False, spacing=10)
                        textstr = "GPIO " + str(i)
                        label = Gtk.Label(label=textstr)
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        opt = Gtk.ComboBoxText()
                        menudefault = 0
                        values = []
                        opt.append_text("Input")
                        opt.append_text("Output")
                        hbox.pack_start(opt, False, False, 0);
                        # val = 0 == input. pin_direction = 1
                        # val = 1 == output. pin_direction = 0 
                        if (cfg_gpio.config.pin_directions & (1 << i) != 0):
                                dir = 0 #input
                        else:
                                dir = 1 #output
                        opt.set_active(dir)
                        opt.connect("changed", self.set_gpio_dir)
                        opt.index = i
                        opt.show()
                        hbox.show()
                        vboxf_dir.pack_start(hbox, False, False, 0);

                vboxf_dir.show()
                frame_dir.add(vboxf_dir)
                vboxf_debounce = Gtk.VBox(homogeneous=False, spacing=10)
                for i in range(0, 8):
                        hbox = Gtk.HBox(homogeneous=False, spacing=10)
                        textstr = "Rising"
                        label = Gtk.Label(label=textstr)
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        entry = Gtk.Entry(text="0", max_width_chars=4)
                        txt = cfg_gpio.config.dbtime[i].rising
                        entry.set_text(str(txt))
                        entry.set_max_width_chars(4)
                        entry.index = i
                        entry.connect("changed", self.set_gpio_rising)
                        entry.show()
                        hbox.pack_start(entry, False, False, 0)

                        textstr = "Falling"
                        label = Gtk.Label(label=textstr, )
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        entry = Gtk.Entry(text="0")
                        txt = cfg_gpio.config.dbtime[i].falling
                        entry.set_text(str(txt))
                        entry.index = i
                        entry.connect("changed", self.set_gpio_falling)
                        entry.show()
                        hbox.pack_start(entry, False, False, 0)
                        hbox.show()
                        vboxf_debounce.pack_start(hbox, False, False, 0);

                frame_debounce.add(vboxf_debounce)
                vboxf_debounce.show()
                
                vboxf_bias = Gtk.VBox(homogeneous=False, spacing=10)
                for i in range(0, 8):
                        hbox = Gtk.HBox(homogeneous=False, spacing=10)
                        textstr = "bias"
                        label = Gtk.Label(label=textstr)
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        opt = Gtk.ComboBoxText()
                        menudefault = 0
                        values = []
                        opt.append_text("Off")
                        opt.append_text("On")
                        hbox.pack_start(opt, False, False, 0);
                        opt.show()
                        opt.index = i;
                        if (cfg_gpio.config.bias_enables & (1 << i) != 0):
                                dir = 1
                        else:
                                dir = 0
                        opt.set_active(dir)
                        opt.connect("changed", self.set_gpio_bias)
                        textstr = "polarity"
                        label = Gtk.Label(label=textstr)
                        hbox.pack_start(label, False, False, 0);
                        label.show()
                        opt = Gtk.ComboBoxText()
                        menudefault = 0
                        values = []
                        opt.append_text("Down")
                        opt.append_text("Up")
                        hbox.pack_start(opt, False, False, 0);
                        opt.show()

                        if (cfg_gpio.config.bias_polarities & (1 << i) != 0):
                                dir = 1
                        else:
                                dir = 0
                        opt.set_active(dir)
                        opt.index = i
                        opt.connect("changed", self.set_gpio_polarity)
                        hbox.show()
                        vboxf_bias.pack_start(hbox, False, False, 0);

                frame_bias.add(vboxf_bias)
                vboxf_bias.show()
                
                vboxf_value = Gtk.VBox(homogeneous=False, spacing=10)
                self.cmb_gpio_val.clear()
                for i in range(0, 8):
                        hbox = Gtk.HBox(homogeneous=False, spacing=10)
                        label = Gtk.Label(label="value")
                        hbox.pack_start(label, False, False, 10);
                        label.show()
                        opt = Gtk.ComboBoxText()
                        opt.append_text("0")
                        opt.append_text("1")
                        # val = 0 == input. pin_direction = 1
                        # val = 1 == output. pin_direction = 0 
                        if (cfg_gpio.config.pin_directions & (1 << i) == 0):
                                #output
                                opt.set_sensitive(True)
                                if (cfg_gpio.output_states & (1 << i)):
                                        opt.set_active(1)
                                else:
                                        opt.set_active(0)
                        else:
                                #input
                                opt.set_sensitive(False)
                                if (states.value & (1 << i)):
                                        opt.set_active(1)
                                else:
                                        opt.set_active(0)
                        opt.show()
                        opt.index = i
                        opt.connect("changed", self.set_gpio_output)
                        self.cmb_gpio_val.append(opt)
                        hbox.pack_start(opt, False, False, 10);
                        hbox.show()
                        vboxf_value.pack_start(hbox, False, False, 0);
                frame_value.add(vboxf_value)
                vboxf_value.show()                

                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                hbox.pack_start(frame_dir, False, False, 0)
                hbox.pack_start(frame_debounce, False, False, 0)
                hbox.pack_start(frame_bias, False, False, 0)
                hbox.pack_start(frame_value, False, False, 0)
                hbox.show()
                vboxf_gpio.pack_start(hbox, False, False, 0);
                # lastly, add Scan Sensors to GPIO frame
                hbox_readline = Gtk.HBox(homogeneous=False, spacing=10)
                vboxf2 = Gtk.VBox(homogeneous=False, spacing=10)
                frame2 = Gtk.Frame(label="Continuous Read(1s refresh)")
                hbox = Gtk.HBox(homogeneous=False, spacing=10)
                lbl_preamble = Gtk.Label(label="Status: ")
                self.lbl_status_gpio = Gtk.Label(label="Idle")
                self.btn_gpio_start = Gtk.Button(label="Start")
                self.btn_gpio_stop = Gtk.Button(label="Stop")
                self.lbl_status_gpio.show()
                self.chk_log_gpio = Gtk.CheckButton(label="LogFile")
                self.btn_gpio_select = Gtk.Button(label="Select")
                self.label_filename_gpio = Gtk.Label(label="")
                self.label_filename_gpio.show()
                demo.lbl_status_gpio.set_width_chars(8)
                demo.lbl_status_gpio.set_max_width_chars(8)
                lbl_preamble.show()
                self.btn_gpio_start.show()
                self.btn_gpio_stop.show()
                self.btn_gpio_start.connect("clicked", self.btn_start_gpio)
                self.btn_gpio_stop.connect("clicked", self.btn_stop_gpio)
                self.btn_gpio_select.connect("clicked", self.btn_select_gpio)
                hbox.pack_start(lbl_preamble, False, False, 10)
                hbox.pack_start(self.lbl_status_gpio, False, False, 10)
                hbox.pack_start(self.btn_gpio_start, False, False, 10)
                hbox.pack_start(self.btn_gpio_stop, False, False, 10)
                hbox.pack_start(self.chk_log_gpio, False, False, 10)
                hbox.pack_start(self.btn_gpio_select, False, False, 10)
                hbox.pack_start(self.label_filename_gpio, False, False, 10)

                vboxf2.pack_start(hbox, False, False, 10)
                frame2.add(vboxf2)
                vboxf2.show()
                hbox.show()

                btn_read = Gtk.Button(label="Refresh")
                btn_read.connect("clicked", self.btn_read_gpio_inputs)
                btn_read.show()
                hbox_readline.pack_end(btn_read, False, False, 10)
                hbox_readline.pack_start(frame2, False, False, 10)
                hbox_readline.show()

                self.chk_log_gpio.show()
                self.btn_gpio_select.show()
                frame2.show()
                vboxf_gpio.pack_start(hbox_readline, False, False, 10)
                frame_gpio.add(vboxf_gpio)
                vboxf_gpio.show();
                frame_gpio.show()
                vbox.pack_start(frame_gpio, False, False, 0)
                self.add_view_frame(vbox)
                vbox.show()
                dprint("test setup_gpio")                
                return vbox
        def main(self):
                Gtk.main()
                self.stop_all_threads()

if __name__ == "__main__":
        demo = Demo()
        demo.main()
