///////////////////////////////////////////////////////////////////////////
// File        : frm2218.h
// Function    : Include file for Sensoray 2218 PIC firmware
// Author      : Jim Lamberson
// Copyright   : Copyright (C) 2017 Sensoray
///////////////////////////////////////////////////////////////////////////

#ifndef _INC_FRM2218_H_
#define	_INC_FRM2218_H_

#include <p33Exxxx.h>       // microcontroller equates
#include <stdbool.h>
#include <string.h>         // memcpy

#include "api2218.h"

#define DEBUG_STACK
///////////////////////////////////////////////////////
// FIRMWARE VERSION NUMBER -- CHANGE EVERY RELEASE   //
//                                                   //
#define FW_VERSION_MAJOR            1                //
#define FW_VERSION_MINOR            0                //
#define FW_VERSION_BUILD            1                //
//                                                   //
// Revision log:                                     //
//  1.0.0 - Initial release.                         //
//  1.0.1 - Update test                              //
//
///////////////////////////////////////////////////////


// Constants ===================================================================

// Configuration constants --------------------------

// Kernel
#define MAX_TASKS                   10          // Maximum number of tasks allowed. This can be set to arbitrary size but must be >= number of created tasks.
#define TASK_STACK_MINWORDS         100         // Minimum number of words to reserve for a task stack
#define TASK_NAMESIZE               4           // Number of chars comprising a task name, including asciiz terminator
#define MEMPOOL_WORDS               1024        // Reserve this much RAM (in 16-bit words) for use as private task stacks.

// Manifest constants -------------------------------

#define TRUE                        1
#define FALSE                       0
#define INFINITE_WAIT               0xFFFF      // Task delay time used to disable task timer

// Timer configuration register constants
#define STOP_TIMER_IN_IDLE_MODE     0x2000      // Halt counter when device in idle mode
#define TIMER_ON                    0x8000      // Enable counting
#define TIMER_PRESCALER_1           0x0000      // No clock prescaling
#define TIMER_PRESCALER_8           0x0010      // Prescale clock x8
#define TIMER_PRESCALER_64          0x0020      // Prescale clock x64
#define TIMER_PRESCALER_256         0x0030      // Prescale clock x256
#define TIMER_CLOCK_SYNC            0x0004
#define TIMER_EXTERNAL_CLOCK        0x0002      // Use external clock input (vs. internal system clock)
// RAM-based pointers to interrupt handler functions. On this CPU, the interrupt vector table is located at a fixed location in flash memory.
// These pointers enable interrupt functions to be relocated without changing the vector table. This makes it possible for application
// firmware to substitute its own ISRs in place of the bootloader firmware ISRs, simply by changing these pointers. Upon interrupt,
// the vector table directs the CPU to a fixed, top-level ISR that belongs to the bootloader firmware but is used by the application firmware
// as well. The top-level ISR will then call the firmware-specific ISR function through a pointer stored in RAM.
#define ISR_PTR(ADDR)               (*((ISR *)(ADDR)))              // Pointers to firmware-specific ISRs:
#define IsrFuncUsb                  ISR_PTR(0x7FF0)                 //  USB
#define IsrFuncTickTimer            ISR_PTR(0x7FF2)                 //  TIMER1 -- available for future use; no need to initialize this unless used in the future

// PERIPHERAL USAGE =========================================================================
//
// Timers:
// TIMER2       unused
// TIMER3       Scheduler tick timer
// TIMER4       Task timer (used for diagnostics only)
// TIMER5       unused
// TIMER6       unused
// TIMER7       unused
// TIMER8       unused
//
// PWMs: not used in this version, but available for future use
// RE0/PWM1L    GPIO0
// RE1/PWM1H    GPIO1
// RE2/PWM2L    GPIO2
// RE3/PWM2H    GPIO3
// RE4/PWM3L    GPIO4
// RE5/PWM3H    GPIO5
// RE6/PWM4L    GPIO6
// RE7/PWM4H    GPIO7
//
// Analog inputs:
// AN14         VMON3 - (+15V monitor)  - nominal 2.63V
// AN10         VMON2 - (+5V monitor)   - nominal 2.5V
// AN9          VMON1 - (+3.3V monitor) - nominal 1.65V
// AN8          VMON0 - (-12V monitor)  - nominal 1.48V
#define CFG_ANSELB  (MASKB_VMON3 | MASKB_VMON2 | MASKB_VMON1 | MASKB_VMON0)
//
//                                              // PIN          TYPE  DIR  BIAS  SIGNAL     FUNCTION
//                                              // -----------  ----  ---  ----  ---------  ---------------------------------
#define MASKB_EEPOT_CS              (1 << 15)   // RB15         out   o    -     EPCSn      EEPOT chip select; held at 0V when EEPOT is not being calibrated
#define MASKB_VMON3                 (1 << 14)   // RB14/AN14    ain   i    -     VMON3      Power supply metering
#define MASKB_DEVID1                (1 << 13)   // RB13         in    i    pu    ID1        USB device address switch
#define MASKB_DEVID2                (1 << 12)   // RB12         in    i    pu    ID2        USB device address switch
#define MASKB_DEVID3                (1 << 11)   // RB11         in    i    pu    ID3        USB device address switch
#define MASKB_VMON2                 (1 << 10)   // RB10/AN10    ain   i    -     VMON2      Power supply metering
#define MASKB_VMON1                 (1 << 9)    // RB9/AN9      ain   i    -     VMON1      Power supply metering
#define MASKB_VMON0                 (1 << 8)    // RB8/AN8      ain   i    -     VMON0      Power supply metering
#define MASKB_PGD                   (1 << 7)    // RB7/PGED1    alt   i    -     (PGD)      Programming data
#define MASKB_PGC                   (1 << 6)    // RB6/PGEC1    alt   i    -     (PGC)      Programming clock
#define MASKB_IRQ                   (1 << 5)    // RB5          in    i    pd    IRQ        FPGA IRQ
#define MASKB_RST                   (1 << 4)    // RB4          out   o    -     RST        FPGA reset (smad reset)
#define MASKB_RD                    (1 << 3)    // RB3          out   o    -     nRD        FPGA read strobe 
#define MASKB_WR                    (1 << 2)    // RB2          out   o    -     nWR        FPGA write strobe
#define MASKB_ADDRBUS               (3 << 0)    // RB1:RB0      out   o    pd    A[1:0]     FPGA address bus
// ---------------------------------------------------------------------------------------------------------------------------
#define MASKC_LED_HEARTBEAT         (1 << 15)   // RC15         out   o    -     HBTLED     Heartbeat LED (active low)
#define MASKC_PWBREV3               (1 << 14)   // RC14         in    i    pd    REV3       PWB rev code
#define MASKC_PWBREV2               (1 << 13)   // RC13         in    i    pd    REV2       PWB rev code
#define MASKC_OSCIN                 (1 << 12)   // RC12/OSC1    alt   i    -     16MHZ      Master clock input
// ---------------------------------------------------------------------------------------------------------------------------
#define MASKD_PWBREV1               (1 << 11)   // RD11         in    i    pd    REV1       PWB rev code
#define MASKD_PWBREV0               (1 << 10)   // RD10         in    i    pd    REV0       PWB rev code
#define MASKD_DEVID0                (1 << 9)    // RD9          in    i    pu    ID0        USB device address switch
#define MASKD_SPARE_RD8             (1 << 8)    // RD8          out   o    -     (RD8)      ** Spare
#define MASKD_DATABUS               (0x00FF)    // RD7:RD0      io    i    pd    D[7:0]     FPGA data bus; tri-stated (configured as input) except when writing to FPGA
// ---------------------------------------------------------------------------------------------------------------------------
#define MASKE_GPIO                  (0x00FF)    // RE7:RE0      io    i    -     GPIO[7:0]  General-purpose I/Os for application use; initially config as floating inputs
// ---------------------------------------------------------------------------------------------------------------------------
#define MASKF_FPGA_CLOCK            (1 << 5)    // RF5/RP101    alt   i    -     FPGACLK    FPGA clock; gated off when FPGA is unpowered
#define MASKF_LED_FAULT             (1 << 4)    // RF4          out   o    -     FLTLED     Fault LED (active low)
#define MASKF_FPGA_CONF_DONE        (1 << 3)    // RF3          in    i    -     CONF_DONE  FPGA configuration completed (active low, external pull-up)
#define MASKF_PPOWER_ENABLE         (1 << 1)    // RF1          out   o    -     APEN       Enable analog positive power supply (active high, external pull-down)
#define MASKF_NPOWER_ENABLE         (1 << 0)    // RF0          out   o    -     ANEN       Enable analog negative power supply (active high, external pull-down)
// ---------------------------------------------------------------------------------------------------------------------------
#define MASKG_EEPOT_DIR             (1 << 9)    // RG9          out   o    -     EPUP       EEPOT wiper direction control; held at 0V when EEPOT is not being calibrated
#define MASKG_EEPOT_STEP            (1 << 8)    // RG8          out   o    -     EPINCn     EEPOT wiper step command (negative pulse to step); held at 0V when EEPOT is not being calibrated
#define MASKG_DPOWER_ENABLE         (1 << 7)    // RG7          out   o    -     DEN        Enable FPGA power supply (active high, external pull-down)
#define MASKG_SMAD_FAULT            (1 << 6)    // RG6          in    i    pd    FLT        Analog subsystem fault (active low)


// Initial pin directions: 1=in, 0=out
#define CFG_TRISB                   (MASKB_VMON3 | MASKB_DEVID1 | MASKB_DEVID2 | MASKB_DEVID3 | MASKB_VMON2 | MASKB_VMON1 | MASKB_VMON0 | MASKB_PGD | MASKB_PGC | MASKB_IRQ)
#define CFG_TRISC                   (MASKC_PWBREV3 | MASKC_PWBREV2 | MASKC_OSCIN)
#define CFG_TRISD                   (MASKD_PWBREV1 | MASKD_PWBREV0 | MASKD_DEVID0 | MASKD_DATABUS)
#define CFG_TRISE                   MASKE_GPIO     // initially config application GPIOs as inputs
#define CFG_TRISF                   (MASKF_FPGA_CLOCK | MASKF_FPGA_CONF_DONE)
#define CFG_TRISG                   MASKG_SMAD_FAULT   // EEPOT DIR

// Weak pull-downs
#define CFG_PD_B                    (MASKB_IRQ | MASKB_ADDRBUS)
#define CFG_PD_C                    (MASKC_PWBREV3 | MASKC_PWBREV2)
#define CFG_PD_D                    (MASKD_PWBREV1 | MASKD_PWBREV0 | MASKD_DATABUS)
#define CFG_PD_E                    0
#define CFG_PD_F                    0
#define CFG_PD_G                    MASKG_SMAD_FAULT

// Weak pull-ups
#define CFG_PU_B                    (MASKB_DEVID1 | MASKB_DEVID2 | MASKB_DEVID3)
#define CFG_PU_C                    0
#define CFG_PU_D                    MASKD_DEVID0
#define CFG_PU_E                    0
#define CFG_PU_F                    0
#define CFG_PU_G                    0


// Types =========================================================================

// Function pointers
typedef void (*TASK)(void);
typedef void (*ISR)(void);

// Word/byte conversions ------------------
// NOTE: THESE TARGET THE LITTLE-ENDIAN PIC -- MUST BE CHANGED IF PORTING TO BIG-ENDIAN CPU!
typedef union HOST16 {  // struct used for efficient u16/u8 dual-use values (host/little endian)
    u16 uword;
    s16 sword;
    u16 rising_falling; // msb_lsb
    u8  b[2];
    struct {    // host byte order
        u8 lsb;
        u8 msb;
    };
    struct {    // host byte order
        u8 falling;
        u8 rising;
    };
} HOST16;

static inline u16 HTON16(u16 src)     // convert host (little) endian u16 to network (big) endian
{
    register HOST16 dst;
    dst.msb = ((HOST16)src).lsb;
    dst.lsb = ((HOST16)src).msb;
    return dst.uword;
}

static inline u8 MSB(u16 src)       // extract MSB from host (little) endian u16
{
    return ((HOST16)src).msb;
}

static inline u8 LSB(u16 src)
{
    return ((HOST16)src).lsb;
}
// GPIO -----------------------------------
typedef struct GPIO_CONFIG {                    // GPIO configurations (MUST BE IDENTICAL TO S2218_GPIO_CONFIG):
    HOST16  DebounceInterval[S2218_NUM_GPIOS];  //   Rising/falling debounce intervals[gpionum]: 0=disabled; else=milliseconds.
    u8      PinDirections;                      //   Pin direction flags: 1=input; 0=output.
    u8      BiasEnables;                        //   Pullup/down enable flags: 1=enable; 0=float;
    u8      BiasPolarities;                     //   Pullup/down type flags: 1=up; 0=down.
    u8      spare;
} GPIO_CONFIG;

typedef struct GPIO_SETTINGS {          // GPIO settings (MUST BE IDENTICAL TO S2218_GPIO_SETTINGS):
    GPIO_CONFIG Config;                 //   GPIO configurations.
    HOST16      CapEnables;             //   Rising/falling edge capture enables.
    u8          OutputStates;           //   Pin output states (valid only for pins with direction=0).
    u8          spare;
} GPIO_SETTINGS;
// Smad image ----------------------------
// flags bits
#define LIMITENABLE_HI  (1 << 3)        // Enable high threshold checking
#define LIMITENABLE_LO  (1 << 2)        // Enable low threshold checking
#define ALARMENABLE_HI  (1 << 1)        // Enable high threshold alarms
#define ALARMENABLE_LO  (1 << 0)        // Enable high threshold alarms
typedef struct CHAN_IMAGE {
    s16         rawData;                // Raw sensor data from most recent scan.
    s16         alarmLimitHi;           // High sensor data limit.
    s16         alarmLimitLo;           // Low sensor data limit.
    u8          apiSensorType;          // API sensor type codes (vs. smad sensor codes).
    u8          flags;                  // Limit config/status: (x, x, VIOLATION_H, VIOLATION_L, LIMITENABLE_HI, LIMITENABLE_LO, ALARMENABLE_HI, ALARMENABLE_LO).
} CHAN_IMAGE;

typedef struct SMAD_IMAGE {             // Copy of current smad configuration/status --------
    HOST16      alarmFlags;             //   Alarm flags: msb=high alarms; lsb=low alarms.
    HOST16      limitVios;              //   Limit violations: msb=high limits; lsb=low limits.
    u8          openFlags;              //   Open sensor flags.
    u8          TcCondEnables;          //   Thermocouple conditioner enables.
    CHAN_IMAGE  chan[S2218_NUM_CHANS];  //   Channel info.
} SMAD_IMAGE;

// Macros ==========================================================
// Hardware-dependent inline functions
#define DisableInterrupts()     SET_CPU_IPL(7)
#define EnableInterrupts()      SET_CPU_IPL(0)

// Program state of RD8 (spare GPIO); may be used for custom diagnostics

#define SetSpareRD8             LATD |= MASKD_SPARE_RD8
#define ClrSpareRD8             LATD &= ~MASKD_SPARE_RD8

// Global functions ====================================================
// System

void    KernelInit(void);
int     CreateTask(char name[4], TASK func, uint stackwords);
int     GetTaskHandle(void);
void    Yield(int nticks);
void    WaitForSignal(void);
void    ExecuteTasks(void);
void    SignalTask(int taskid);
char    *TaskName(int taskid);
#ifdef DEBUG_STACK
uint    StackMargin(int taskid);
#endif
uint    TaskExecTime(int taskid);
uint    LoopExecTime(void);

// Task threads

void    TaskUsb(void);              // USB manager
void    TaskCmdProc(void);          // Command processor
void    TaskGpioMon(void);          // GPIO input monitor
void    TaskSmadScanner(void);      // ADC sampler
void    TaskStateManager(void);     // System state manager
void    TaskUpdateLeds(void);       // Status LED manager
void    TaskPowerMon(void);         // Power monitor
// Utility (main)
u8      ReadPwbVersion(void);
u8      *BuildNotification(void);
void    ScheduleCmdProc(u8 *rxbuf);
// Smad command wrappers and utilities (commands.c)
u8      CmdReset(void);
u8      CmdReadPort(int addr, u8 *data);  // returns errcode
u8      CmdWritePort(int addr, u8 data);  // returns errcode
u8      CmdSetSensorType(u8 chan, u8 typecode);
u8      CmdSetTcondEnables(u8 mask, u8 enables);
u8      CmdAdjustEepot(u8 dir, u8 nsteps);

// Utility

void    RegisterSensorType(u8 chan, u8 typecode);
void    SetLimitItem(u8 chan, u8 itemtype, s16 value);
u8      SmadInPort(int addr);  // returns data
void    SmadOutPort(int addr, u8 data);

// Smad commands and other smad-related low-level info (smadcomm.c)

u8      SmadReadChannel(u8 chan, s16 *data);
u8      SmadSetSensorType(u8 chan, u8 typecode);
u8      SmadReadAllChannels(s16 data[S2218_NUM_CHANS]);
u8      SmadSetOpenValues(u8 flags);
u8      SmadSetGaugeZero(u8 chan);
u8      SmadSetGaugeSpan(u8 chan, u16 data);
u8      SmadTareGauge(u8 chan);
u8      SmadSetGaugeCal(u8 chan, u8 *data);
u8      SmadGetGaugeCal(u8 chan, u8 *data);
u8      SmadReadModel(u16 *model);
u8      SmadReadFwVersion(u16 *ver);
u8      SmadReadBoardTemperature(u16 *temp);
u8      SmadCalibrate(u8 chan, u8 calcode, u16 data, u8 *ref);
u8      SmadHaltOnSetup(u8 ctrlreg);
u8      SmadGetExternalSnapshot(u8 chan, u8 *buf);  // buf[8]
u8      SmadReadEeprom(u8 *buf);                    // buf[4*7]
u8      SmadGetInternalSnapshot(u8 *buf);           // buf[4*9]
u8      SmadReadModel(u16 *model);
u8      SmadReadFwVersion(u16 *ver);
u8      SmadHaltOnChannel(u8 logchan);              // logchan: 0 to 7 = external, 16+0 to 16+8 = internal
u8      SmadWriteSN(u8 *sn);                        // sn[4]
u8      SmadReadSN(u8 *sn);
u8      SmadSetTcondEnables(u8 enables);
boole   SmadAdjustEepot(u8 dir, u8 nsteps);

// Sensor scanner (scanner.c)

void    SuspendScanning(void);
void    RestoreScanning(void);
boole   ScanningEnabled(void);
u16     GetPumpInterval(void);
u8      SetPumpInterval(u16 interval);
void    ScannerTick(void);
void    ExecScan(void);

// GPIO (gpiomon.c)

void    GpioWriteConfig(GPIO_CONFIG *cfg);
u8      GpioSetDatareportTrigMask(u16 mask);
u16     GpioGetDatareportTrigMask(void);
void    GpioReadSettings(GPIO_SETTINGS *settings);
void    GpioWriteOutputs(u8 states);
u8      GpioReadInputs(void);
void    GpioWriteCapEnables(HOST16 flags);
HOST16  GpioReadCapEvents(void);

// Global storage ============================================================

extern volatile u16     SystemTime;             // System uptime: resolution = 1 millisecond; overflows every 65.536 seconds.
extern volatile u16     TimestampMsec;          // System uptime milliseconds (rolls over every 1000 ms).
extern volatile u16     TimestampSecLo;         // System uptime seconds, low word (increments every TimestampMsec reset).
extern volatile u16     TimestampSecHi;         // System uptime seconds, high word (increments every TimestampSecLo rollover).
extern volatile u16     Meter[4];               // Power supply metering raw adc values.
extern volatile boole   ReqReset;               // Smad reset trigger.
extern volatile boole   TrigEnable;             // Permissive for transmitting data reports.
extern volatile boole   ReqScan;                // Scan trigger.
extern volatile boole   SmadHalted;             // While smad halted-on-chan/setup, prohibits all smad comm except reset.
extern volatile boole   SmadDiagLock;           // During smad diagnostic port read/write sequences, prohibits all smad comm except direct port read/write.
extern volatile u8      SmadState;              // Smad system state.
extern volatile u8      PowerEnables;           // Power enable requests: (x, x, x, x, x, +3.3V, +15V, -12.25V).
extern volatile u8      DPowerState;            // FPGA power state.
extern volatile u8      PPowerState;            // Analog positive (+15V) power state.
extern volatile u8      NPowerState;            // Analog negative (-12.25V) power state.
extern volatile u8      ComState;               // Smad communication state.
extern volatile u8      PowerGoodFlags;         // Smad PowerGood flags: (0, 0, 0, 0, +15V, +5V, +3.3V, -12V).
extern volatile u8      NotifyPending;          // Flags that indicate a notification message must be sent to the host.
extern volatile u8      *PendingReplyBuf;       // Pending command response message. Set to NULL to indicate no message pending.
extern u8               LogicalDeviceNumber;    // 2218 device identifier read from dip switches.
extern u8               PwbVersion;             // Board hardware version: 0=rev0, 1=revA, 2=revB, etc.
extern SMAD_IMAGE       smad;                   // Smad (analog subsystem, including FPGA) shadow image and status info.

#endif	/* _INC_FRM2218_H_ */

