/////////////////////////////////////////////////////////////////////////////////////
// File        : mid2218.h
// Function    : Common middleware/firmware defines for Sensoray 2218
// Author      : Jim Lamberson
// Copyright   : Copyright (C) 2017-2023 Sensoray
/////////////////////////////////////////////////////////////////////////////////////

#ifndef _INC_MID2218_H_
#define	_INC_MID2218_H_

// Constants ======================================================================

#define USB_MSG_SIZE                    64          // USB tx/rx packet size

// Flag bits in NotifyPending that indicate notification reason
#define NOTIFY_NONE                     0           // Don't send a notification message
#define NOTIFY_DATA                     (1 << 0)    // Sensor data report
#define NOTIFY_EDGE                     (1 << 1)    // GPIO edge(s) captured
#define NOTIFY_ALARM                    (1 << 2)    // Alarm limit violation
#define NOTIFY_STATE                    (1 << 3)    // Automatic status report (smad system state transitioned to power off, online, or fatal error).
#define NOTIFY_SOFT                     (1 << 4)    // Software-triggered status report (invoked by host)

// Bit masks for PowerEnable request flags: (x, x, x, x, x, +3.3V, +15V, -12.25V)
#define REQ_POWER_N                     (1 << 0)    // Analog -12.25V
#define REQ_POWER_P                     (1 << 1)    // Analog +15V
#define REQ_POWER_D                     (1 << 2)    // FPGA +3.3V

// Bit masks for PowerGoodFlags: (x, x, x, x, VMON3, VMON2, VMON1, VMON0)
#define POWERGOOD_N12V                  (1 << 0)    // VMON0 : -12.25V
#define POWERGOOD_3R3V                  (1 << 1)    // VMON1 : +3.3V
#define POWERGOOD_5V                    (1 << 2)    // VMON2 : +5V
#define POWERGOOD_P15V                  (1 << 3)    // VMON3 : _15V

#define POWERGOOD_DIGITAL               POWERGOOD_3R3V
#define POWERGOOD_ANALOG                (POWERGOOD_N12V | POWERGOOD_P15V)

// Flags used for requesting soft-triggered reports
#define REQ_STATUS_REPORT               (1 << 0)    // Request status report
#define REQ_DATA_REPORT                 (1 << 1)    // Request data report

// Maximum message sizes
#define CMD2218_CMD_MAXSIZE             62          // Command message from host
#define CMD2218_RSP_MAXSIZE             62          // Command response message to host

// Outgoing USB message types, used to demux messages at host
#define MSGTYPE_REPLY                   0           // Command reply
#define MSGTYPE_NOTIFY                  1           // Event notification

// 2218 command codes ---------------------------------------------------

#define CMD2218_NOP                     0           // Command processor is idle. DO NOT CHANGE THIS VALUE!

// Configuration and control
#define CMD2218_GET_ANALOG_CFG          1           // Read smad configuration
#define CMD2218_SET_POWER_ENABLES       2           // Enable/disable power supplies
#define CMD2218_SET_SENSOR_TYPE         3           // Configure sensor type
#define CMD2218_SET_TCCOND_ENABLES      4           // Enable/disable thermocouple conditioners

// DataReport triggering
#define CMD2218_SET_TRIG_CONFIG         5           // Configure DataReport trigger
#define CMD2218_GET_TRIG_CONFIG         6           // Get DataReport trigger config
#define CMD2218_SET_TRIG_ENABLE         7           // Enable DataReport trigger
#define CMD2218_GET_TRIG_ENABLE         8           // Get DataReport trigger enable

// Sensor data
#define CMD2218_GET_SENSOR_DATA         9           // Manually read sensor data
#define CMD2218_GET_CJ_TEMPERATURE      10          // Read cold junction temperature sensor

// Open-sensors and alarms
#define CMD2218_SET_OPEN_VALUES         11          // Set open-sensor data values to min or max
#define CMD2218_SET_LIMIT_ITEM          12          // Configure limit item for one channel
#define CMD2218_GET_LIMITS              13          // Read limits for one channel

// Strain gauge configuration
#define CMD2218_SET_GAUGE_ZERO          14          // Assert zero-load applied
#define CMD2218_SET_GAUGE_SPAN          15          // Assert full-load applied
#define CMD2218_TARE_GAUGE              16          // Adjust offset
#define CMD2218_SET_GAUGE_CALIBRATION   17          // Activate gauge calibration
#define CMD2218_GET_GAUGE_CALIBRATION   18          // Read gauge calibration

// GPIO
#define CMD2218_SET_GPIO_OUTPUTS        19          // Program outputs
#define CMD2218_GET_GPIO_INPUTS         20          // Read pin states
#define CMD2218_SET_GPIO_CAP_ENABLES    21          // Program edge capture enables
#define CMD2218_SET_GPIO_CONFIG         22          // Program pin directions/bias and debounce
#define CMD2218_GET_GPIO_SETTINGS       23          // Read gpio settings

// Calibration
#define CMD2218_CALIBRATE_REF           24          // Calibrate smad internal reference
#define CMD2218_CALIBRATE_CJ            25          // Calibrate cold-junction temperature sensor

// Diagnostic/other
#define CMD2218_GET_VERSIONS            26          // Get firmware and hardware version numbers
#define CMD2218_RAW_WRITE               27          // Write fpga port
#define CMD2218_RAW_READ                28          // Read fpga port
#define CMD2218_RAW_RW_LOCK             29          // Set/clear lock for port read/write
#define CMD2218_TRIG_NOTIFICATION       30          // Manually trigger a notification message
#define CMD2218_GET_TASK_STATISTICS     31          // Read task statistics
#define CMD2218_SMAD_RESET              32          // Invoke fpga bus reset
#define CMD2218_GET_POWER_ENABLES       33          // Get power enables
#define CMD2218_HALT_ON_CHAN            34          // Halt scanning on logical channel
#define CMD2218_HALT_ON_SETUP           35          // Halt scanning with manual front-end configuration
#define CMD2218_READ_EEPROM             36          // Read calibration data
#define CMD2218_GET_INTERNAL_SNAPSHOT   37          // Get raw data for reference standards
#define CMD2218_GET_EXTERNAL_SNAPSHOT   38          // Get raw data for sensor channel
#define CMD2218_SN_WRITE                39          // Program SN
#define CMD2218_SN_READ                 40          // Get SN

// Item select codes for CMD2218_SET_LIMIT_ITEM --------------------

#define SEL_LO_THRESHOLD    0
#define SEL_HI_THRESHOLD    1
#define SEL_LO_LIMIT_ENAB   2
#define SEL_HI_LIMIT_ENAB   3
#define SEL_LO_ALARM_ENAB   4
#define SEL_HI_ALARM_ENAB   5

// USB notification message ----------------------------

typedef struct NOTIFY_MSG {     // **** ALL MULTI-BYTE VALUES ARE BIG-ENDIAN! ****
    u8  msgtype;                // Message type: always MSGTYPE_NOTIFY
    u8  EventFlags;             // Flags indicating portions of message that need attention.
                                // Group 16-bit values for alignment in packed struct:
    u16 NotificationID;         //  Message ID; incremented by device every time it sends a notification message.
    u16 TimestampMsec;          //  Timestamp fractional seconds.
    u16 TimestampSecHi;         //  Timestamp 32-bit seconds, high word.
    u16 TimestampSecLo;         //  Timestamp 32-bit seconds, low word.
    u16 Meter[4];               //  Voltage metering raw adc values.
    u8  AlarmFlags[2];          //  High|low alarm flags (auto reset upon read).
    u8  LimitVios[2];           //  High|low limit violation flags.
    u16 SensorData[8];          //  Data from 8 sensor channels.
    u8  GpioEvents[2];          //  Rise|fall GPIO captured edges.
                                // Remaining byte values:
    u8  GpioPinStates;          //  GPIO pin states.
    u8  SmadState;              //  SMAD_STATE.
    u8  ComState;               //  SMAD_COMSTATE.
    u8  PowerEnables;           //  Power supply enable permissives: (x,x,x,x,x,x,digital,analog).
    u8  PowerGood;              //  PowerGood flags.
    u8  DPowerState;            //  Digital POWER_STATE
    u8  PPowerState;            //  +Smad POWER_STATE
    u8  NPowerState;            //  -Smad POWER_STATE
} NOTIFY_MSG;

// Command response message ----------------------------

#define RSPHDR_LEN      4       // Response header size

typedef struct RESPONSE_MSG {               // Header:
    u8 msgtype;                             //   Message type: always MSGTYPE_REPLY.
    u8 opcode;                              //   Echoed command code.
    u8 rc;                                  //   Command execution error/return code.
    u8 datalen;                             //   Response data size in bytes.
                                            // Payload:
    u8 data[USB_MSG_SIZE - RSPHDR_LEN];     //   Command data.
} RESPONSE_MSG;

typedef union RXPKT {                   // RECEIVED PACKET VIEWS ----------
    u32             align_u32;          // (ensure quadlet-aligned struct)
    u8              buf[USB_MSG_SIZE];  // Raw message buffer.
    u8              msgtype;            // All messages start with message_type code.
    NOTIFY_MSG      ev;                 // Event message.
    RESPONSE_MSG    rsp;                // Command response message.
} RXPKT;

// USB command message -----------------------

#define CMDHDR_LEN     2       // Command header size

typedef union CMD {
    u16 align_u16;                      // (ensure word-aligned struct)
    u8  buf[USB_MSG_SIZE];              // Raw message buffer.
    struct {                            // Structured message:
                                            // Header:
        u8 opcode;                          //   Command code.
        u8 datalen;                         //   Command data size in bytes (excluding opcode).
                                            // Payload:
        u8 data[USB_MSG_SIZE - CMDHDR_LEN]; //   Command data.
    };
} CMD;

// USB command transaction ---------------------

typedef union XACT {
    u32             align_u16;          // (ensure word-aligned struct)
    struct {
        CMD             cmd;
        RESPONSE_MSG    rsp;
    };
} XACT;

// Sensor data thresholds --------------------------

typedef struct DEV_ALARM_CONFIG {           // Sensor alarm configuration for one sensor channel:
    s16                 LoLimit;            //   Low sensor data threshold (smad native units).
    s16                 HiLimit;            //   High sensor data threshold (smad native units).
    u8                  Flags;              //   Limit/alarm enable flags: (x, x, x, x, LIMITENABLE_HI, LIMITENABLE_LO, ALARMENABLE_HI, ALARMENABLE_LO).
} DEV_ALARM_CONFIG;

// FSM states ---------------------------

typedef enum POWER_STATE {                  // Power state:
    PS_UNPOWERED        = 0,                //   Normal: Unpowered; waiting for permissive
    PS_HOLDOFF_DELAY    = 1,                //   Normal: Unpowered; waiting for Dpower settle (applies only to +15V and -12.25V analog supplies)
    PS_POWER_UP         = 2,                //   Normal: Powering up; waiting for power good
    PS_POWERED          = 3,                //   Normal: Powered and operating normally
    PS_POWER_DROPPED    = 4,                //   Fault:  Power unexpectedly dropped out
    PS_POWERUP_TIMEOUT  = 5                 //   Fault:  Power failed to start
} POWER_STATE;

typedef enum SMAD_STATE {
    SS_UNPOWERED        = 0,                // Normal: Unpowered
    SS_RESETTING        = 1,                // Normal: Initializing
    SS_BOOTING          = 2,                // Normal: Booting; waiting for FLT to clear
    SS_RUNNING          = 3,                // Normal: Online
    SS_FAULT            = 4,                // Fault:  Boot failed (persistent FLT) or spurious FLT
} SMAD_STATE;

// Smad communication states ---------------------

// Note: these also serve as error codes for some command functions.

typedef enum SMAD_COMSTATE {
    CS_ONLINE       = S2218_ERR_OK,       // Normal operation
    CS_OFFLINE      = S2218_ERR_OFFLINE,  // Smad system state != SS_RUNNING
    CS_CRMT_FAIL    = S2218_ERR_CRMT,     // CRMT handshake timeout
    CS_DAV_FAIL     = S2218_ERR_DAV       // DAV handshake timeout
} SMAD_COMSTATE;

#endif // ifdef _INC_MID2218_H_
