# Python bindings for the 2218 middleware api
# Copyright(c) 2024 Sensoray Co., Inc.

from ctypes import *
import os

debug = False
def dprint(*args):
        if debug:
                for x in args:
                        print(x,)
                print


# Constants ----------------------------------------
S2218_MAX_DEVICES           = 16      # Maximum number of 2218s connected to a host (limited by device selector rotary switch).
S2218_NUM_CHANS             = 8       #of sensor channels.
S2218_NUM_REFS              = 9       # Number of internal reference standards.
S2218_NUM_GPIOS             = 8       # Model 2218 has 8 GPIOs.
S2218_INFINITE_WAIT         = (~0)    # Infinite wait time for blocking functions.
## Error codes
S2218_ERR_OK                = 0       # Success
## Device errors
S2218_ERR_OPCODE            = 1    # Command not recognized.
S2218_ERR_ARG               = 2       # Illegal argument value.
S2218_ERR_UNPOWERED         = 3       # Analog circuitry unpowered.
S2218_ERR_OFFLINE           = 4       # Offline (smad.state != SS_RUNNING)
S2218_ERR_CRMT              = 5       # Logged timeout waiting for CRMT.
S2218_ERR_DAV               = 6       # Logged timeout waiting for DAV.
S2218_ERR_NO_EFFECT         = 7       # Command has no effect (e.g., will not cause state change or send associated notification)
S2218_ERR_LOCKOUT           = 8       # Command prohibited due to in-progress diagnostic
S2218_ERR_SCANNING          = 9       # Command prohibited during analog scanning (e.g., SetSensorType)
## API errors
S2218_ERR_TIMEOUT           = 20      # Timed out waiting for device. If relevant, use more specific LOCK or SIG timeout
S2218_ERR_FAULT             = 21      # Couldn't execute command due to fault condition (determine fault(s) by checking FaultFlags)
S2218_ERR_REOPEN            = 22      # Attempted to open smad when it's already open (i.e., smad.state != SS_POWER_OFF).
S2218_ERR_CREATE_RESOURCE   = 23      # Failed to create mutex/semaphore.
S2218_ERR_RESOURCE_CLOSED   = 24      # Resource is closed or unavailable.
S2218_ERR_DEVICE_CLOSED     = 25      # Device is closed.
S2218_ERR_DEVICE_OPEN       = 26      # Device is already open.
S2218_ERR_WAS_SHUTDOWN      = 27      # USB connection was shutdown.
S2218_ERR_READ              = 28      # USB read error.
S2218_ERR_WRITE             = 29      # USB write error.
S2218_ERR_OPENHIDLIB        = 30      # Failed to open HIDLIB.
S2218_ERR_DEVICE_NOT_FOUND  = 31      # Specified device is not detected.
S2218_ERR_COMMAND_MISMATCH  = 32      # Response didn't match last issued command.
S2218_ERR_API_OPEN          = 33      # API is already open.
S2218_ERR_API_CLOSED        = 34      # API is closed.
S2218_ERR_LOCK_TIMEOUT      = 35      # Failed to get resource lock
S2218_ERR_SIG_TIMEOUT       = 36      # Timed out waiting for message from device.
S2218_ERR_NOT_ALLOWED       = 37      # Prohibited by current settings.
S2218_ERR_DUPLICATE_ADDRESS = 38      # Two modules have the same address.
S2218_ERR_CANT_OPEN_DEVICE  = 39      # Failed to open device.
S2218_ERR_MALLOC            = 40      # Memory allocation failed.
S2218_ERR_SYSTEM            = 41      # Unknown system error.
                
# SENSOR_TYPE
STR_SENSOR_TYPE = ["Disabled",
                     "5V",
                     "500mV",
                     "100mV",
                     "400 ohms",
                     "4 Kohms",
                     "600 Kohms",
                     "B thermocouple",
                     "C thermocouple",
                     "E thermocouple",
                     "J thermocouple",
                     "K thermocouple",
                     "N thermocouple",
                     "T thermocouple",
                     "S thermocouple",
                     "R thermocouple",
                     "10K thermistor",
                     "RTD Pt 385 400C",
                     "Pt 385 800C",
                     "Pt 392 400C",
                     "Pt 392 800C",
                     "Ni 200",
                     "Ni 1000",
                     "Cu 10"]
                     
SDC_DISABLED                = 0
SDC_VOLTS_5V                = 1    #  5 V range
SDC_VOLTS_500MV             = 2    #  500 mV range
SDC_VOLTS_100MV             = 3    #  100 mV range
SDC_OHMS_400                = 4    #  400 ohms
SDC_OHMS_4K                 = 5    #  4 Kohms
SDC_OHMS_600K               = 6    #  600 Kohms
SDC_TEMP_THERMOCOUPLE_B     = 7    #  B thermocouple
SDC_TEMP_THERMOCOUPLE_C     = 8    #  C
SDC_TEMP_THERMOCOUPLE_E     = 9    #  E
SDC_TEMP_THERMOCOUPLE_J     = 10   #  J
SDC_TEMP_THERMOCOUPLE_K     = 11   #  K
SDC_TEMP_THERMOCOUPLE_N     = 12   #  N
SDC_TEMP_THERMOCOUPLE_T     = 13   #  T
SDC_TEMP_THERMOCOUPLE_S     = 14   #  S
SDC_TEMP_THERMOCOUPLE_R     = 15   #  R
SDC_TEMP_THERMISTOR         = 16   #  10K thermistor
SDC_TEMP_RTD_PT100_385_400  = 17   #  RTD Pt 385 400C
SDC_TEMP_RTD_PT100_385_800  = 18   #  Pt 385 800C
SDC_TEMP_RTD_PT100_392_400  = 19   #  Pt 392 400C
SDC_TEMP_RTD_PT100_392_800  = 20   #  Pt 392 800C
SDC_TEMP_RTD_NI200          = 21   #  Ni 200
SDC_TEMP_RTD_NI1000         = 22   #  Ni 1000
SDC_TEMP_RTD_CU10           = 23   #  Cu 10
S2218_NUM_SENSOR_TYPES      = 24
NUM_GPIOS = 8

#powergood valid power state
#from class status_report_struct
POWERGOOD_N12V = (1 << 0)
POWERGOOD_3R3V = (1 << 1)
POWERGOOD_5V = (1 << 2)
POWERGOOD_P15V = (1 << 3)

s2218mid = cdll.LoadLibrary("./libs2218.so")
CALLBACKTYPE = CFUNCTYPE(c_int, c_void_p, c_int, c_int, c_int)

def OpenApi():
        devflags = c_int(0)
        s2218mid.s2218_OpenApi(byref(devflags))
        return devflags
def CloseApi():
        s2218mid.s2218_CloseApi()
        return 0

def OpenDevice(devaddr):
        online = c_int(0)
        rc = s2218mid.s2218_OpenDevice(devaddr, online)
        return rc, online

def CloseDevice(devaddr):
        rc = s2218mid.s2218_CloseDevice(devaddr)
        return rc

def ShutdownDevice(devaddr):
        rc = s2218mid.s2218_ShutdownDevice(devaddr)
        return rc

class devVersionStruct(Structure):
        _fields_ = (("major", c_ubyte),
                    ("minor", c_ubyte),
                    ("build", c_short),
                    ("smad_fw", c_short),
                    ("pwb_rev", c_ubyte))
        def __repr__(self):
                return f'versionStruct(major={self.dev_major},minor={self.dev_minor},build={self.dev_build},smad_fw={self.smad_fw},pwb_rev={self.pwb_rev})'

class versionStruct(Structure):
        _fields_ = (("dev_major", c_char),
                    ("dev_minor", c_char),
                    ("dev_build", c_short))
        def __repr__(self):
                return f'versionStruct(major={self.dev_major},minor={self.dev_minor},build={self.dev_build})'

        
def GetApiVersion():
        dprint("api version")
        ver = versionStruct(0,0,0)
        s2218mid.s2218_GetApiVersion(byref(ver))
        print(ver)

def GetDevVersions(devaddr):
        dprint("dev version")
        ver = devVersionStruct(0,0,0,0,0)
        rc = s2218mid.s2218_GetDeviceVersions(devaddr, byref(ver))
        return rc, ver

def ErrString(errcode):
        s2218mid.s2218_ErrString.restype = ctypes.c_char_p
        return s2218mid.s2218_ErrString(errcode)

def SetSensorType(devaddr, chan, sdc):
        rc = s2218mid.s2218_SetSensorType(devaddr, chan, sdc)
        return rc

def SetOpenSensorValues(devaddr, fail_high_flags):
        rc = s2218mid.s2218_SetOpenSensorValues(devaddr, fail_high_flags)
        return rc

class analog_config_struct_old(Structure):
        _fields_ = (("t0", c_char),
                    ("t1", c_char),
                    ("t2", c_char),
                    ("t3", c_char),
                    ("t4", c_char),
                    ("t5", c_char),
                    ("t6", c_char),
                    ("t7", c_char),
                    ("smad_state", c_char),
                    ("tc_cond_enables", c_char),
                    ("open_flags", c_char))

class analog_config_struct(Structure):
        _fields_ = (("sensor_type", c_int*8),
                    ("smad_state", c_ubyte),
                    ("tc_cond_enables", c_ubyte),
                    ("open_flags", c_ubyte),
                    ("padding", c_ubyte))
        
#        def __repr__(self):
##               return f'analog_config_struct(type={self.type},smad_state={self.#smad_state},tc_cond_enables={self.tc_cond_enables},open_flags={self.open_flags})'

class analog_config_python:
        def __init__(self, s0, s1,s2,s3,s4,s5,s6,s7,
                     smad_state, tc_cond_enables, open_flags):
                self.sensor_type = [s0, s1, s2, s3, s4, s5,
                                    s6, s7]
                self.smad_state = smad_state
                self.tc_cond_enables = tc_cond_enables
                self.open_flags = open_flags

def GetAnalogConfig(devaddr):
        cfg = analog_config_struct()
        rc = s2218mid.s2218_GetAnalogConfig(devaddr, byref(cfg))
        return rc, cfg

class status_report_struct(Structure):
        _fields_ = (("soft_trig", c_int),
                    ("n_dropped", c_int),
                    ("meter_mv", c_short*4),
                    ("power_enables", c_ubyte),
                    ("power_good", c_ubyte),
                    ("d_power_state", c_ubyte),
                    ("p_power_state", c_ubyte),
                    ("n_power_state", c_ubyte),
                    ("smad_state", c_ubyte),
                    ("com_state", c_ubyte))


def GrabStatus(devaddr, tmax):
        status = status_report_struct()
        rc = s2218mid.s2218_GrabStatus(devaddr, byref(status), tmax)
        return rc, status


def SetPowerEnables(devaddr, flags):
        rc = s2218mid.s2218_SetPowerEnables(devaddr, c_ubyte(flags))
        return rc

def GetPowerEnables(devaddr):
        flags = c_ubyte()
        rc = s2218mid.s2218_GetPowerEnables(devaddr, byref(flags))
        return rc, flags

class s2218_datareport(Structure):
        _fields_ = (("samples", c_double*8),
                    ("timestamp_s", c_uint),
                    ("timestamp_ms", c_uint),
                    ("n_dropped", c_int),
                    ("hi_limit_vios", c_ubyte),
                    ("low_limit_vios", c_ubyte),
                    ("padding", c_int * 32)
                    )
       
def WaitForDataReport(devaddr, report, tmax):
        rc = s2218mid.s2218_WaitForDataReport(devaddr, byref(report), tmax)
        return rc

class trig_config_python:
        def __init__(self, rate, gpio_flags_rising, gpio_flags_falling):
                self.gpio_flags_rising = gpio_flags_rising
                self.gpio_flags_falling = gpio_flags_falling
                self.rate = rate

class s2218_trig_config(Structure):
        _fields_ = (("rate", c_ushort),
                    ("gpio_flags_rising", c_ubyte),
                    ("gpio_flags_falling", c_ubyte))

def SetTrigConfig(devaddr, cfgp):
        cfg = s2218_trig_config()
        cfg.gpio_flags_rising = cfgp.gpio_flags_rising
        cfg.gpio_flags_falling = cfgp.gpio_flags_falling
        cfg.rate = cfgp.rate
        rc = s2218mid.s2218_SetTrigConfig(devaddr, byref(cfg))
        return rc

def GetTrigConfig(devaddr):
        rc = s2218mid.s2218_GetTrigConfig(devaddr, byref(cfg))
        return rc

def SetTrigEnable(devaddr, enable):
        rc = s2218mid.s2218_SetTrigEnable(devaddr, c_ubyte(enable))
        return rc;

def GetTrigEnable(devaddr):
        enable = c_ubyte(0)
        rc = s2218mid.s2218_GetTrigEnable(devaddr, byref(enable))
        return rc, enable




class s2218_gpio_pair(Structure):
        _fields_ = (("rising", c_ubyte),
                    ("falling", c_ubyte))
        
class s2218_gpio_config(Structure):
        _fields_ = (("dbtime", s2218_gpio_pair * 8),
                    ("pin_directions", c_ubyte),
                    ("bias_enables", c_ubyte),
                    ("bias_polarities", c_ubyte),
                    ("spare", c_ubyte))

class gpio_settings(Structure):
        _fields_ = (("config", s2218_gpio_config),
                    ("cap_enables", s2218_gpio_pair),
                    ("output_states", c_ubyte),
                    ("spare", c_ubyte))
        
class s2218_alarm_config(Structure):
        _fields_ = (("lo_limit", c_double),
                   ("hi_limit", c_double),
                   ("flags", c_ubyte))

class s2218_alarm_report(Structure):
        _fields_ = (("timestamp_s", c_uint),
                    ("timestamp_ms", c_uint),
                    ("n_dropped", c_uint),
                    ("flags", c_ushort))

class s2218_gpio_report(Structure):
        _fields_ = (("timestamp_s", c_uint),
                    ("timestamp_ms", c_uint),
                    ("flags", s2218_gpio_pair),
                    ("pinstates", c_ubyte))
        
def SetGpioConfig(devaddr, gpio):
        rc = s2218mid.s2218_SetGpioConfig(devaddr, byref(gpio))
        return rc;

def SetGpioOutputs(devaddr, outputstates):
        rc = s2218mid.s2218_SetGpioOutputs(devaddr, outputstates)
        return rc;

def GetGpioSettings(devaddr, cfg):
        rc = s2218mid.s2218_GetGpioSettings(devaddr, byref(cfg))
        return rc

def GetGpioInputs(devaddr):
        states = c_ubyte(8)
        rc = s2218mid.s2218_GetGpioInputs(devaddr, byref(states))
        return rc,states

def GetAlarmConfig(devaddr, chan):
        cfg = s2218_alarm_config()
        rc = s2218mid.s2218_GetAlarmConfig(devaddr, chan, byref(cfg))
        print(cfg.flags)
        return rc, cfg

def SetLoThreshold(devaddr, chan, limval):
        rc = s2218mid.s2218_SetLoThreshold(devaddr, c_ubyte(chan), c_double(limval))
        return rc

def SetHiThreshold(devaddr, chan, limval):
        rc = s2218mid.s2218_SetHiThreshold(devaddr, c_ubyte(chan), c_double(limval))
        return rc

def SetLoLimitEnable(devaddr, chan, enab):
        rc = s2218mid.s2218_SetLoLimitEnable(devaddr, c_ubyte(chan),
                                             c_short(enab))
        print("lo limit " + str(rc))
        return rc

def SetHiLimitEnable(devaddr, chan, enab):
        rc = s2218mid.s2218_SetHiLimitEnable(devaddr, c_ubyte(chan),
                                             c_short(enab))
        return rc

def SetLoAlarmEnable(devaddr, chan, enab):
        rc = s2218mid.s2218_SetLoAlarmEnable(devaddr, c_ubyte(chan),
                                             c_short(enab))
        return rc

def SetHiAlarmEnable(devaddr, chan, enab):
        rc = s2218mid.s2218_SetHiAlarmEnable(devaddr, c_ubyte(chan),
                                             c_short(enab))
        return rc

# read device temperature degrees Celcius
def ReadDeviceTemp(devaddr):
        degc = c_double()
        rc = s2218mid.s2218_ReadDeviceTemp(devaddr, byref(degc))
        return rc, degc

def WaitForLimitAlarm(devaddr, tmax):
        report = s2218_alarm_report()
        rc = s2218mid.s2218_WaitForLimitAlarm(devaddr, report, tmax)
        return rc, report

def SetGpioCapEnables(devaddr, enables):
        rc = s2218mid.s2218_SetGpioCapEnables(devaddr, enables)
        
def WaitForGpioEvent(devaddr, tmax):
        report = s2218_gpio_report()
        rc = s2218mid.s2218_WaitForGpioEvent(devaddr, report, tmax)
        return rc, report






