///////////////////////////////////////////////////////////////////////////////////////
// File      : s2218_core.c
// Project   : API for Sensoray model 2218
// Function  : Portable API core (operating system/processor independent)
// Author    : Jim Lamberson
// Copyright : (C) 2018 Sensoray
// set max column width 100
///////////////////////////////////////////////////////////////////////////////////////
// To build the API:
//  * Revise the HAL interface (s2218_platform.h) and implementation (s2218_platform.c)
//     as required to be compatible to the target operating system and CPU.
//  * Compile and link this module and the HAL interface to create the API library.
///////////////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <math.h>               // pow()
#include <pthread.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include "s2218_platform.h"     // System-dependent HAL
#include "api2218.h"            // Command application/middleware defines
#include "mid2218.h"            // Common middleware/firmware defines
#include "s2218_core.h"
#include "s2218_version.h"       // ********** UPDATE EVERY RELEASE ***********
// Constants ============================
// Configuration constants ----------
// Times in milliseconds
#define FALSE 0
#define TRUE  1
#define MAX_RESPONSE_MSEC           500     // Maximum time (ms) to wait for command response.
#define MAX_LOCK_WAIT               1000    // Maximum time (ms) to wait for a resource lock.
#define MAX_STATUS_WAIT             500     // Maximum time (ms) to wait for a soft status report.
// Queue capacities in items
#define STREAMQ_CAPACITY            20      // Data reports.
#define ALARMQ_CAPACITY             20      // Alarm reports.
#define GPIOEVENTQ_CAPACITY         100     // GPIO event reports.
#define NOTIFYQ_CAPACITY            16      // Status reports.
// Manifest constants ---------------
// USB IDs
#define VENDOR_ID                   0x1943  // USB vendor ID (Sensoray)
#define DEVICE_ID                   0x2218  // USB device ID (Model number)
// Incoming usb message types
#define MSGTYPE_REPLY               0       // Command reply
#define MSGTYPE_NOTIFY              1       // Event notification
// Types ============================

typedef struct SENSOR {           // Channel configuration info:
	SENSOR_TYPE     TypeCode; //   API sensor type code.
	double          Scalar;   //   Raw data scalar (EngrUnits = raw * scalar).
} SENSOR;

typedef struct RESOURCE {         // Resource control:
	pthread_mutex_t lock;     // Resource lock (mutex).
	pthread_cond_t cond;      // Data pending signal (condition variable)
	int init;
	int ShutdownReq; // Boolean: Shutdown requested.
} RESOURCE;

typedef struct QUEUE {     // Generic queue:
	RESOURCE Resource; // Resource control.
	int MemberSize; // Size of queue item in bytes.
	void *Ringbuf; // Queue base pointer.
	int Capacity; // Item capacity of queue.
	int WriteIndex; // Queue NextWrite index.
	int ReadIndex; // Queue NextRead index.
	int Level; // Number of items in queue.
	u32 nDropped; // Dropped items count.
} QUEUE;

// Command processor:
typedef struct CMDPROC {
	RESOURCE   Resource; //   Resource control.
	RXPKT rxpkt; //Response buffer. This holds the entire response message (header+payload).
} CMDPROC;

// 2218 DEVICE ------------
typedef struct DEVICE {
        //Pointer to HID device control struct.
	//The device is officially open if this is not NULL.
	HID_DEVICE      *HidDev;
	int             devaddr;        // Logical device number.
	//S2218_HANDLE    hRxThread;    // Response/notification receiver thread handle.
	pthread_t       rx_thread;
	int init;
	int quit;
	SENSOR          Sensor[S2218_NUM_CHANS];        // Channel configuration info.
	QUEUE           *StreamFifo;                    // Data report FIFO.
	QUEUE           *GpioEventQ;                    // GPIO captures events FIFO.
	QUEUE           *AlarmQ;                        // Sensor limit alarm FIFO.
	QUEUE           *AsyncNotifyQ;                  // Soft-triggered status notification FIFO.
	QUEUE           *SoftNotifyQ;                   // Asnyc status notification FIFO.
	CMDPROC         Cmdproc;                        // Command processor.
	int             ShutdownSig;                    // Shutdown request.
} DEVICE;

typedef struct QITEM_GPIOEVENT {                    // GPIO edge detection event queue item:
	S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
	S2218_GPIO_PAIR     Flags;          //   EdgeDetected flags
	// (MSB = rising edges; LSB = falling edges).
	u8                  PinStates;                  //   Pin states.
} QITEM_GPIOEVENT;

typedef struct QITEM_ALARMEVENT {                   // Limit alarm event queue item:
	S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
	u16                 Flags;            //   Alarm flags (MSB = high alarms; LSB = low alarms).
} QITEM_ALARMEVENT;

typedef struct QITEM_DATAEVENT {                    // Sensor sample event queue item:
	double              Samples[S2218_NUM_CHANS];   //   Sensor data.
	S2218_TIMESTAMP     Timestamp;                  //   Data timestamp.
	u8                  HiLimitVios;                //   Limit violations.
	u8                  LoLimitVios;
} QITEM_DATAEVENT;


// Forward declarations ===============

static void *rx_thread(void *arg);
static void get_absolute_ts_msec(struct timespec *ts, int msec);

// Storage ============================

static int      ApiOpen = FALSE;
static DEVICE   DeviceInfo[S2218_MAX_DEVICES] = {0,};


/////////////////////////////////////////////////////////////////////////////////////////
// Sensor scalar lookup table; maps API logical sensor type code to API scalar.
// This table is used to convert a channel's raw sensor data to engineering units
// as follows: rawdata * sensor_attr[logical_sdc] = engr_units.

#define SCALAR_TODEV_LG     5000
#define SCALAR_TODEV_HG     50000
#define SCALAR_TODEV_RL     50
#define SCALAR_TODEV_RM     8
#define SCALAR_TODEV_RH     (1.0 / SCALAR_TOAPI_RH)
#define SCALAR_TOAPI_LG     (1.0 / SCALAR_TODEV_LG)
#define SCALAR_TOAPI_HG     (1.0 / SCALAR_TODEV_HG)
#define SCALAR_TOAPI_RL     (1.0 / SCALAR_TODEV_RL)
#define SCALAR_TOAPI_RM     (1.0 / SCALAR_TODEV_RM)
#define SCALAR_TOAPI_RH     31

typedef struct SENSOR_ATTR {
	int     tcCondEnab;// Boolean: enable thermocouple conditioner when sensor type is declared.
	double  scalar;    // Raw sensor scalar.
} SENSOR_ATTR;


const SENSOR_ATTR sensor_attr[] = {
	{0, 0.0     },          //   0: Disabled channel
	// Volts
	{0, SCALAR_TOAPI_LG },  //   1: 5 V range
	{0, SCALAR_TOAPI_HG },  //   2: 500 mV range
	{0, 0.000005},          //   3: 100 mV range
	// Ohms
	{0, SCALAR_TOAPI_RL },  //   4: 400 ohms
	{0, SCALAR_TOAPI_RM },  //   5: 4 Kohms
	{0, SCALAR_TOAPI_RH },  //   6: 600 Kohms
	// Degrees C
	{1, 0.1     },          //   7: B thermocouple
	{1, 0.1     },          //   8: C
	{1, 0.1     },          //   9: E
	{1, 0.1     },          //  10: J
	{1, 0.1     },          //  11: K
	{1, 0.1     },          //  12: N
	{1, 0.1     },          //  13: T
	{1, 0.1     },          //  14: S
	{1, 0.1     },          //  15: R
	{0, 0.01    },          //  16: 10K thermistor
	{0, 0.0125  },          //  17: RTD Pt 385 400C
	{0, 0.05    },          //  18: Pt 385 800C
	{0, 0.0125  },          //  19: Pt 392 400C
	{0, 0.05    },          //  20: Pt 392 800C
	{0, 0.05    },          //  21: Ni 200
	{0, 0.05    },          //  22: Ni 2000
	{0, 0.1     }           //  23: Cu 10
};

// Endian conversion ============================

///////////////////////////////////////////////////////////////////////////////////////////
// Fetch nbytes from received device message as unsigned (device is big endian).

#define B0      (((u8 *)buf)[0])
#define B1      (((u8 *)buf)[1])
#define B2      (((u8 *)buf)[2])
#define B3      (((u8 *)buf)[3])

static u16 DevBufToU16(void *buf)  // note: buf is treated as *u8
{
	return ( ((u16)B0) << 8 ) | B1;
}

static u32 DevBufToU32(void *buf)  // note: buf is treated as *u8
{
	return ((((( ((u32)B0) << 8 ) | B1 ) << 8 ) | B2 ) << 8) | B3;
}

static void U32ToDevBuf(u32 n, u8 *buf)
{
	*buf++ = (u8)(n >> 24);
	*buf++ = (u8)(n >> 16);
	*buf++ = (u8)(n >> 8);
	*buf++ = (u8)(n);
}

// Host to device
#define LSB(N)                  ((u8)((N) & 0xFF))     // Extract bytes from host u16
#define MSB(N)                  ((u8)((N) >> 8))

//////////////////////////////////////////////////////////////////////////
//////////////////  RESOURCE MANAGEMENT  /////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Lock/unlock a resource.

static int Lock(RESOURCE *r)
{
	int rc;
	//struct timespec ts;

	//get_absolute_ts_msec(&ts, MAX_LOCK_WAIT);

	//rc = pthread_mutex_timedlock(&r->lock, &ts);
	rc = pthread_mutex_lock(&r->lock); //lint !e454
	return rc;
	// If timed out, return more specific rc
	//return (rc == ETIMEDOUT) ? S2218_ERR_LOCK_TIMEOUT : rc;
}

static void Unlock(RESOURCE *r)
{
	pthread_mutex_unlock(&r->lock); //lint !e455
	return;
}

/////////////////////////////////////////////////
// Send/receive a resource signal.
static int Signal(RESOURCE *r)
{
	int rc;
	//  if (r->type == 0)
	rc = pthread_cond_broadcast(&r->cond);
	//  else
	//rc = sem_post(&r->sem);
	return rc;
}

static void get_absolute_ts_msec(struct timespec *ts, int msec)
{
	clock_gettime(CLOCK_REALTIME, ts);
	if (msec >= 1000) {
		ts->tv_sec += msec / 1000; // 1 second
		ts->tv_nsec += (msec % 1000) * 1000 * 1000;
	} else {
		ts->tv_nsec += msec * 1000 * 1000;
	}
	while (ts->tv_nsec > 1000000000) {
		ts->tv_sec += 1;
		ts->tv_nsec -= 1000000000;
	}
}

#if 0
static int WaitSignal(RESOURCE *r, int msec)
{
	int rc;
	struct timespec ts;

	pthread_mutex_lock(&r->lock);
	get_absolute_ts_msec(&ts, msec);
	rc = pthread_cond_timedwait(&r->cond, &r->lock, &ts);
	pthread_mutex_unlock(&r->lock);
	if (r->ShutdownReq)
		return S2218_ERR_WAS_SHUTDOWN;
        // If timed out, return more specific rc
	return (rc == ETIMEDOUT) ? S2218_ERR_SIG_TIMEOUT : rc;
}
#endif
// wait signal, resource already locked
static int WaitSignalLocked(RESOURCE *r, int msec)
{
	int rc;
	struct timespec ts;
	get_absolute_ts_msec(&ts, msec);
	rc = pthread_cond_timedwait(&r->cond, &r->lock, &ts);
	if (r->ShutdownReq) {
		return S2218_ERR_WAS_SHUTDOWN;
	}
	// If timed out, return more specific rc
	return (rc == ETIMEDOUT) ? S2218_ERR_SIG_TIMEOUT : rc;
}

/////////////////////////////////////////////////////////////
// Open a resource management struct.
static int OpenResource(RESOURCE *r, int maxcount)
{
	r->ShutdownReq = 0;
	pthread_mutex_init(&r->lock, NULL);
	pthread_cond_init(&r->cond, NULL);
	return S2218_ERR_OK;
}

///////////////////////////////////////////////////////////////
// Cancel a resource's blocking wait (if any).

static void CancelResourceWait(RESOURCE *r)
{
	r->ShutdownReq = 1; // Signal resource shutdown request.
	if (!r->init)
		return;
	pthread_cond_broadcast(&r->cond);
}

///////////////////////////////////////////////////////////////
// Close resource manager objects and free their handles.
static void DestroyResource(RESOURCE *r)  // destroy resource handles
{
	/*
	  if (r->hPending != S2218_INVALID_HANDLE) {   // Resource signal.
	  sysDestroySemaphore(r->hPending);
	  r->hPending = S2218_INVALID_HANDLE;
	  }
	  if (r->hLock != S2218_INVALID_HANDLE) {   // Resource lock.
	  sysDestroyMutex(r->hLock);
	  r->hLock = S2218_INVALID_HANDLE;
	  }
	*/
	r->init = 0;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////  QUEUE MANAGEMENT  ////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////



// Cancel a queue's blocking wait (if any).
static void QueueCancelWait(QUEUE **pq)
{
	QUEUE *q = *pq;                         // Get queue pointer.
	if (q != NULL)                          // If the queue exists
		CancelResourceWait(&q->Resource);   //   cancel blocking wait on the queue (if any).
}

static void QueueDestroy(QUEUE **pq)
{
	QUEUE *q = *pq;                     // Get queue pointer.
        // Free malloc'd memory and close all handles.
	if (q != NULL)
	{
		DestroyResource(&q->Resource);  // Close resource manager and free its handles.
		if (q->Ringbuf != NULL)         // Free queue's ring buffer.
			free(q->Ringbuf);
		free(q);                        // Free the queue struct.
		*pq = NULL;                     // Mark queue closed.
	}
}

// Create and open a queue.
static int QueueOpen(int itemsize, int capacity, QUEUE **pq)
{
	QUEUE *q;

	*pq = NULL; // indicate queue not yet created

	q = malloc(sizeof(QUEUE));                  // Allocate QUEUE.
	if (q == NULL)
		return S2218_ERR_MALLOC;

	q->Ringbuf = malloc(itemsize * capacity);   // Allocate ring buffer.
	if (q->Ringbuf == NULL) {
		free(q);
		return S2218_ERR_MALLOC;
	}
	if (OpenResource(&q->Resource, capacity) != S2218_ERR_OK) { // Create queue lock and signal.
		free(q->Ringbuf);
		free(q);
		return S2218_ERR_CREATE_RESOURCE;
	}
	// Initialize queue:
	q->ReadIndex    = 0;                    // Init read/write indices.
	q->WriteIndex   = 0;
	q->Level        = 0;                    // Queue is empty.
	q->nDropped     = 0;                    // No dropped items yet.
	q->MemberSize   = itemsize;             // Item size in bytes.
	q->Capacity     = capacity;             // Queue capacity in items.
	*pq = q;
	return S2218_ERR_OK;
}

//////////////////////////////////////////////////////////////////////
// Add item to queue.
// Fails silently if queue doesn't exist or can't be locked.

// helper: add item at index, then advance index (index += 1, rollover-corrected)
static void QueueInsertItem(QUEUE *q, void *item)
{
	memcpy((u8 *)(q->Ringbuf) + q->WriteIndex * q->MemberSize, item, q->MemberSize);
	if (++(q->WriteIndex) == q->Capacity)   // Bump NextWrite index and wrap if end of ring buffer.
		q->WriteIndex = 0;
}

// Add item to queue.
static void QueueWrite(QUEUE *q, void *item)
{
	// abort if queue doesn't exist
	if (q != NULL) {
		if (Lock(&q->Resource) == S2218_ERR_OK) {
			int full = (q->Level == q->Capacity); // Determine whether queue is full.
			QueueInsertItem(q, item); // Add item to queue;
			//this overwrites (i.e., drops) oldest item if queue is full.
			if (q->Level < q->Capacity) // If queue not full
				q->Level++; //   Bump queue level.
			else { // Else drop oldest item:
				q->ReadIndex = q->WriteIndex; // Set NextRead to next oldest item.
				q->nDropped++; // Bump dropped count.
			}
			if (!full) // If item was added to queue
				Signal(&q->Resource); //   increment queue level.
			Unlock(&q->Resource); // Leave critical section -----
		}
	}
}

///////////////////////////////////////////////////////////////////////////
// Extract oldest item from queue.
// Imports:
//  q           - pointer to queue.
//  item        - buffer that will receive queue item. Set to NULL if not needed.
//  nDropped    - buffer that will receive dropped item count. Set to NULL if not needed.
//  tmax        - maximum milliseconds to wait for item.
static int QueueRead(QUEUE *q, void *item, int *nDropped, int tmax)
{
	int rc;// = WaitSignal(&q->Resource, tmax);   // Wait for queue item available.
	if (q->Resource.ShutdownReq)                    // Abort if shutdown requested.
		return S2218_ERR_WAS_SHUTDOWN;
	//if (rc != S2218_ERR_OK)                    // Abort if timeout or error.
	//  return rc;
	rc = Lock(&q->Resource);                   // Enter critical section ----------------------
	if (rc == S2218_ERR_OK)
	{
		if (q->Level <= 0) {
			// Wait for queue item available.
			rc = WaitSignalLocked(&q->Resource, tmax);
			//printf("wait signal locked %d, queue level %d\n", rc, q->Level);
			if (rc != S2218_ERR_OK) {
				// Exit critical section --------------------------
				Unlock(&q->Resource);
				return rc;
			}
		}
		// If caller is interested in item content
		//   copy oldest queue item to user buffer.
		if (item != NULL)
			memcpy(item, (u8 *)q->Ringbuf + q->ReadIndex * q->MemberSize,
				q->MemberSize);
		// If caller is interested in dropped item count
                //   send the count.
		if (nDropped != NULL)
			*nDropped = q->nDropped;
		// Bump queue's item count.
		q->Level--;
		// Bump NextRead index and wrap if at end of buffer.
		if (++(q->ReadIndex) == q->Capacity)
			q->ReadIndex = 0;
                // Exit critical section --------------------------
		Unlock(&q->Resource);
	}
	return rc;
}

///////////////////////////////////////////////////////////////////////////
// Discard all items in queue.
#if 0
static void QueueFlush(QUEUE *q)
{
	int rc;

	do {
		rc = QueueRead(q, NULL, NULL, 0);
		if (rc == S2218_ERR_SIG_TIMEOUT)
			return;
	}
	while (rc == S2218_ERR_OK);
}
#endif
///////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////  QUEUE TYPE ENFORCEMENT  ///////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

// ALARM FIFO ---------------------------------------------------------------------
static int CreateAlarmFifo(DEVICE *dev)
{
	return QueueOpen(sizeof(QITEM_ALARMEVENT), ALARMQ_CAPACITY, &dev->AlarmQ);
}

static void WriteAlarmFifo(DEVICE *dev, QITEM_ALARMEVENT *ev)
{
	QueueWrite(dev->AlarmQ, ev);
}

static int ReadAlarmFifo(DEVICE *dev, S2218_ALARMREPORT *report, int tmax)
{
	QITEM_ALARMEVENT ev;
	int nDropped;
	int rc = QueueRead(dev->AlarmQ, &ev, &nDropped, tmax);
	report->Flags       = ev.Flags;
	report->Timestamp   = ev.Timestamp;
	report->nDropped    = nDropped;
	return rc;
}

// GPIO FIFO ---------------------------------------------------------------------
static int CreateGpioFifo(DEVICE *dev)
{
	return QueueOpen(sizeof(QITEM_GPIOEVENT), GPIOEVENTQ_CAPACITY, &dev->GpioEventQ);
}

static void WriteGpioFifo(DEVICE *dev, QITEM_GPIOEVENT *ev)
{
	QueueWrite(dev->GpioEventQ, ev);
}

static int ReadGpioFifo(DEVICE *dev, S2218_GPIOREPORT *report, int tmax)
{
	QITEM_GPIOEVENT ev;
	int nDropped;
	int rc = QueueRead(dev->GpioEventQ, &ev, &nDropped, tmax);
	report->Flags       = ev.Flags;
	report->PinStates   = ev.PinStates;
	report->Timestamp   = ev.Timestamp;
	report->nDropped    = nDropped;
	return rc;
}

// DATA FIFO ---------------------------------------------------------------------
static int CreateDataFifo(DEVICE *dev)
{
	return QueueOpen(sizeof(QITEM_DATAEVENT), STREAMQ_CAPACITY, &dev->StreamFifo);
}

static void WriteDataFifo(DEVICE *dev, QITEM_DATAEVENT *ev)
{
	QueueWrite(dev->StreamFifo, ev);
}

static int ReadDataFifo(DEVICE *dev, S2218_DATAREPORT *report, int tmax)
{
	QITEM_DATAEVENT ev;
	int nDropped;
	int rc = QueueRead(dev->StreamFifo, &ev, &nDropped, tmax);
	memcpy(report->Samples, ev.Samples, sizeof(ev.Samples));
	report->HiLimitVios = ev.HiLimitVios;
	report->LoLimitVios = ev.LoLimitVios;
	report->Timestamp   = ev.Timestamp;
	report->nDropped    = nDropped;
	return rc;
}

// STATUS FIFOS ---------------------------------------------------------------------
// Async fifo
static int CreateAsyncStatusFifo(DEVICE *dev)
{
	return QueueOpen(sizeof(S2218_STATUSREPORT), NOTIFYQ_CAPACITY, &dev->AsyncNotifyQ);
}

static void WriteAsyncStatusFifo(DEVICE *dev, S2218_STATUSREPORT *report)
{
	QueueWrite(dev->AsyncNotifyQ, report);
}

static int ReadAsyncStatusFifo(DEVICE *dev, S2218_STATUSREPORT *report, int tmax)
{
	return QueueRead(dev->AsyncNotifyQ, report, &report->nDropped, tmax);
}

// Soft fifo
static int CreateSoftStatusFifo(DEVICE *dev)
{
	return QueueOpen(sizeof(S2218_STATUSREPORT), NOTIFYQ_CAPACITY, &dev->SoftNotifyQ);
}

static void WriteSoftStatusFifo(DEVICE *dev, S2218_STATUSREPORT *report)
{
	QueueWrite(dev->SoftNotifyQ, report);
}

static int ReadSoftStatusFifo(DEVICE *dev, S2218_STATUSREPORT *report, int tmax)
{
	return QueueRead(dev->SoftNotifyQ, report, &report->nDropped, tmax);
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Read/write a device's response queue (Cmdproc.Rsp), which is effectively a one-deep fifo.
// USB receiver thread calls this to write command reply to response queue
static void WriteCmdRspQueue(DEVICE *dev, u8 *data, int len)
{
	if (len > (int) sizeof(RXPKT))  // avoid buf overflow
		len = sizeof(RXPKT);

	// Copy command response header and payload to module's dedicated response buffer
	memcpy(&dev->Cmdproc.rxpkt, data, len);
	// Signal app that command response is available
	Signal(&dev->Cmdproc.Resource);
}

// Client thread calls this blocking function to receive command reply from response queue
static int WaitCmdRspQueue(DEVICE *dev, XACT *x, int msec)
{
	// Wait for device response; if timeout or error
	int rc = WaitSignalLocked(&dev->Cmdproc.Resource, msec);
	if (rc != S2218_ERR_OK)
		return rc;
	// Else if device responded with error message
	if (dev->Cmdproc.rxpkt.rsp.rc != S2218_ERR_OK)
		return dev->Cmdproc.rxpkt.rsp.rc;
	// Else if response is for an obviously unrelated command
	if (x->cmd.opcode != dev->Cmdproc.rxpkt.rsp.opcode)
		return S2218_ERR_COMMAND_MISMATCH;
	// Else if command response contains a payload and caller is interested in the payload
	//   copy payload to app buffer
	if (x->rsp.datalen > 0)
		memcpy(x->rsp.data, dev->Cmdproc.rxpkt.rsp.data, x->rsp.datalen);
	return S2218_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////  DEVICE OPEN/CLOSE UTILITIES  /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Initialize device control structure to startup condition.

static void InitDevStruct(DEVICE *dev)
{
	RESOURCE *r = &dev->Cmdproc.Resource; // Cache pointer to CmdProc's resource struct.
	// Mark HID device closed, queues uncreated, sensor chans disabled
	// (doesn't actually disable them; just forces sensor data to zero).
	memset(dev, 0, sizeof(DEVICE));
	r->init = 0;
	dev->init = 0;
}

/////////////////////////////////////////////
// Free all resources used by device.
static void ShutdownDevResources(DEVICE *dev)
{
	CancelResourceWait(&dev->Cmdproc.Resource);
	QueueCancelWait(&dev->StreamFifo);
	QueueCancelWait(&dev->AlarmQ);
	QueueCancelWait(&dev->GpioEventQ);
	QueueCancelWait(&dev->AsyncNotifyQ);
	QueueCancelWait(&dev->SoftNotifyQ);
}

static void DestroyDevResources(DEVICE *dev)
{
	DestroyResource(&dev->Cmdproc.Resource);
	QueueDestroy(&dev->StreamFifo);
	QueueDestroy(&dev->AlarmQ);
	QueueDestroy(&dev->GpioEventQ);
	QueueDestroy(&dev->AsyncNotifyQ);
	QueueDestroy(&dev->SoftNotifyQ);
}

// Start/stop USB receiver thread.
static int StartUsbReceiver(DEVICE *dev)
{
	int rc;
	rc = pthread_create(&dev->rx_thread, NULL, rx_thread, dev);
	if (rc < 0)
		return S2218_ERR_CREATE_RESOURCE;
	dev->init = 1;
	return S2218_ERR_OK;
}

static void StopUsbReceiver(DEVICE *dev)
{
	// If USB receiver thread was created
	if (dev->init != 0) {
		//   Signal shutdown
		dev->ShutdownSig = 1;
		dev->quit = 1;
		//pthread_cancel(dev->rx_thread);
		hid_device_stop_poll(dev->HidDev);
		// Wait for receiver thread to terminate.
		//pthread_cancel(dev->rx_thread);
		pthread_join(dev->rx_thread, NULL);
		dev->init = 0; // mark thread closed
	}
}

//////////////////////////////////////////////////////////////////////////////
// Terminate USB receiver and force all blocking functions to return error.
static int ShutdownDevice(DEVICE *dev)
{
	// Terminate USB receiver thread if it's running.
	StopUsbReceiver(dev);
	return S2218_ERR_OK;
}

// Close 2218 device.
static int CloseDevice(DEVICE *dev)
{
	int devaddr = dev->devaddr;
	// Abort if device is already closed.
	if (dev->HidDev == NULL)
		return S2218_ERR_DEVICE_CLOSED;
        // Terminate USB receiver thread if it's running.
	StopUsbReceiver(dev);
	DestroyDevResources(dev);
	// Close HID device.
	hid_device_close(dev->HidDev);
	// Reset device control structure to startup condition (marks HID device closed).
	InitDevStruct(dev);
	// start-up condition has correct device address and initdevstruct clears
	// it completely, restore it here.
	dev->devaddr = devaddr;
	return S2218_ERR_OK;
}

///////////////////////////////////////////////////////////
// Open 2218 device.
static int OpenDevice(DEVICE *dev)
{
	// Find and open device.
	int rc = hid_device_open(VENDOR_ID, DEVICE_ID, dev->devaddr, NULL, &dev->HidDev);
	if (rc == S2218_ERR_OK) {
		// Create resource locks/signals and start USB receiver.
                // create alarm queue, cmdproc resources, datareport fife, gpio queue
		// create async (triggered by state change) statusQ
		// create soft-triggered statusQ
		// Start USB receiver thread running.
		if ((rc = CreateAlarmFifo(dev)) == S2218_ERR_OK)
			if ((rc = OpenResource(&dev->Cmdproc.Resource, 1)) == S2218_ERR_OK)
				if ((rc = CreateDataFifo(dev)) == S2218_ERR_OK)
					if ((rc = CreateGpioFifo(dev)) == S2218_ERR_OK)
						if ((rc = CreateAsyncStatusFifo(dev)) == S2218_ERR_OK)
							if ((rc = CreateSoftStatusFifo(dev)) == S2218_ERR_OK)
								rc = StartUsbReceiver(dev);
		// If something went wrong,
		//    close everything that's been opened on the device (ignore return value).
		if (rc != S2218_ERR_OK)
			CloseDevice(dev);
	}
	return rc;
}

/////////////////////////////////////////////////////////////////////////////////////////
///////////////////  USB RECEIVER THREAD  ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
// USB RECEIVER THREAD
// Each open 2218 device is associated with a dedicated instance of this thread.
// This thread receives and demultiplexes packets from the 2218. Each packet is forwarded to an
// appropriate handler which executes in an application context as required by the application.
// Imports:
//   arg - 2218 module's DEVICE to associate with this thresd
static void *rx_thread(void *arg)
{
	int     rc = 0;
	DEVICE  *dev = (DEVICE *)arg;   // device associated with this thread instance
	RXPKT   pkt;                    // packet buffer
	//u16     notificationID;         // device-generated message id (not used in this version)
	int     i;

	while (!dev->quit)    // Repeat until device closed or usb read error:
	{
		// Receive packet from device ---------------------------------------------------
		// Wait for next message from device or error, whichever occurs first.
		rc = hid_device_read(dev->HidDev, pkt.buf, S2218_BUFSIZE);
                // Terminate thread if read error.
		if (rc != S2218_ERR_OK) {
			break;
		}
		// Terminate thread if read was canceled due to shutdown request.
		if (dev->ShutdownSig) {
			rc = S2218_ERR_WAS_SHUTDOWN;
			break;
		}

		// Decode message type and forward message as required ---------------------------
		if (pkt.msgtype == MSGTYPE_REPLY) {
			// COMMAND RESPONSE MESSAGE
			// Send command response message to app.
			WriteCmdRspQueue(dev, pkt.buf, RSPHDR_LEN + pkt.rsp.datalen);

		} else if (pkt.msgtype == MSGTYPE_NOTIFY) {
			// ASYNC NOTIFICATION MESSAGE
			//notificationID = DevBufToU16(&pkt.ev.NotificationID);
			// If packet contains a data report
			if (pkt.ev.EventFlags & NOTIFY_DATA) {
				QITEM_DATAEVENT ev;
				// Construct data event message:
				//   Timestamp.
				//   Limit violations.
				ev.Timestamp.ms = DevBufToU16(&pkt.ev.TimestampMsec);
				ev.Timestamp.s  = DevBufToU32(&pkt.ev.TimestampSecHi);
				ev.HiLimitVios  = pkt.ev.LimitVios[0];
				ev.LoLimitVios  = pkt.ev.LimitVios[1];
				//   Sensor data (convert raw to engineering units)
				for (i = 0; i < S2218_NUM_CHANS; i++)
					ev.Samples[i] = dev->Sensor[i].Scalar * (s16)DevBufToU16(&pkt.ev.SensorData[i]);
				// Send data event message to app.
				WriteDataFifo(dev, &ev);
			}

			// If gpio edge(s) detected
			if (pkt.ev.EventFlags & NOTIFY_EDGE) {
				QITEM_GPIOEVENT ev;
				// Build gpio event message:
				//   Timestamp.
				ev.Timestamp.ms     = DevBufToU16(&pkt.ev.TimestampMsec);
				ev.Timestamp.s      = DevBufToU32(&pkt.ev.TimestampSecHi);
				//   Rising/falling edge event flags.
				ev.Flags.Rising     = pkt.ev.GpioEvents[0];
				ev.Flags.Falling    = pkt.ev.GpioEvents[1];
				ev.PinStates        = pkt.ev.GpioPinStates;                     //   Pin states.
				// Send gpio event message to app.
				WriteGpioFifo(dev, &ev);
			}

			// If sensor limit alarm(s) sounded
			if (pkt.ev.EventFlags & NOTIFY_ALARM) {
				QITEM_ALARMEVENT ev;
				// Build alarm event message:
				ev.Timestamp.ms     = DevBufToU16(&pkt.ev.TimestampMsec);
				ev.Timestamp.s      = DevBufToU32(&pkt.ev.TimestampSecHi);
				ev.Flags            = DevBufToU16(pkt.ev.AlarmFlags);
				// Send alarms to app.
				WriteAlarmFifo(dev, &ev);
			}

			// If message contains a status report
			if (pkt.ev.EventFlags & (NOTIFY_SOFT | NOTIFY_STATE)) {
				s16 rawadc[4];
				s16 raw_mV[4];
				S2218_STATUSREPORT status;
				// Gather status info starting with power supply voltages:
				//   Power supply voltages:
				//   Meter raw data:
				for (i = 0; i < 4; i++) {
					// adc value.
					rawadc[i] = DevBufToU16(&pkt.ev.Meter[i]);
					//     adc millivolts.
					raw_mV[i] = (s16)((rawadc[i] * (u32)3300) / 4095);
				}
				//   Power supply voltages:
				status.Meter_mV[3]  = (s16)(5.7 * raw_mV[3]); // +15V
				status.Meter_mV[2]  = 2 * raw_mV[2]; // +5V
				status.Meter_mV[1]  = 2 * raw_mV[1]; // +3.3V
				//-12.25V below
				status.Meter_mV[0]  = (s16)(5000 - ((raw_mV[0] + 12250) * 49) / 39);
				status.SmadState    = pkt.ev.SmadState; // Smad system state.
				status.ComState     = pkt.ev.ComState; // Smad communication state.
				status.PowerEnables = pkt.ev.PowerEnables; // Power enable flags.
				status.PowerGood    = pkt.ev.PowerGood; // All power supplies good.
				status.DPowerState  = pkt.ev.DPowerState; // Digital power state.
				status.PPowerState  = pkt.ev.PPowerState; // +Analog power state.
				status.NPowerState  = pkt.ev.NPowerState; //  -Analog power state.
				status.SoftTrig     = ((pkt.ev.EventFlags & NOTIFY_SOFT) != 0);
				// Event that triggered the report.
				// Send soft-triggered report to app,
				// if report was requested.
				if (pkt.ev.EventFlags & NOTIFY_SOFT)
					WriteSoftStatusFifo(dev, &status);

				// If report was not requested (e.g., sent due to state change)
				//   send async report to app.
				if (pkt.ev.EventFlags & NOTIFY_STATE)
					WriteAsyncStatusFifo(dev, &status);
			}
		}
	}
	// Shutdown all resources, thereby canceling any associated waits.
	ShutdownDevResources(dev);
	// Exit thread with reason for thread termination.
	return (void *) (unsigned long) rc;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////  COMMAND FUNCTIONS  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
// Send command to 2218 and wait for reply. All commands generate a response
// from the module consisting of header and optional payload.
// Command structure: (opcode),(payload_size)[,(payload ...)]
// Imports:
//  dev     - device control struct.
//  rspdata - buffer that receives optional response payload. This may be set to NULL if rsplen==0.
// rsplen - number of payload bytes to receive; set to 0 if payload not needed or expected.
//  cmd - command to send to 2218.

static int Transact(DEVICE *dev, XACT *x)
{
	RESOURCE *r = &dev->Cmdproc.Resource;     // cmdproc control struct
	int rc;

	rc = Lock(r);
	// Lock device's cmdproc to ensure thread safety.
	// if no timeout or errors
	if (rc == S2218_ERR_OK) {
		if (dev->ShutdownSig)
			rc = S2218_ERR_WAS_SHUTDOWN; // Abort if device closed or closing.
		else {
			// Send command to device (consisting of header + data).
			rc = hid_device_write(dev->HidDev, x->cmd.buf, CMDHDR_LEN + x->cmd.datalen);
			//printf("hid_device_write buf[0,1]: %x, %x rc: %d datalen: %d\n",
			//		x->cmd.buf[0], x->cmd.buf[1], rc, x->cmd.datalen);
			if (dev->ShutdownSig)
				rc = S2218_ERR_WAS_SHUTDOWN;// Abort if device_close requested while sending command.
		}
		if (rc == S2218_ERR_OK) {
			// Receive reply from device.
			rc = WaitCmdRspQueue(dev, x, MAX_RESPONSE_MSEC);
		}
	}
	Unlock(r); // Release lock.
	x->rsp.rc = rc; // Cache and return rc
	return rc;
}

// Execute command which has no response payload
static int TransactEmpty(DEVICE *dev, XACT *x)
{
	x->rsp.datalen = 0;
	return Transact(dev, x);
}

// Execute command which returns verbatim payload to app
static int TransactDirect(DEVICE *dev, XACT *x, u8 rsplen, void *rspbuf)
{
	x->rsp.datalen = rsplen;
	Transact(dev, x);
	memcpy(rspbuf, x->rsp.data, rsplen);
	return x->rsp.rc;
}

// Channel configuration --------------------------------------------------------
// Sync local image to device config
static void SyncChanImage(DEVICE *dev, int chan, SENSOR_TYPE sdc)
{
	dev->Sensor[chan].TypeCode = sdc;
	dev->Sensor[chan].Scalar   = sensor_attr[sdc].scalar;
}

// Declare sensor type
static int SetSensorType(DEVICE *dev, int chan, SENSOR_TYPE sdc)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_SENSOR_TYPE;
	x.cmd.datalen = 2;
	x.cmd.data[0] = (u8)chan;
	x.cmd.data[1] = (u8)sdc;
	return TransactEmpty(dev, &x);
}

// Enable/disable thermocouple conditioners
static int SetTcCondEnables(DEVICE *dev, int mask, int enables)
{
	XACT x;
	x.cmd.opcode    = CMD2218_SET_TCCOND_ENABLES;
	x.cmd.datalen   = 2;
	x.cmd.data[0]   = (u8)mask;
	x.cmd.data[1]   = (u8)enables;
	return TransactEmpty(dev, &x);
}

// Enable/disable thermocouple conditioner on one channel
static int SetChanTcCondEnable(DEVICE *dev, int chan, int enable)
{
	return SetTcCondEnables(dev, 1 << chan, (enable != 0) << chan);
}

/////////////////////////////////////////////////////////////////////
// Get analog configuration.

static int GetAnalogConfig(DEVICE *dev, S2218_ANALOG_CONFIG *cfg)
{
	XACT x;
	int i;
	x.cmd.opcode    = CMD2218_GET_ANALOG_CFG;
	x.cmd.datalen   = 0;
	x.rsp.datalen   = sizeof(S2218_ANALOG_CONFIG);
	if (Transact(dev, &x) == S2218_ERR_OK) {
		u8 *pbuf = x.rsp.data;
		// API sensor types for all channels:
		for (i = 0; i < S2218_NUM_CHANS; i++) {
			//   Fetch sensor type from received msg.
			cfg->SensorType[i] = *pbuf++;
			//   Sync local chan config image to device.
			SyncChanImage(dev, i, cfg->SensorType[i]);
		}
		// Smad state (PowerEnabled is indicated by state != SS_POWER_OFF).
		cfg->SmadState      = *pbuf++;
		// Thermocouple conditioner enables.
		cfg->TcCondEnables  = *pbuf++;
		// Open-sensor data values.
		cfg->OpenFlags      = *pbuf++;
	}
	return x.rsp.rc;
}

// DataReport trigger --------------------------------------------------------
// Set/get trigger config
static int SetTrigConfig(DEVICE *dev, S2218_TRIG_CONFIG *cfg)
{
	XACT x;
	x.cmd.opcode    = CMD2218_SET_TRIG_CONFIG;
	x.cmd.datalen   = 4;
	// periodic rate; 0 = disabled
	x.cmd.data[0]   = MSB(cfg->Rate);
	x.cmd.data[1]   = LSB(cfg->Rate);
	// edge selector flags
	x.cmd.data[2]   = cfg->GpioFlags.Rising;
	x.cmd.data[3]   = cfg->GpioFlags.Falling;
	return TransactEmpty(dev, &x);
}

static int GetTrigConfig(DEVICE *dev, S2218_TRIG_CONFIG *cfg)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_TRIG_CONFIG;
	x.cmd.datalen = 0;
	x.rsp.datalen = 4;
	Transact(dev, &x);
	cfg->Rate               = DevBufToU16(&x.rsp.data[0]);
	cfg->GpioFlags.Rising   = x.rsp.data[2];
	cfg->GpioFlags.Falling  = x.rsp.data[3];
	return x.rsp.rc;
}

// Set/get trigger enable
static int SetTrigEnable(DEVICE *dev, u8 enable)
{
	XACT x;
	x.cmd.opcode    = CMD2218_SET_TRIG_ENABLE;
	x.cmd.datalen   = 1;
	x.cmd.data[0]   = enable;
	return TransactEmpty(dev, &x);
}

static int GetTrigEnable(DEVICE *dev, u8 *enable)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_TRIG_ENABLE;
	x.cmd.datalen = 0;
	x.rsp.datalen = 1;
	Transact(dev, &x);
	*enable = x.rsp.data[0];
	return x.rsp.rc;
}

/////////////////////////////////////////////////////////////
// Get hardware/firmware version numbers
static int GetDeviceVersions(DEVICE *dev, S2218_VERSION_INFO *ver)
{
	//                            0123456789ABCDEF01234
	static const char revmap[] = "0ABCDEFGHJKLMNPRTUWXY";
	XACT x;
	u8 pwb;

	x.cmd.opcode  = CMD2218_GET_VERSIONS;
	x.cmd.datalen = 0;
	x.rsp.datalen   = sizeof(S2218_VERSION_INFO);
	// Get version info from device and copy to app buffer:
	Transact(dev, &x);
	// Device firmware version.
	ver->DevFw.Major = x.rsp.data[0];
	ver->DevFw.Minor = x.rsp.data[1];
	ver->DevFw.Build = DevBufToU16(&x.rsp.data[2]);
	// Smad firmware version (x100) MSB
	ver->SmadFw = DevBufToU16(&x.rsp.data[4]);
	pwb = x.rsp.data[6];
        // Board version number: 0=rev0, 1=revA, 2=revB, etc.
	ver->PwbRev = (pwb < 0x15) ? revmap[pwb] : '?';
	return x.rsp.rc;
}

// Data acquisition -----------------------------------------------------------
// Helper: extract V/F snapshot info
static void ExtractVFinfo(u8 *rsp, u8 acc0, S2218_VFINFO *s)
{
	// Number of 16 MHz cycles during V/F integration: concat ACC<16:1> to ACC0
	// (x,x,x,x,x,x,x,ACC<0>) and adj for rollover.
	s->acc = (((((u32)rsp[0] << 8) + rsp[1]) ^ 0xFFFF) << 1) | (acc0 & 1) | 0x40000;
	s->iters    = ((((u32)rsp[2] << 8) + rsp[3]) ^ 0xFFFF) + 1;  // V/F iterations
	s->freq     = (s->acc != 0) ? (16000000.0 * s->iters / s->acc) : 0.0; // V/F frequency.
}

// Get reference standards raw data
static int GetInternalSnapshot(DEVICE *dev, S2218_INT_SNAPSHOT *s)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_INTERNAL_SNAPSHOT;
	x.cmd.datalen = 0;
	x.rsp.datalen = 4 * S2218_NUM_REFS;
	// Fetch 4 bytes per std: acchi, acclo, iterhi, iterlo.
	if (Transact(dev, &x) == S2218_ERR_OK) {
		int i;
		u8 *pbuf = x.rsp.data;
		S2218_VFINFO *pinfo = s->vfinfo;
		for (i = 0; i < S2218_NUM_REFS; i++, pbuf += 4, pinfo++)
			ExtractVFinfo(pbuf, 0, pinfo);
	}
	return x.rsp.rc;
}

// Get sensor channel raw data
static int GetExternalSnapshot(DEVICE *dev, u8 chan, S2218_EXT_SNAPSHOT *s)
{
	XACT x;

	x.cmd.opcode  = CMD2218_GET_EXTERNAL_SNAPSHOT;
	x.cmd.datalen = 1;
	x.cmd.data[0] = chan;
	x.rsp.datalen = 8;
	// Fetch sensor raw data: tsettlehi, tsettlelo, ctrlreg, acc0, acchi, acclo, iterhi, iterlo
	Transact(dev, &x);
	s->tsettle  = (x.rsp.data[0] << 8) + x.rsp.data[1];
	s->ctrlreg  = x.rsp.data[2];
	s->acc0     = x.rsp.data[3];
	ExtractVFinfo(&x.rsp.data[4], s->acc0, &s->vfinfo); // extract acc, iters; compute freq
	return x.rsp.rc;
}

// Write/read serial number
static int WriteSN(DEVICE *dev, u32 n)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SN_WRITE;
	x.cmd.datalen = 4;
	x.rsp.datalen = 0;
	U32ToDevBuf(n, x.cmd.data);
	return Transact(dev, &x);
}
static int ReadSN(DEVICE *dev, u32 *n)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SN_READ;
	x.cmd.datalen = 0;
	x.rsp.datalen = 4;
	Transact(dev, &x);
	*n = DevBufToU32(x.rsp.data);
	return x.rsp.rc;
}

// Halt scanning on logical channel
static int HaltOnChannel(DEVICE *dev, u8 logchan)
{
	XACT x;

	x.cmd.opcode  = CMD2218_HALT_ON_CHAN;
	x.cmd.datalen = 1;
	x.cmd.data[0] = logchan;
	return TransactEmpty(dev, &x);
}

// Halt scanning on manual control_register config
static int HaltOnSetup(DEVICE *dev, u8 setup)
{
	XACT x;

	x.cmd.opcode  = CMD2218_HALT_ON_SETUP;
	x.cmd.datalen = 1;
	x.cmd.data[0] = setup;
	return TransactEmpty(dev, &x);
}

// Read EEPROM ---------------------

static double ExtractReal(u8 *buf)
{
	int m = (int)((((u32)buf[2] << 8) | (u32)buf[1]) << 8) | (u32)buf[0];
	int x = (int)(u32)buf[3];
	if (x == 0xFF)      return 0.0;     // indicate unprogrammed
	else if (x > 228)   x = 76;
	else if (x < 28)    x = -124;
	else                x -= 152;
	return ((buf[2] & 0x80) ? -1.0 : 1.0) * (m | 0x800000) * pow(2.0, x);
}

// Fetch EEPROM values
static int ReadEeprom(DEVICE *dev, S2218_EEPROM_DATA *eedata)
{
	XACT x;

	x.cmd.opcode  = CMD2218_READ_EEPROM;
	x.cmd.datalen = 0;
	x.rsp.datalen = 7 * 4; // response buf: 7 reals
	Transact(dev, &x);
	eedata->V5    = ExtractReal(&x.rsp.data[0]); // LG ref
	eedata->MV500 = ExtractReal(&x.rsp.data[4]); // HG ref
	eedata->R0    = ExtractReal(&x.rsp.data[8]); // R ref (400 ohm range); next 2 reals ignored
	eedata->R1    = ExtractReal(&x.rsp.data[20]);// R ref (4 kohm range)
	eedata->R2    = ExtractReal(&x.rsp.data[24]);// R ref (600 kohm range)
	return x.rsp.rc;
}

// Manually read sensor data
static int ReadSensorData(DEVICE *dev, int chan, u16 *rawdata)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_SENSOR_DATA;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)chan;
	x.rsp.datalen = sizeof(u16);
	Transact(dev, &x);
	*rawdata = DevBufToU16(x.rsp.data);
	return x.rsp.rc;
}

// Read cold junction temperature sensor
static int ReadDeviceTemp(DEVICE *dev, u16 *rawdata)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_CJ_TEMPERATURE;
	x.cmd.datalen = 0;
	x.rsp.datalen = sizeof(u16);
	Transact(dev, &x);
	*rawdata = DevBufToU16(x.rsp.data);  // (host endian)rawdata
	return x.rsp.rc;
}

// Data alarms -------------------------------------------------------------------
// Set open-sensor data values to min/max
static int SetOpenSensorValues(DEVICE *dev, u8 failHighFlags)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_OPEN_VALUES;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)failHighFlags;
	return TransactEmpty(dev, &x);
}

// Set one limit configuration item for one sensor channel
static int SetAlarmConfig(DEVICE *dev, u8 chan, u8 itemtype, s16 itemval)
{
	XACT x;
	u8 *pbuf = x.cmd.data;
	x.cmd.opcode = CMD2218_SET_LIMIT_ITEM;
	*pbuf++ = chan;
	*pbuf++ = itemtype;
	*pbuf++ = MSB(itemval);
	*pbuf++ = LSB(itemval);
	x.cmd.datalen = pbuf - x.cmd.data;
	return TransactEmpty(dev, &x);
}

// Get all limit configuration info for one sensor channel
static int GetAlarmConfig(DEVICE *dev, u8 chan, DEV_ALARM_CONFIG *cfg)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_LIMITS;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)chan;
	x.rsp.datalen = 5;
	Transact(dev, &x);
	cfg->Flags = x.rsp.data[0];
	cfg->LoLimit = DevBufToU16(&x.rsp.data[1]); // endian correct
	cfg->HiLimit = DevBufToU16(&x.rsp.data[3]); // endian correct
	return x.rsp.rc;
}

// Strain gauges -----------------------------------------------------------

// Assert zero-load applied
static int SetGaugeZero(DEVICE *dev, int chan)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_GAUGE_ZERO;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)chan;
	return TransactEmpty(dev, &x);
}

// Assert full-load applied
static int SetGaugeSpan(DEVICE *dev, int chan, int span)
{

	XACT x;
	x.cmd.opcode  = CMD2218_SET_GAUGE_SPAN;
	x.cmd.datalen = 3;
	x.cmd.data[0] = (u8)chan;
	x.cmd.data[1] = MSB(span);
	x.cmd.data[2] = LSB(span);
	x.rsp.datalen = 3;
	return TransactEmpty(dev, &x);
}

// Adjust offset
static int TareGauge(DEVICE *dev, int chan)
{
	XACT x;
	x.cmd.opcode  = CMD2218_TARE_GAUGE;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)chan;
	return TransactEmpty(dev, &x);
}

// Activate gauge calibration
static int SetGaugeCal(DEVICE *dev, int chan, u8 caldata[6])
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_GAUGE_CALIBRATION;
	x.cmd.datalen = 7;
	x.cmd.data[0] = (u8)chan;
	memcpy(&x.cmd.data[1], caldata, 6);
	return TransactEmpty(dev, &x);
}

// Read gauge calibration
static int GetGaugeCal(DEVICE *dev, int chan, u8 caldata[6])
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_GAUGE_CALIBRATION;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)chan;
	return TransactDirect(dev, &x, 6, caldata);
}

// GPIOs ----------------------------------------------------------------------
// Program gpio output states
static int SetGpioOutputs(DEVICE *dev, u8 states)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_GPIO_OUTPUTS;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)states;
	return TransactEmpty(dev, &x);
}

// Read gpio pin states
static int GetGpioInputs(DEVICE *dev, u8 *states)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_GPIO_INPUTS;
	x.cmd.datalen = 0;
	return TransactDirect(dev, &x, 1, states);
}

// Program gpio edge capture enables
static int SetGpioCapEnables(DEVICE *dev, S2218_GPIO_PAIR *enables)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_GPIO_CAP_ENABLES;
	x.cmd.datalen = 2;
	x.cmd.data[0] = enables->Rising;
	x.cmd.data[1] = enables->Falling;
	return TransactEmpty(dev, &x);
}

// Program gpio configurations (pin direction flags,
//                              datatrig enable flags, rise/fall debounce times[]).
static int SetGpioConfig(DEVICE *dev, S2218_GPIO_CONFIG *cfg)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_GPIO_CONFIG;
	x.cmd.datalen = sizeof(S2218_GPIO_CONFIG);
	// cfg[] members don't require endian correction because there are no 16-bit values
	memcpy(x.cmd.data, cfg, sizeof(S2218_GPIO_CONFIG));
	return TransactEmpty(dev, &x);
}

// Read gpio settings
static int GetGpioSettings(DEVICE *dev, S2218_GPIO_SETTINGS *settings)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_GPIO_SETTINGS;
	x.cmd.datalen = 0;
	//   (endian correction not necessary)
	return TransactDirect(dev, &x, sizeof(S2218_GPIO_SETTINGS), settings);
}

// Calibrate ----------------------------------------------------------------
// Calibrate internal reference
static int CalibrateStandard(DEVICE *dev, int chan, int calcode, int rawdata, double *rval)
{
	XACT x;
	x.cmd.opcode  = CMD2218_CALIBRATE_REF;
	x.cmd.datalen = 4;
	x.cmd.data[0] = (u8)chan;
	x.cmd.data[1] = (u8)calcode;
	x.cmd.data[2] = MSB(rawdata);
	x.cmd.data[3] = LSB(rawdata);
	x.rsp.datalen = 4;
	Transact(dev, &x);
	*rval = ExtractReal(x.rsp.data);
	return x.rsp.rc;
}

// Calibrate cold-junction temperature sensor
static int CalibrateCJ(DEVICE *dev, u8 dir, u8 stepcount)
{
	XACT x;
	x.cmd.opcode  = CMD2218_CALIBRATE_CJ;
	x.cmd.datalen = 2;
	x.cmd.data[0] = dir;
	x.cmd.data[1] = stepcount;
	return TransactEmpty(dev, &x);
}

// Diagnostic/other --------------------------------------------------------
// Write fpga port
static int WriteReg(DEVICE *dev, int addr, int data)
{
	XACT x;
	x.cmd.opcode  = CMD2218_RAW_WRITE;
	x.cmd.datalen = 2;
	x.cmd.data[0] = (u8)addr;
	x.cmd.data[1] = (u8)data;
	return TransactEmpty(dev, &x);
}

// Read fpga port
static int ReadReg(DEVICE *dev, int addr, u8 *data)
{
	XACT x;
	x.cmd.opcode  = CMD2218_RAW_READ;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)addr;
	return TransactDirect(dev, &x, 1, data);
}

// Lock/unlock diagnostic r/w port accesses
static int LockReg(DEVICE *dev, u8 lock)
{
	XACT x;
	x.cmd.opcode  = CMD2218_RAW_RW_LOCK;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)lock;
	return TransactEmpty(dev, &x);
}

// Manually trigger a status or data report
static int TrigReport(DEVICE *dev, u8 flags)
{
	XACT x;
	x.cmd.opcode  = CMD2218_TRIG_NOTIFICATION;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)flags;
	return TransactEmpty(dev, &x);
}

// Enable/disable smad power
static int SetPowerEnables(DEVICE *dev, u8 flags)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SET_POWER_ENABLES;
	x.cmd.datalen = 1;
	x.cmd.data[0] = flags;
	return TransactEmpty(dev, &x);
}

static int GetPowerEnables(DEVICE *dev, u8 *flags)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_POWER_ENABLES;
	x.cmd.datalen = 0;
	return TransactDirect(dev, &x, 1, flags);
}

// Reset analog system
static int AnalogReset(DEVICE *dev)
{
	XACT x;
	x.cmd.opcode  = CMD2218_SMAD_RESET;
	x.cmd.datalen = 0;
	return TransactEmpty(dev, &x);
}

// Read task statistics
static int GetStatistics(DEVICE *dev, int taskid, S2218_TASKINFO *info)
{
	XACT x;
	x.cmd.opcode  = CMD2218_GET_TASK_STATISTICS;
	x.cmd.datalen = 1;
	x.cmd.data[0] = (u8)taskid;
	x.rsp.datalen = sizeof(S2218_TASKINFO);
	Transact(dev, &x);
	memcpy(info->Name, x.rsp.data, 3);
	info->Name[3] = 0; // string terminator
	info->StackMargin = DevBufToU16(&x.rsp.data[4]);
	info->ExecTime    = DevBufToU16(&x.rsp.data[6]);
	info->LoopTime    = DevBufToU16(&x.rsp.data[8]);
	return x.rsp.rc;
}

// Command helper function: Map device ID to associated DEVICE structure.
static int GetDev(int devaddr, DEVICE **ppdev)
{
	DEVICE *dev;
	int rc = 0;

	// Abort if api is closed
	if (!ApiOpen)
		return S2218_ERR_API_CLOSED;
         // Abort if illegal device address
	if ((uint)devaddr >= S2218_MAX_DEVICES)
		return S2218_ERR_ARG;
	dev = &DeviceInfo[devaddr];
	// Abort if HID device is closed
	if (dev->HidDev == NULL)
		return S2218_ERR_RESOURCE_CLOSED;

	// Return pointer to device's control structure
	if (rc == S2218_ERR_OK)
		*ppdev = dev;
	return rc;
}

// Command helper function: Map device ID to control structure and validate channel number.
static int GetDevValidateChan(int devaddr, DEVICE **ppdev, int chan)
{
	int rc = GetDev(devaddr, ppdev);
	if (rc == S2218_ERR_OK)
		if ((uint)chan >= S2218_NUM_CHANS)
			rc = S2218_ERR_ARG;
	return rc;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  PUBLIC API FUNCTIONS  ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Open API and enumerate all 2218 devices.
// Returns:
//   if error     - error code (negative value)
//   if no error  - set of flags (zero or positive),
//                  one per logical device address: 1 = device detected, 0 = not detected

int s2218_OpenApi(int *devflags)
{
	int i;
	int rc;
	DEVICE *dev = DeviceInfo;

	*devflags = 0;
	// Abort if api is already open.
	if (ApiOpen) {
		return S2218_ERR_API_OPEN;
	}
	rc = hid_system_open();
	if (rc != S2218_ERR_OK) {
		return rc;
	}
	// Init API storage:
	for (i = 0; i < S2218_MAX_DEVICES; i++, dev++) {
		//   address reverse-lookup
		InitDevStruct(dev);
		dev->devaddr = i;
	}
	// Enumerate devices and get detected_module flags.
	hid_devices_detect(VENDOR_ID, DEVICE_ID, devflags);
	ApiOpen = TRUE;

	return S2218_ERR_OK;
}

//////////////////////////////////////////
// Close API.
int s2218_CloseApi(void)
{
	int i;
	DEVICE *dev = DeviceInfo;

	if (!ApiOpen)
		return S2218_ERR_API_CLOSED;

	for (i = 0; i < S2218_MAX_DEVICES; i++)
		CloseDevice(dev++);

	hid_system_close();
	ApiOpen = FALSE;
	return S2218_ERR_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Open 2218 device.
// Imports:
//  devaddr  - logical device address
//  online   - returns TRUE if smad online

int s2218_OpenDevice(int devaddr, int *online)
{
	int rc;
	DEVICE *dev;
	S2218_STATUSREPORT status;

	if (!ApiOpen)                                   // Abort if api closed.
		return S2218_ERR_API_CLOSED;

	if (devaddr >= S2218_MAX_DEVICES)               // Abort if device address invalid.
		return S2218_ERR_ARG;

	dev = &DeviceInfo[devaddr];

	rc = OpenDevice(dev);                      // Open HID device.
	if (rc != S2218_ERR_OK)
		return rc;

	rc = TrigReport(dev, REQ_STATUS_REPORT);   // Request a device status report.
	if (rc != S2218_ERR_OK)
		return rc;

	// Wait for the requested report.
	rc = ReadSoftStatusFifo(dev, &status, MAX_STATUS_WAIT);
	if (rc != S2218_ERR_OK)
		return rc;
	//   *online = (status.ComState == CS_ONLINE);   // Get smad state.
	return rc;
}

/////////////////////////////////////////////////////
// Shutdown 2218 device.
// Imports:
//  devaddr  - logical device address
int s2218_ShutdownDevice(int devaddr)
{
	if (!ApiOpen)                               // Abort if api is closed
		return S2218_ERR_API_CLOSED;
	if (devaddr >= S2218_MAX_DEVICES)           // Abort if device id invalid
		return S2218_ERR_ARG;
	return ShutdownDevice(&DeviceInfo[devaddr]);   // Shutdown device
}

/////////////////////////////////////////////////////
// Close 2218 device.
// Imports:
//  devaddr  - logical device address
int s2218_CloseDevice(int devaddr)
{
	if (!ApiOpen)                               // Abort if api is closed
		return S2218_ERR_API_CLOSED;
	if (devaddr >= S2218_MAX_DEVICES)           // Abort if device id invalid
		return S2218_ERR_ARG;
	return CloseDevice(&DeviceInfo[devaddr]);   // Close device if it's open
}

////////////////////////////////////////////////////
// Map rc to descriptive string.

const char *s2218_ErrString(int rc)
{
	switch (rc) {
	case S2218_ERR_OK:
		return "No error";
	case S2218_ERR_OPCODE:
		return "Command not recognized";
	case S2218_ERR_ARG:
		return "Illegal argument value";
	case S2218_ERR_UNPOWERED:
		return "Analog circuitry unpowered";
	case S2218_ERR_OFFLINE:
		return "Analog system offline";
	case S2218_ERR_CRMT:
                return "CRMT timeout";
	case S2218_ERR_DAV:
		return "DAV timeout";
	case S2218_ERR_NO_EFFECT:
		return "Command has no effect";
	case S2218_ERR_LOCKOUT:
		return "Command prohibited in diagnostic mode";
	case S2218_ERR_SCANNING:
		return "Command prohibited while scanning";
		// API
	case S2218_ERR_TIMEOUT:
		return "Time-out waiting for device. If relevant, use more specific LOCK or SIG timeout";
	case S2218_ERR_FAULT:
		return "Couldn't execute command due to fault (check FaultFlags)";
	case S2218_ERR_REOPEN:
		return "Attempted to open smad when already open";
	case S2218_ERR_CREATE_RESOURCE:
		return "Failed to create mutex/semaphore/thread";
	case S2218_ERR_RESOURCE_CLOSED:
		return "Resource closed or unavailable";
	case S2218_ERR_DEVICE_CLOSED:
		return "Device closed";
	case S2218_ERR_DEVICE_OPEN:
		return "Device already open";
	case S2218_ERR_WAS_SHUTDOWN:
		return "USB connection was shutdown";
	case S2218_ERR_READ:
                return "USB read error";
	case S2218_ERR_WRITE:
		return "USB write error";
	case S2218_ERR_OPENHIDLIB:
		return "Failed to open HIDLIB";
	case S2218_ERR_DEVICE_NOT_FOUND:
		return "Device not found";
	case S2218_ERR_COMMAND_MISMATCH:
		return "Response didn't match last issued command";
	case S2218_ERR_API_OPEN:
		return "API already open";
	case S2218_ERR_API_CLOSED:
		return "API closed";
	case S2218_ERR_LOCK_TIMEOUT:
		return "Failed to get resource lock";
	case S2218_ERR_SIG_TIMEOUT:
		return "Time-out waiting for message from device";
	case S2218_ERR_NOT_ALLOWED:
		return "Prohibited by current settings";
	case S2218_ERR_DUPLICATE_ADDRESS:
		return "Two modules have the same address";
	case S2218_ERR_CANT_OPEN_DEVICE:
		return "Failed to open device";
	case S2218_ERR_MALLOC:
		return "Memory allocation failed";
	case S2218_ERR_SYSTEM:
		return "System error";
	default:
		return "Unknown error";
	}
}

// Analog configuration -------------------------------------------------
// Configure sensor type
int s2218_SetSensorType(int devaddr, int chan, SENSOR_TYPE sdc)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	if (rc != S2218_ERR_OK)
		return rc;
	if (sdc >= S2218_NUM_SENSOR_TYPES)
		return S2218_ERR_ARG;  // illegal sensor type

	// Configure channel's thermocouple conditioner
	rc = SetChanTcCondEnable(dev, chan, sensor_attr[sdc].tcCondEnab);
	if (rc != S2218_ERR_OK)
		return rc;
	rc = SetSensorType(dev, chan, sdc);    // Program sensor type
	if (rc == S2218_ERR_OK)
		SyncChanImage(dev, chan, sdc);          // Sync local image to device config
	return rc;
}

// Enable/disable a channel's thermocouple conditioner
// Note: a channel's conditioners are automatically enabled when configured for thermocouples and
//       disabled for all other sensor types (via s2218_SetSensorType).
int s2218_SetTcCondEnable(int devaddr, int chan, int enable)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	return (rc == S2218_ERR_OK) ? SetChanTcCondEnable(dev, chan, enable) : rc;
}

// Read smad configuration
int s2218_GetAnalogConfig(int devaddr, S2218_ANALOG_CONFIG *cfg)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? GetAnalogConfig(dev, cfg) : rc;
}

// Data acquisition -------------------------------------------------
// Enable gpio-triggered data reporting and set periodic report rate (as multiple of scan interval)

/*
  int s2218_SetStreamRate(int devaddr, u16 rate)
  {
  DEVICE *dev;
  int rc = GetDev(devaddr, &dev);
  return (rc == S2218_ERR_OK) ? SetStreamControl(dev, rate, TRUE) : rc;//and start stream.
  }
*/

int s2218_SetTrigConfig(int devaddr, S2218_TRIG_CONFIG *cfg)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? SetTrigConfig(dev, cfg) : rc;
}

int s2218_GetTrigConfig(int devaddr, S2218_TRIG_CONFIG *cfg)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? GetTrigConfig(dev, cfg) : rc;
}

int s2218_SetTrigEnable(int devaddr, u8 enable)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? SetTrigEnable(dev, enable) : rc;
}

int s2218_GetTrigEnable(int devaddr, u8 *enable)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? GetTrigEnable(dev, enable) : rc;
}

/*
// Manually trigger a data report
int s2218_TrigDataReport(int devaddr)
{
DEVICE *dev;
int rc = GetDev(devaddr, &dev);
return (rc == S2218_ERR_OK) ? TrigReport(dev, REQ_DATA_REPORT) : rc;
}
*/

////////////////////////////////////////////////////////////////////////////////
// Wait for a data report from stream receiver FIFO.
// Imports:
//  samples  - receives data report (sensor data and timestamp).
//  tmax     - max time (in milliseconds) to wait for data report.
// Returns:
//  S2218_ERR_OK            - Success. This is the only case in which the returned data is valid.
//  S2218_ERR_LOCK_TIMEOUT  - Failed because another thread is hogging the resource lock.
//  S2218_ERR_SIG_TIMEOUT   - Failed: timed out waiting for data to arrive from the device.
//  Other                   - Failed; see error code descriptions.

int s2218_WaitForDataReport(int devaddr, S2218_DATAREPORT *report, int tmax)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	return (rc == S2218_ERR_OK) ? ReadDataFifo(dev, report, tmax) : rc;
}

// Manually read sensor data
int s2218_ReadSensorData(int devaddr, int chan, double *data, int tmax)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	if (rc == S2218_ERR_OK) {
		u16 raw; // raw sensor data
		rc = ReadSensorData(dev, chan, &raw);
		*data = raw * dev->Sensor[chan].Scalar;  // convert raw to engr
	}
	return rc;
}

// Read cold junction temperature sensor
int s2218_ReadDeviceTemp(int devaddr, double *degC)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);

	if (rc == S2218_ERR_OK) {
		u16 raw;
		rc = ReadDeviceTemp(dev, &raw);
		*degC = raw * 0.10;  // rawdata * scalar
	}
	return rc;
}

// Alarms --------------------------------------------------------------

// Set open-sensor data values to min/max (configures all sensor channels).
int s2218_SetOpenSensorValues(int devaddr, u8 failHighFlags)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? SetOpenSensorValues(dev, failHighFlags) : rc;
}

// Helper: Convert sensor data value (engr units) to smad native integer value.
static short SensorEngrToRaw(double engr, double scalar)
{
	if (scalar == 0.0)   // Default if sensor channel disabled
		return 0;
	else {
		double raw = engr / scalar;                     // native smad sensor data
		if (raw > 32767.0)          return 32767;       // max raw value
		else if (raw < -32768.0)    return -32768;      // min raw value
		else                        return (s16)raw;
	}
}

// Configure high or low threshold for specified sensor channel.
static int SetThreshold(int devaddr, u8 chan, u8 itemtype, double limval)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	return (rc == S2218_ERR_OK) ?
		SetAlarmConfig(dev, chan, itemtype, SensorEngrToRaw(limval, dev->Sensor[chan].Scalar)) : rc;
}

int s2218_SetLoThreshold(int devaddr, u8 chan, double limval)
{
	return SetThreshold(devaddr, chan, SEL_LO_THRESHOLD, limval);
}

int s2218_SetHiThreshold(int devaddr, u8 chan, double limval)
{
	return SetThreshold(devaddr, chan, SEL_HI_THRESHOLD, limval);
}

// Configure one limit or alarm enable for specified sensor channel.
static int SetThreshEnable(int devaddr, u8 chan, u8 itemtype, short enab)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	return (rc == S2218_ERR_OK) ? SetAlarmConfig(dev, chan, itemtype, enab) : rc;
}

int s2218_SetLoLimitEnable(int devaddr, u8 chan, short enab)
{
	return SetThreshEnable(devaddr, chan, SEL_LO_LIMIT_ENAB, enab);
}

int s2218_SetHiLimitEnable(int devaddr, u8 chan, short enab)
{
	return SetThreshEnable(devaddr, chan, SEL_HI_LIMIT_ENAB, enab);
}

int s2218_SetLoAlarmEnable(int devaddr, u8 chan, short enab)
{
	return SetThreshEnable(devaddr, chan, SEL_LO_ALARM_ENAB, enab);
}

int s2218_SetHiAlarmEnable(int devaddr, u8 chan, short enab)
{
	return SetThreshEnable(devaddr, chan, SEL_HI_ALARM_ENAB, enab);
}

// Read alarm configuration for specified sensor channel.
int s2218_GetAlarmConfig(int devaddr, u8 chan, S2218_ALARM_CONFIG *appcfg)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);

	if (rc == S2218_ERR_OK) {
		DEV_ALARM_CONFIG devcfg;
		if ((rc = GetAlarmConfig(dev, chan, &devcfg)) == S2218_ERR_OK) {
			double scalar   = dev->Sensor[chan].Scalar;
			appcfg->Flags   = devcfg.Flags;
			appcfg->LoLimit = devcfg.LoLimit * scalar;    // convert raw to engr
			appcfg->HiLimit = devcfg.HiLimit * scalar;
		}
	}
	return rc;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Wait for limit alarm.
// Imports:
//  alarms   - receives alarm report.
//  tmax     - max time (in milliseconds) to wait for alarms.
// Returns:
//  S2218_ERR_OK            - Success. This is the only case in which the returned data is valid.
//  S2218_ERR_LOCK_TIMEOUT  - Failed because another thread is hogging the resource lock.
//  S2218_ERR_SIG_TIMEOUT   - Failed: timed out waiting for data to arrive from the device.
//  Other                   - Failed; see error code descriptions.
int s2218_WaitForLimitAlarm(int devaddr, S2218_ALARMREPORT *report, int tmax)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadAlarmFifo(dev, report, tmax) : rc;
}

// Strain gauge ------------------------------------------------------
// Assert zero-load applied
int s2218_SetGaugeZero(int devaddr, int chan)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	return (rc == S2218_ERR_OK) ? SetGaugeZero(dev, chan) : rc;
}

// Assert full-load applied
int s2218_SetGaugeSpan(int devaddr, int chan, double span)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	return (rc == S2218_ERR_OK) ? SetGaugeSpan(dev, chan, (int) span) : rc;
}

// Tare gauge
int s2218_TareGauge(int devaddr, int chan)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	return (rc == S2218_ERR_OK) ? TareGauge(dev, chan) : rc;
}

// Install gauge calibration
int s2218_SetGaugeCal(int devaddr, int chan, u8 caldata[6])
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	return (rc == S2218_ERR_OK) ? SetGaugeCal(dev, chan, caldata) : rc;
}

// Read gauge calibration
int s2218_GetGaugeCal(int devaddr, int chan, u8 caldata[6])
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	return (rc == S2218_ERR_OK) ? GetGaugeCal(dev, chan, caldata) : rc;
}

// GPIO ---------------------------------------------------------------

// Program output states
int s2218_SetGpioOutputs(int devaddr, u8 states)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? SetGpioOutputs(dev, states) : rc;
}

// Read pin states
int s2218_GetGpioInputs(int devaddr, u8 *states)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetGpioInputs(dev, states) : rc;
}

// Program gpio configurations
int s2218_SetGpioConfig(int devaddr, S2218_GPIO_CONFIG *cfg)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? SetGpioConfig(dev, cfg) : rc;
}

// Read gpio settings
int s2218_GetGpioSettings(int devaddr, S2218_GPIO_SETTINGS *settings)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetGpioSettings(dev, settings) : rc;
}

// Program edge capture enables
int s2218_SetGpioCapEnables(int devaddr, S2218_GPIO_PAIR *enables)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? SetGpioCapEnables(dev, enables) : rc;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Wait for captured GPIO events.
// Imports:
//  report   - receives GPIO event report.
//  tmax     - max time (in milliseconds) to wait for events.
// Returns:
//  S2218_ERR_OK            - Success. This is the only case in which the returned data is valid.
//  S2218_ERR_LOCK_TIMEOUT  - Failed because another thread is hogging the resource lock.
//  S2218_ERR_SIG_TIMEOUT   - Failed: timed out waiting for data to arrive from the device.
//  Other                   - Failed; see error code descriptions.
int s2218_WaitForGpioEvent(int devaddr, S2218_GPIOREPORT *report, int tmax)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadGpioFifo(dev, report,tmax) : rc;
}

// Calibration ----------------------------------------------
/////////////////////////////////////////////////////////////////////////////
// Calibrate internal reference
// Imports:
//  chan        - channel number used to measure external reference standard
//  calID       - range ID (0 to 4)
//  sensval     - expected after-cal sensor data value
//  refval      - receives computed value of reference standard

#define NUM_CAL_RANGES  5

static const int CalCode[NUM_CAL_RANGES] = {0, 1, 2, 5, 6};
static const double CalScalar[NUM_CAL_RANGES] = {  // smad rawdata scalars
	SCALAR_TODEV_LG,    // 5V Range
	SCALAR_TODEV_HG,    // 500mV Range
	SCALAR_TODEV_RL,    // 400 Ohm Range
	SCALAR_TODEV_RM,    // 4K Ohm Range
	SCALAR_TODEV_RH     // 600K Ohm Range
};

int s2218_CalibrateStandard(int devaddr, int chan, int calID, double sensval, double *refval)
{
	DEVICE *dev;
	int rc = GetDevValidateChan(devaddr, &dev, chan);
	// Check chan and devaddr
	if (rc != S2218_ERR_OK)
		return rc;
	// Check range
	if ((uint)calID >= NUM_CAL_RANGES)
		return S2218_ERR_ARG;
	// Convert sensor data to smad native rawdata
	sensval *= CalScalar[calID];
	if ((sensval < -32000) || (sensval > 32000))
		return S2218_ERR_ARG;
	// Invoke calibrate
	return CalibrateStandard(dev, chan, CalCode[calID], (int)sensval, refval);
}

//////////////////////////////////////////////////////////
// Calibrate cold-junction temperature sensor
int s2218_CalibrateCJ(int devaddr, u8 dir, u8 stepcount)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? CalibrateCJ(dev, dir, stepcount) : rc;
}

// Version info ----------------------------------------------
// Read middleware version
int s2218_GetApiVersion(S2218_VERSION *ver)
{
	ver->Major  = S2218_API_VERSION_MAJOR;
	ver->Minor  = S2218_API_VERSION_MINOR;
	ver->Build  = S2218_API_VERSION_BUILD;
	return S2218_ERR_OK;
}

// Read device versions info
int s2218_GetDeviceVersions(int devaddr, S2218_VERSION_INFO *ver)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetDeviceVersions(dev, ver) : rc;
}

// Status reports ----------------------------------------------
// Manually trigger and receive a synchronous status report.
int s2218_GrabStatus(int devaddr, S2218_STATUSREPORT *status, int tmax)
{
	DEVICE *dev;
	int rc;
        // Ask 2218 to send a synchronous status report?
	// Receive the report
	if ((rc = GetDev(devaddr, &dev)) == S2218_ERR_OK)
		if ((rc = TrigReport(dev, REQ_STATUS_REPORT)) == S2218_ERR_OK)
			rc = ReadSoftStatusFifo(dev, status, tmax);
	return rc;
}

////////////////////////////////////////////////////////////////////////////////////
// Wait for async status change.
// The device automatically sends a status update message whenever its system
// state changes to Online, PowerOff, BootTimeout, PowerTimeout, or PowerDropout.
// Imports:
//  status              - receives status info.
//  status->nDropped    - receives number of dropped messages.
//  tmax                - max time (in milliseconds) to wait for status update message.
// Returns:
//  S2218_ERR_OK            - Success. This is the only case in which the returned data is valid.
//  S2218_ERR_LOCK_TIMEOUT  - Failed because another thread is hogging the resource lock.
//  S2218_ERR_SIG_TIMEOUT   - Failed: timed out waiting for data to arrive from the device.
//  Other                   - Failed; see error code descriptions.
int s2218_WaitForAsyncStatusReport(int devaddr, S2218_STATUSREPORT *status, int tmax)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadAsyncStatusFifo(dev, status, tmax) : rc;
}

// Diagnostic/other ----------------------------------------------
// Write fpga port
int s2218_WriteReg(int devaddr, int addr, int data)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? WriteReg(dev, addr, data) : rc;
}

// Read fpga port
int s2218_ReadReg(int devaddr, int addr, u8 *data)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadReg(dev, addr, data) : rc;
}

// Lock/unlock diagnostic access to fpga ports
int s2218_LockReg(int devaddr, u8 lock)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? LockReg(dev, lock) : rc;
}

// Enable/disable power supplies
int s2218_SetPowerEnables(int devaddr, u8 enab)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? SetPowerEnables(dev, enab) : rc;
}

// Get smad power settings
int s2218_GetPowerEnables(int devaddr, u8 *enab)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetPowerEnables(dev, enab) : rc;
}

// Reset analog system
int s2218_AnalogReset(int devaddr)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? AnalogReset(dev) : rc;    // Invoke smad reset.
}

// Read task statistics
int s2218_GetStatistics(int devaddr, int taskid, S2218_TASKINFO *info)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetStatistics(dev, taskid, info) : rc;
}

int s2218_HaltOnChan(int devaddr, u8 chan)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? HaltOnChannel(dev, chan) : rc;
}

int s2218_HaltOnSetup(int devaddr, u8 setup)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? HaltOnSetup(dev, setup) : rc;
}

int s2218_ReadEeprom(int devaddr, S2218_EEPROM_DATA *eedata)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadEeprom(dev, eedata) : rc;
}

int s2218_GetIntSnapshot(int devaddr, S2218_INT_SNAPSHOT *s)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetInternalSnapshot(dev, s) : rc;
}

int s2218_GetExtSnapshot(int devaddr, u8 chan, S2218_EXT_SNAPSHOT *s)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? GetExternalSnapshot(dev, chan, s) : rc;
}

int s2218_ReadSN(int devaddr, u32 *sn)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? ReadSN(dev, sn) : rc;
}

int s2218_WriteSN(int devaddr, u32 sn)
{
	DEVICE *dev;
	int rc = GetDev(devaddr, &dev);
	return (rc == S2218_ERR_OK) ? WriteSN(dev, sn) : rc;
}
