////////////////////////////////////////////////////////////////////////////////
// File     : s2218_platform.c
// System   : Linux
// Function : OS-dependent functions for USB and mutexes
////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "s2218_platform.h"
#include "api2218.h"
#include "s2218_core.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "hidapi.h"

#define HID_OUT_CTL_CODE(id)						\
	CTL_CODE(FILE_DEVICE_KEYBOARD, (id), METHOD_OUT_DIRECT, FILE_ANY_ACCESS)

#ifdef __cplusplus
} // extern "C"
#endif

// Windows inserts a report number (0x00) at the beginning of the report
// Linux may not, however.
#define USB_PACKETSIZE      (S2218_BUFSIZE + 1)

struct HID_DEVICE_ {
	hid_device      *hid_dev;
	//HANDLE          rd_handle;
	//HANDLE          wr_handle;
};

typedef struct HIDD_ATTRIBUTES {
	unsigned long     Size;
	unsigned short    VendorID;
	unsigned short    ProductID;
	unsigned short    VersionNumber;
} HIDD_ATTRIBUTES;

typedef struct HID_DEVINFO {                // Device enumeration info
	int             nMatches;
	u16             HwVersion;
	char            DevicePath[256];
} HID_DEVINFO;

// Storage -----------------------------------------------------------
static HID_DEVINFO  DevInfo[16] = {0,};             // Device enumeration info

// Functions ---------------------------------------------------------
////////////////////////////////////////////////////////
// Open/close the HID module.

// Close the HID API.
int hid_system_close(void)
{
	hid_exit();
	return 0;
}

// Open the HID API.
int hid_system_open(void)
{
	hid_init();
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Determine if specified device is the one we seek
// (i.e., matching vendor/product IDs and switch address).
// If so, get its hw_version and return TRUE, else return FALSE.
#define SN_VALUE(STR)   (10 * (STR[0] - 0x30) + (STR[1] - 0x30))    // Convert SN to integer
////////////////////////////////////////////////////////////////////////////////////////////////////
// Enumerate module at specified path.
static void EnumDevice(char *DevicePath, u16 vendor_id, u16 product_id)
{
	wchar_t SnString[126];      // Device serial number string
	hid_device *dev;
	struct hid_device_info *info;
	struct hid_device_info *devs;;
	uint devaddr;
	dev = hid_open_path(DevicePath);
	if (dev != NULL) {
		info = hid_enumerate(vendor_id, product_id);
		devs = info;
		while (info != NULL) {
			if (strcmp(info->path, DevicePath) == 0)
				// found the a match to the path, vendor_id, product_id
				break;
			info = info->next;
		}
		if (devs == NULL) {
			hid_close(dev);
			return;
		}
		if (info == NULL) {
			hid_free_enumeration(devs);
			hid_close(dev);
			return;
		}
		hid_get_serial_number_string(dev, SnString, sizeof(SnString));
		devaddr = SN_VALUE(SnString);  // Get device address (from SN).
		if (devaddr < S2218_MAX_DEVICES) {
			if (++(DevInfo[devaddr].nMatches) == 1) // If device doesn't have a duplicate address
			{
				DevInfo[devaddr].HwVersion = info->release_number; // cache device version number
				strcpy(DevInfo[devaddr].DevicePath, DevicePath);
			}
		}
	}
	hid_close(dev);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Enumerate/re-enumerate all 2218 modules. This must be called before opening devices.
static int hid_enum_devices(u16 vendor_id, u16 product_id)
{
	struct hid_device_info *info, *info_first;

	info = hid_enumerate(vendor_id, product_id);
	info_first = info;

	if (info == NULL)
		return S2218_ERR_OPENHIDLIB;

	memset(DevInfo, 0, sizeof(DevInfo));        // Erase existing enumeration if there is one.
	while (info != NULL) {
		// Get basic device interface info.
		// If device is 2218, cache details to local enum list.
		EnumDevice(info->path, vendor_id, product_id);
		info = info->next;
	}
	hid_free_enumeration(info_first);
	return S2218_ERR_OK;
}

/////////////////////////////////////////////////////////////
// Find device with matching vendor id, product id and device address (in serial number).
// Returns errcode if error, else S2218_ERR_OK.
// Imports:
//  vendor_id   - target vendor id
//  product_id  - target product id
//  devaddr     - target address (must match module's rotary address switch)

static int hid_device_find(u16 vendor_id, u16 product_id, uint devaddr)
{
	// Re-enumerate 2218's in case devices have been plugged/unplugged since last enumeration.
	int errcode = hid_enum_devices(vendor_id, product_id);
	if (errcode != S2218_ERR_OK)
		return errcode;
	switch (DevInfo[devaddr].nMatches)
	{
	case 1:
		return S2218_ERR_OK;
	case 0:
		return S2218_ERR_DEVICE_NOT_FOUND;
	default:
		return S2218_ERR_DUPLICATE_ADDRESS;
	}
}

//////////////////////////////////////////////////////////////////
// Return set of flags indicating all detected 2218 modules.
// Returns errcode if error, else S2218_ERR_OK.
// Imports:
//  vendor_id   - target vendor id
//  product_id  - target product id
//  devflags    - receives detected_module flags

int hid_devices_detect(u16 vendor_id, u16 product_id, int *devflags)
{
	int devaddr;
	int flags = 0;  // no modules found yet
	int errcode;

	// Re-enumerate 2218's in case devices have been plugged/unplugged since last enumeration.
	errcode = hid_enum_devices(vendor_id, product_id);
	if (errcode == S2218_ERR_OK) {
		// Enumerate devices. For each possible device address:
		//   Set Detected flag if device was detected
		for (devaddr = S2218_MAX_DEVICES - 1; devaddr >= 0; devaddr--)
			flags = (flags << 1) | (DevInfo[devaddr].nMatches == 1);
	}
	*devflags = flags;
	return errcode;
}

//////////////////////////////////////////////////////////////////////////////////////////////
// Open 2218 device with matching vendor id, product id and device address (in serial number)
// and get its hid_device handle.
int hid_device_open(u16 vendor_id, u16 product_id, uint devaddr, u16 *hw_version, HID_DEVICE **pdev)
{
	int errcode = hid_device_find(vendor_id, product_id, devaddr);  // find device
	if (errcode == S2218_ERR_OK) {
		// Create hid_device object and fill it with zeros.
		HID_DEVICE *dev = calloc(1, sizeof(HID_DEVICE));
		dev->hid_dev = hid_open_path(DevInfo[devaddr].DevicePath);
		if (dev->hid_dev == NULL) {
			errcode = S2218_ERR_CANT_OPEN_DEVICE;
		} else {
			// Return hid_device struct pointer to caller.
			*pdev = dev;
			// If requested,
			//   pass device version number to app
			if (hw_version != NULL)
				*hw_version = DevInfo[devaddr].HwVersion;
		}
	}
	return errcode;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Write an Output report to a HID device.
// Send data on the first OUT endpoint if one exists, else through Control Endpoint (Endpoint 0).
// Imports:
//  dev     - device handle returned from hid_open().
//  data    - data to send.
//  datalen - length in bytes of data to send.
//  Returns number of bytes written, -1 upon error (with error code in dev->last_error).

int hid_device_write(HID_DEVICE *dev, const u8 *data, size_t datalen)
{
	int bytes_written;       // always 0x41
	u8 packetbuf[USB_PACKETSIZE];

	// Copy data to output buffer.
	// Windows WriteFile() expects number of bytes in the longest report
	// (plus one for report number) even if data is a shorter report.
	packetbuf[0] = 0;                           // Insert report number to tx buf
#ifdef IS_WINDOWS_BUILD
	// Windows...
	// Copy command to tx buf (unused buf bytes with zeros are ignored by the 2218)
	//    memcpy(&packetbuf[1], data, datalen);
	// Linux
	// Copy command to tx buf (unused buf bytes with zeros are ignored by the 2218)
#else
	memcpy(&packetbuf[0], data, datalen);
#endif
	// Invoke USB write.
	bytes_written = hid_write(dev->hid_dev, packetbuf, USB_PACKETSIZE);

	if (bytes_written != USB_PACKETSIZE)
		return S2218_ERR_WRITE;

	return S2218_ERR_OK;
}

////////////////////////////////////////////////////////////////
// Read an Input report from a HID device.
// Input reports are returned to host through INTERRUPT IN endpoint.
// The first byte contains Report number.
// Exports:
//  device   - Device handle returned from hid_open().
//  data     - Buffer to put the read data into.
//  length   - Number of bytes to read. For devices with multiple reports,
//             make sure to read an extra byte for the report number.
// Returns S2218_ERR_OK upon success, else errcode.

int hid_device_read(HID_DEVICE *dev, u8 *data, size_t length)
{
	int bytes_read;       // always 0x41
	u8 packetbuf[USB_PACKETSIZE];

	bytes_read = hid_read(dev->hid_dev, packetbuf, USB_PACKETSIZE);
	if (bytes_read != USB_PACKETSIZE)
		return S2218_ERR_READ;  // unexpected packet size
#ifdef IS_WINDOWS_BUILD
	// Windows inserts a report number (0x00) at the beginning of the report
	// but 2218 doesn't use or understand it so skip over this byte.
	//    memcpy(data, &packetbuf[1], length);
#else
	memcpy(data, &packetbuf[0], length);
#endif
	return S2218_ERR_OK;
}

void hid_device_close(HID_DEVICE *dev)
{
	if (dev) { // If device is open
		hid_close(dev->hid_dev);
		dev->hid_dev = NULL;
		free(dev); //   Free device info struct
	}
}

int hid_device_stop_poll(HID_DEVICE *dev)
{
	int rc;
	rc = hid_stop_poll(dev->hid_dev);
	return rc;
}

