/////////////////////////////////////////////////////////////////////////////
// File      : s2218alarm.c
// Function  : Simple C demo for testing/demo of alarm limits on
//             Sensoray model 2218
// Note: Scan rate per channel is 5Hz (aggregate is 40Hz), which means
//       worst case latency for alarm trigger is 200ms
// Copyright : (C) 2024 Sensoray
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include "api2218.h"
#include "mid2218.h"

static int G_sensor_type = 0;
// Function prototypes ----------
void changemode(int dir);
int kbhit(void);
void UseTheDevice(int devid);

//////////////////////////////////////////////////////
// Simple error detector/handler.
int IsError(int errcode)
{
	if (errcode == S2218_ERR_OK)
		return 0;
	printf("Error: %s\n", s2218_ErrString(errcode));
	return errcode;
}

int SetupThresholds(double low, double high, int mask, int devid)
{
	int i;
	for (i = 0; i < S2218_NUM_CHANS; i++) {
		if ((1 << i) & mask) {
			s2218_SetLoThreshold(devid, i, low);
			s2218_SetHiThreshold(devid, i, high);
			s2218_SetLoLimitEnable(devid, i, 1);
			s2218_SetHiLimitEnable(devid, i, 1);
			s2218_SetLoAlarmEnable(devid, i, 1);
			s2218_SetHiAlarmEnable(devid, i, 1);
		}
	}
	return 0;
}

int SetupSensorsFiveVolts(int devid)
{
	int i;

	// Configure all chans for +/-5V input range
	for (i = 0; i < S2218_NUM_CHANS; i++) {
		if (IsError(s2218_SetSensorType(devid, i, SDC_VOLTS_5V)))
			return -1;
		s2218_SetOpenSensorValues(devid, 0);
	}
	G_sensor_type = SDC_VOLTS_5V;
	return 0;
}

int main(int argc, char **argv)
{
	int devflags;
	int devid = 0;
	int devmask = 1;
	int online;

	// for keyboard input use
	changemode(1);
	if (!IsError(s2218_OpenApi(&devflags))) {        // Open the API
		// Abort if no devices discovered
		if (devflags == 0)
			printf("No devices found\n");
		else {
			// Find the first available device (device with lowest address)
			for (devid = 0; devid < S2218_MAX_DEVICES; devid++, devmask <<= 1) {
				if (devflags & devmask) {
					if (!IsError(s2218_OpenDevice(devid, &online))) {
						// Open the device
						UseTheDevice(devid);
						s2218_CloseDevice(devid);
					}
					break;
				}
			}
		}
		// close the API
		s2218_CloseApi();
	}
	changemode(0);
	return 0;
}

char getch() {
	char buf = 0;
	struct termios old = { 0 };
	fflush(stdout);
	if (tcgetattr(0, &old) < 0) perror("tcsetattr()");
	old.c_lflag    &= ~ICANON;   // local modes = Non Canonical mode
	old.c_lflag    &= ~ECHO;     // local modes = Disable echo.
	old.c_cc[VMIN]  = 1;         // control chars (MIN value) = 1
	old.c_cc[VTIME] = 0;         // control chars (TIME value) = 0 (No time)
	if (tcsetattr(0, TCSANOW, &old) < 0) perror("tcsetattr ICANON");
	if (read(0, &buf, 1) < 0) perror("read()");
	old.c_lflag    |= ICANON;    // local modes = Canonical mode
	old.c_lflag    |= ECHO;      // local modes = Enable echo.
	if (tcsetattr(0, TCSADRAIN, &old) < 0) perror ("tcsetattr ~ICANON");
	return buf;
}

#if 0
static char getkey(void)
{
	return (char)getch();
}
#endif

void changemode(int dir)
{
	static struct termios oldt, newt;

	if ( dir == 1 )
	{
		tcgetattr( STDIN_FILENO, &oldt);
		newt = oldt;
		newt.c_lflag &= ~( ICANON | ECHO );
		tcsetattr( STDIN_FILENO, TCSANOW, &newt);
	}
	else
		tcsetattr( STDIN_FILENO, TCSANOW, &oldt);
}

int kbhit(void)
{
	struct timeval tv;
	fd_set rdfs;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	FD_ZERO(&rdfs);
	FD_SET (STDIN_FILENO, &rdfs);
	select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
	return FD_ISSET(STDIN_FILENO, &rdfs);

}

static int keypress(void)
{
	if (!kbhit()) {
		return 0;
	}
	getch();
	return 1;
}


int ReadPowerEnables(int devid)
{
	u8 flags;  // (x, x, x, x, x, +3.3V, +15V, -12.25V)

	if (IsError(s2218_GetPowerEnables(devid, &flags)))
		return -1;

	printf("3.3 %s\n+15 %s\n-12 %s\n",
	       (flags & REQ_POWER_D) ? "ON" : "OFF",
	       (flags & REQ_POWER_P) ? "ON" : "OFF", (flags & REQ_POWER_N) ? "ON" : "OFF");
	return 0;
}

int WritePowerEnables(int devid, u8 mask, u8 flags)
{
	u8 curflags;  // (x, x, x, x, x, +3.3V, +15V, -12.25V)
	u8 newflags;

	if (IsError(s2218_GetPowerEnables(devid, &curflags)))
		return -1;
	newflags = (flags & mask) | (curflags & ~mask);
	if (IsError(s2218_SetPowerEnables(devid, newflags)))
		return -1;
	return ReadPowerEnables(devid);
}

// Status ---------------------------
int GetStatus(int devid)
{
	S2218_STATUSREPORT r;

	if (IsError(s2218_GrabStatus(devid, &r, 2000)))
		return -1;
	printf("meter:\n");
	printf(" DIG  = %6.2f (%s)\n",
	       r.Meter_mV[1] * 0.001, (r.PowerGood & POWERGOOD_3R3V) ? "ok" : "bad");
	printf(" ANA- = %6.2f (%s)\n",
	       r.Meter_mV[0] * 0.001, (r.PowerGood & POWERGOOD_N12V) ? "ok" : "bad");
	printf(" ANA+ = %6.2f (%s)\n",
	       r.Meter_mV[3] * 0.001, (r.PowerGood & POWERGOOD_P15V) ? "ok" : "bad");
	printf(" VREF = %6.2f (%s)\n",
	       r.Meter_mV[2] * 0.001, (r.PowerGood & POWERGOOD_5V)   ? "ok" : "bad");
	if (r.SmadState > SS_FAULT)
		printf("SmadState = ILLEGAL\n");
	else {
		const char *ss[] = {"Unpowered", "Initializing", "Booting", "Online", "Fault"};
		printf("SmadState = %s\n", ss[r.SmadState]);
	}
	if (r.ComState > CS_DAV_FAIL)
		printf("SmadState = ILLEGAL\n");
	else {
		const char *ss[] = {"Normal", "ILLEGAL", "ILLEGAL", "ILLEGAL",
			"Offline", "CRMT timeout", "DAV timeout"};
		printf("ComState = %s\n", ss[r.ComState]);
	}
	return 0;
}

int ShowVersionInfo(int devid)
{
	S2218_VERSION_INFO ver;
	if (IsError(s2218_GetDeviceVersions(devid, &ver)))
		return -1;
	printf("Version info: ");
	printf("dev %d.%d.%d, ", ver.DevFw.Major, ver.DevFw.Minor, ver.DevFw.Build);
	printf("smad %d.%d, ", ver.SmadFw/100, ver.SmadFw - 100*(ver.SmadFw/100));
	printf("pwb rev%c\n", ver.PwbRev);
	return 0;
}

void print_time(char *str)
{
    struct timespec ts;
    timespec_get(&ts, TIME_UTC);
    char buff[100];
    strftime(buff, sizeof buff, "%D %T", gmtime(&ts.tv_sec));
    printf("%s Current time: %s.%09ld UTC\n", str, buff, ts.tv_nsec);
}

void UseTheDevice(int devid)
{
	S2218_DATAREPORT    report;
	S2218_TRIG_CONFIG   trig;
	int rc;
	printf("\nUsing 2218 board %d\n", devid);
	WritePowerEnables(devid, REQ_POWER_P, REQ_POWER_P);
	WritePowerEnables(devid, REQ_POWER_N, REQ_POWER_N);
	ShowVersionInfo(devid);
	printf("Setting all channels to 5V sensor type\n");
	SetupSensorsFiveVolts(devid);
	usleep(1000);
	SetupThresholds(0.3, 0.9, 2, devid);
	memset(&report, 0, sizeof(S2218_DATAREPORT));
	memset(&trig, 0, sizeof(S2218_TRIG_CONFIG));
	// Set scan and report interval
	trig.Rate               = 1; // at scan rate frequency
	trig.GpioFlags.Rising   = 0;
	trig.GpioFlags.Falling  = 0;
	if (IsError(s2218_SetTrigConfig(devid, &trig)))
		return;
	if (IsError(s2218_SetTrigEnable(devid, 1)))
		return;
	printf("press any key to stop\n");
	while (!keypress()) {
		S2218_ALARMREPORT rep_alarm;
		print_time("waiting 5s for alarm: ");
		rc = s2218_WaitForLimitAlarm(devid, &rep_alarm, 5000);
		if (rc == 0) {
			print_time("alarm triggered:      ");
			printf("alarm flags: 0x%x\n", rep_alarm.Flags);
		} else {
			print_time("timedout waiting      ");
		}
		// re-activate alarms
		SetupThresholds(0.3, 0.9, 2, devid);
		printf("press any key to stop\n");
	}
	return;
}
