////////////////////////////////////////////////////////////////////////////////////////
// File      : s2218demo.c
// Function  : Simple C demo for Sensoray model 2218
// Copyright : (C) 2017-2023 Sensoray
////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include "api2218.h"
#include "mid2218.h"

// Function prototypes ----------
void changemode(int dir);
int kbhit(void);
void UseTheDevice(int devid);
static int ToggleSensorType(int devid);

//////////////////////////////////////////////////////
// Simple error detector/handler.
int IsError(int errcode)
{
	if (errcode == S2218_ERR_OK)
		return 0;
	printf("Error: %s\n", s2218_ErrString(errcode));
	return errcode;
}

int main(int argc, char **argv)
{
	int devflags;
	int devid = 0;
	int devmask = 1;
	int online;

	// for keyboard input use
	changemode(1);
	if (!IsError(s2218_OpenApi(&devflags))) {        // Open the API
		// Abort if no devices discovered
		if (devflags == 0)
			printf("No devices found\n");
		else {
			// Find the first available device (device with lowest address)
			for (devid = 0; devid < S2218_MAX_DEVICES; devid++, devmask <<= 1) {
				if (devflags & devmask) {
					if (!IsError(s2218_OpenDevice(devid, &online))) {
						// Open the device
						UseTheDevice(devid);
						s2218_CloseDevice(devid);
					}
					break;
				}
			}
		}
		// close the API
		s2218_CloseApi();
	}
	//ToggleSensorType(devid);
	changemode(0);
	return 0;
}

char getch() {
	char buf = 0;
	struct termios old = { 0 };
	fflush(stdout);
	if (tcgetattr(0, &old) < 0) perror("tcsetattr()");
	old.c_lflag    &= ~ICANON;   // local modes = Non Canonical mode
	old.c_lflag    &= ~ECHO;     // local modes = Disable echo.
	old.c_cc[VMIN]  = 1;         // control chars (MIN value) = 1
	old.c_cc[VTIME] = 0;         // control chars (TIME value) = 0 (No time)
	if (tcsetattr(0, TCSANOW, &old) < 0) perror("tcsetattr ICANON");
	if (read(0, &buf, 1) < 0) perror("read()");
	old.c_lflag    |= ICANON;    // local modes = Canonical mode
	old.c_lflag    |= ECHO;      // local modes = Enable echo.
	if (tcsetattr(0, TCSADRAIN, &old) < 0) perror ("tcsetattr ~ICANON");
	return buf;
}


static char getkey(void)
{
	return (char)getch();
}

void changemode(int dir)
{
	static struct termios oldt, newt;

	if ( dir == 1 )
	{
		tcgetattr( STDIN_FILENO, &oldt);
		newt = oldt;
		newt.c_lflag &= ~( ICANON | ECHO );
		tcsetattr( STDIN_FILENO, TCSANOW, &newt);
	}
	else
		tcsetattr( STDIN_FILENO, TCSANOW, &oldt);
}

int kbhit(void)
{
	struct timeval tv;
	fd_set rdfs;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	FD_ZERO(&rdfs);
	FD_SET (STDIN_FILENO, &rdfs);
	select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
	return FD_ISSET(STDIN_FILENO, &rdfs);

}

static int keypress(void)
{
	if (!kbhit()) {
		return 0;
	}
	getch();
	return 1;
}


int ReadPowerEnables(int devid)
{
	u8 flags;  // (x, x, x, x, x, +3.3V, +15V, -12.25V)

	if (IsError(s2218_GetPowerEnables(devid, &flags)))
		return -1;

	printf("3.3 %s\n+15 %s\n-12 %s\n",
	       (flags & REQ_POWER_D) ? "ON" : "OFF",
	       (flags & REQ_POWER_P) ? "ON" : "OFF", (flags & REQ_POWER_N) ? "ON" : "OFF");
	return 0;
}

int WritePowerEnables(int devid, u8 mask, u8 flags)
{
	u8 curflags;  // (x, x, x, x, x, +3.3V, +15V, -12.25V)
	u8 newflags;

	if (IsError(s2218_GetPowerEnables(devid, &curflags)))
		return -1;
	newflags = (flags & mask) | (curflags & ~mask);
	if (IsError(s2218_SetPowerEnables(devid, newflags)))
		return -1;
	return ReadPowerEnables(devid);
}

// Status ---------------------------
int GetStatus(int devid)
{
	S2218_STATUSREPORT r;

	if (IsError(s2218_GrabStatus(devid, &r, 2000)))
		return -1;
	printf("meter:\n");
	printf(" DIG  = %6.2f (%s)\n",
	       r.Meter_mV[1] * 0.001, (r.PowerGood & POWERGOOD_3R3V) ? "ok" : "bad");
	printf(" ANA- = %6.2f (%s)\n",
	       r.Meter_mV[0] * 0.001, (r.PowerGood & POWERGOOD_N12V) ? "ok" : "bad");
	printf(" ANA+ = %6.2f (%s)\n",
	       r.Meter_mV[3] * 0.001, (r.PowerGood & POWERGOOD_P15V) ? "ok" : "bad");
	printf(" VREF = %6.2f (%s)\n",
	       r.Meter_mV[2] * 0.001, (r.PowerGood & POWERGOOD_5V)   ? "ok" : "bad");
	if (r.SmadState > SS_FAULT)
		printf("SmadState = ILLEGAL\n");
	else {
		const char *ss[] = {"Unpowered", "Initializing", "Booting", "Online", "Fault"};
		printf("SmadState = %s\n", ss[r.SmadState]);
	}
	if (r.ComState > CS_DAV_FAIL)
		printf("SmadState = ILLEGAL\n");
	else {
		const char *ss[] = {"Normal", "ILLEGAL", "ILLEGAL", "ILLEGAL",
			"Offline", "CRMT timeout", "DAV timeout"};
		printf("ComState = %s\n", ss[r.ComState]);
	}
	return 0;
}

// GPIOs -------------------------
// GPIO outputs
int GpioBinaryCounter(int devid)
{
	int i;
	u8 gpiodata = 0;
	int rc;
	// Configure all GPIOs
	S2218_GPIO_CONFIG cfg;
	cfg.PinDirections               = 0;    // pin directions: all outputs
	cfg.BiasEnables                 = 0;
	cfg.BiasPolarities              = 0;
	cfg.spare                       = 0;
	for (i = 0; i < S2218_NUM_GPIOS; i++) {
		cfg.DbTime[i].Rising  = 0;
		cfg.DbTime[i].Falling = 0;
	}
	if (IsError(s2218_SetGpioConfig(devid, &cfg)))
		return -1;
	// Output binary counter on GPIOs until key pressed
	printf("SetGpioOutputs (press Enter to stop):\n");
	do {
		rc = s2218_SetGpioOutputs(devid, gpiodata++);
		if (IsError(rc)) {
			printf("error set gpio outputs %d\n", rc);
			return -1;
		}
		printf("%c%c%c%d\n", 8, 8, 8,gpiodata);
	}
	while (!keypress() && (gpiodata != 0));
	printf("%c%c%c   ", 8, 8, 8);
	return 0;
}

// GPIO inputs
int GpioScanInputs(int devid)
{
	int i;
	u8 gpiodata = 0;
	int rc;
	// Configure all GPIOs
	S2218_GPIO_CONFIG cfg;
	cfg.PinDirections               = 0xff;    // pin directions: all inputs
	cfg.BiasEnables                 = 0xff;  // enable pull-up/down
	cfg.BiasPolarities              = 0x00;  // pull-down
	cfg.spare                       = 0x00;
	for (i = 0; i < S2218_NUM_GPIOS; i++) {
		cfg.DbTime[i].Rising  = 0;
		cfg.DbTime[i].Falling = 0;
	}
	if (IsError(s2218_SetGpioConfig(devid, &cfg)))
		return -1;
	// Output binary counter on GPIOs until key pressed
	printf("GetGpioInputs (press Enter to stop):\n");
	do {
		gpiodata = 0;
		rc = s2218_GetGpioInputs(devid, &gpiodata);
		if (IsError(rc)) {
			printf("error get gpio inputs %d\n", rc);
			return -1;
		}
		printf("GPIO[in] 0x%x\n", gpiodata);
		usleep(1000*1300);
	}
	while (!keypress());
	printf("GPIO input acquisition complete\n");
	return 0;
}

static int G_sensor_type = 0;

static void print_sensor_type()
{
	switch (G_sensor_type) {
	case SDC_DISABLED:
		printf("SDC Disabled\n");
		break;
	case SDC_VOLTS_5V:
		printf("SDC 5V\n");
		break;
	case SDC_VOLTS_500MV:
		printf("SDC 500mV\n");
		break;
	case SDC_VOLTS_100MV:
		printf("SDC 100mV\n");
		break;
	case SDC_OHMS_400:
		printf("SDC 400 ohms\n");
		break;
	case SDC_OHMS_4K:
		printf("SDC 4K ohms\n");
		break;
	case SDC_OHMS_600K:
		printf("SDC 600K ohms\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_B:
		printf("SDC Thermocouple B\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_C:
		printf("SDC Thermocouple C\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_E:
		printf("SDC Thermocouple E\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_J:
		printf("SDC Thermocouple J\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_K:
		printf("SDC Thermocouple K\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_N:
		printf("SDC Thermocouple N\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_T:
		printf("SDC Thermocouple T\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_S:
		printf("SDC Thermocouple S\n");
		break;
	case SDC_TEMP_THERMOCOUPLE_R:
		printf("SDC Thermocouple R\n");
		break;
	case SDC_TEMP_THERMISTOR:
		printf("SDC Thermistor\n");
		break;
	case SDC_TEMP_RTD_PT100_385_400:
		printf("SDC RTD Pt 385 400C\n");
		break;
	case SDC_TEMP_RTD_PT100_385_800:
		printf("SDC RTD Pt 385 800C\n");
		break;
	case SDC_TEMP_RTD_PT100_392_400:
		printf("SDC RTD Pt 392 400C\n");
		break;
	case SDC_TEMP_RTD_PT100_392_800:
		printf("SDC RTD Pt 392 800C\n");
		break;
	case SDC_TEMP_RTD_NI200:
		printf("SDC RTD Ni 200\n");
		break;
	case SDC_TEMP_RTD_NI1000:
		printf("SDC RTD Ni 1000\n");
		break;
	case SDC_TEMP_RTD_CU10:
		printf("SDC RTD Cu 10\n");
		break;
	}
	return;
}

static int ToggleSensorType(int devid)
{
	int i;
	G_sensor_type++;
	if (G_sensor_type == S2218_NUM_SENSOR_TYPES) {
		G_sensor_type = SDC_DISABLED;
	}
	print_sensor_type();
	for (i = 0; i < S2218_NUM_CHANS; i++) {
		if (IsError(s2218_SetSensorType(devid, i, G_sensor_type)))
			return -1;
	}
	return 0;
}

int ScanSensors(int bchangetype, int devid)     // GET SENSOR DATA -----------
{
	int i;
	S2218_DATAREPORT    report;
	S2218_TRIG_CONFIG   trig;
	int count = 0;
	// Configure all chans for +/-5V input range
	if (bchangetype) {
		for (i = 0; i < S2218_NUM_CHANS; i++) {
			if (IsError(s2218_SetSensorType(devid, i, SDC_VOLTS_5V)))
				return -1;
			s2218_SetOpenSensorValues(devid, 0);
		}
		G_sensor_type = SDC_VOLTS_5V;
	}
	memset(&report, 0, sizeof(S2218_DATAREPORT));
	// Set scan and report interval
	trig.Rate               = 5;    // report once per second
	trig.GpioFlags.Rising   = 0;
	trig.GpioFlags.Falling  = 0;
	if (IsError(s2218_SetTrigConfig(devid, &trig)))
		return -1;
	if (IsError(s2218_SetTrigEnable(devid, 1)))
		return -1;
	// Display incoming reports until keypress
	if (bchangetype)
		printf("scanning sensors 5V range(press Enter to stop)\n");
	else
		printf("reading sensors(press Enter to stop)\n");
	do {
		if (IsError(s2218_WaitForDataReport(devid, &report, 2000))) {
			s2218_SetTrigEnable(devid, 0);  // stop stream
			return -1;
		}
		printf("\n%5d.%3d   ", report.Timestamp.s, report.Timestamp.ms);
		for (i = 0; i < S2218_NUM_CHANS; i++)
            printf("%1.3f   ", report.Samples[i]);
		printf(", scans: %d", count++);
    }
    while (!keypress());

    if (IsError(s2218_SetTrigEnable(devid, 0)))     // turn off sensor data stream
        return -1;

    printf("\n");

    return 0;
}



int ScanSensorResistance(int devid)     // GET SENSOR DATA -----------
{
	int i;
	S2218_DATAREPORT    report;
	S2218_TRIG_CONFIG   trig;
	int count = 0;
	// Configure all chans for +/-5V input range
	for (i = 0; i < S2218_NUM_CHANS; i++) {
		if (IsError(s2218_SetSensorType(devid, i, SDC_OHMS_400)))
			return -1;
		s2218_SetOpenSensorValues(devid, 0);
	}
	G_sensor_type = SDC_OHMS_400;
	memset(&report, 0, sizeof(S2218_DATAREPORT));
	// Set scan and report interval
	trig.Rate               = 5;    // report once per second
	trig.GpioFlags.Rising   = 0;
	trig.GpioFlags.Falling  = 0;
	if (IsError(s2218_SetTrigConfig(devid, &trig)))
		return -1;
	if (IsError(s2218_SetTrigEnable(devid, 1)))
		return -1;
	// Display incoming reports until keypress
	printf("scanning sensor resistance range(press Enter to stop)\n");
	do {
		if (IsError(s2218_WaitForDataReport(devid, &report, 2000))) {
			s2218_SetTrigEnable(devid, 0);  // stop stream
			return -1;
		}
		printf("\n%5d.%3d   ", report.Timestamp.s, report.Timestamp.ms);
		for (i = 0; i < S2218_NUM_CHANS; i++)
			printf("%lf  ", report.Samples[i]);
		printf(", scans: %d", count++);
		//usleep(1000*500);
	}
	while (!keypress());
	if (IsError(s2218_SetTrigEnable(devid, 0)))     // turn off sensor data stream
		return -1;
	return 0;
}

int ShowVersionInfo(int devid)
{
	S2218_VERSION_INFO ver;
	if (IsError(s2218_GetDeviceVersions(devid, &ver)))
		return -1;
	printf("Version info: ");
	printf("dev %d.%d.%d, ", ver.DevFw.Major, ver.DevFw.Minor, ver.DevFw.Build);
	printf("smad %d.%d, ", ver.SmadFw/100, ver.SmadFw - 100*(ver.SmadFw/100));
	printf("pwb rev%c\n", ver.PwbRev);
	return 0;
}

int Calibrate5V(int devid)
{
	int rc;
	double ref;
	int i;
	for (i = 0; i < S2218_NUM_CHANS; i++)
		s2218_SetSensorType(devid, 0, SDC_VOLTS_5V);
	G_sensor_type = SDC_VOLTS_5V;
	usleep(100*1000);
	rc = s2218_CalibrateStandard(devid, 0, 0, 4.9815, &ref);
	printf("[%d]sent calibrate 5V on channel 0, ref %lf\n", rc, ref);
	usleep(100*1000);
	s2218_AnalogReset(devid);
	printf("analog reset\n");
	return 0;
}

int Calibrate500mV(int devid)
{
	int rc;
	double ref;
	int i;
	for (i = 0; i < S2218_NUM_CHANS; i++)
		s2218_SetSensorType(devid, 0, SDC_VOLTS_500MV);
	G_sensor_type = SDC_VOLTS_500MV;
	s2218_SetSensorType(devid, 0, SDC_VOLTS_500MV);
	ref = 0;
	usleep(300*1000);
	rc = s2218_CalibrateStandard(devid, 0, 1, 0.48105, &ref);
	printf("[%d]sent calibrate 500mV on channel 0, ref %lf\n", rc, ref);
	usleep(1000*1000);
	s2218_AnalogReset(devid);
	printf("analog reset\n");
	return 0;
}


int Calibrate400ohm(int devid)
{
	int rc;
	double ref;
	int i;
	for (i = 0; i < S2218_NUM_CHANS; i++)
		s2218_SetSensorType(devid, 0, SDC_OHMS_400);
	G_sensor_type = SDC_OHMS_400;
	ref = 0;
	printf("set sensor type to 400 ohm\n");
	usleep(1000*1000);
	rc = s2218_CalibrateStandard(devid, 0, 2, 400.0, &ref);
	printf("[%d]sent calibrate 400ohm on channel 0, ref %lf\n", rc, ref);
	usleep(1000*1000);
	s2218_AnalogReset(devid);
	printf("analog reset\n");
	return 0;
}

void UseTheDevice(int devid)
{
	char cmd;
	int ok = 0;

    printf("\nUsing 2218 board %d\n", devid);

    ShowVersionInfo(devid);

    while (ok == 0)
    {
        // Display menu
        printf("\nMENU **************\n");
        printf("? - read version info\n");
        printf("A - pos V on\n");
        printf("B - pos V off\n");
        printf("C - neg V on\n");
        printf("D - neg V off\n");
        printf("F - full status\n");
        printf("G - gpio binary counter\n");
		printf("I - gpIo scan inputs\n");
        printf("P - read power enables\n");
		printf("V - scan sensors as Voltage(5V range)\n");
		printf("R - scan sensors as Resistance(400ohm range)\n");
		// For manufacturing, calibration.
		//printf("0 - Calibrate 5V (SL0, SH0)\n");
		//printf("1 - Calibrate 500mV (SL0, SH0)\n");
		//printf("2 - Calibrate 400ohm (SL0, PL0 <--> SH0, PH0)\n");
		printf("T - Toggle Sensor Types\n");
		printf("S - Scan Sensors (use T to change type)\n");
        printf("Q - quit\n");
		printf("Current Sensors Type[%d/%d]: ",
		       G_sensor_type, S2218_NUM_SENSOR_TYPES-1);
		print_sensor_type();
		printf("cmd>");
		cmd = toupper(getkey());    // Input menu choice
		printf("Command: %c\n\n", cmd);
		// Decode and execute
		switch (cmd) {
		case '?':
			ok = ShowVersionInfo(devid);
			break;
		case 'A':
			ok = WritePowerEnables(devid, REQ_POWER_P, REQ_POWER_P);
			break;
		case 'B':
			ok = WritePowerEnables(devid, REQ_POWER_P, 0);
			break;
		case 'C':
			ok = WritePowerEnables(devid, REQ_POWER_N, REQ_POWER_N);
			break;
		case 'D':
			ok = WritePowerEnables(devid, REQ_POWER_N, 0);
			break;
		case 'F':
			ok = GetStatus(devid);
			break;
		case 'G':
			ok = GpioBinaryCounter(devid);
			break;
		case 'I':
			ok = GpioScanInputs(devid);
			break;
		case 'P':
			ok = ReadPowerEnables(devid);
			break;
		case 'V':
			ok = ScanSensors(1, devid);
			break;
		case 'R':
			ok = ScanSensorResistance(devid);
			break;
		case '0':
			ok = Calibrate5V(devid);
			break;
		case '1':
			ok = Calibrate500mV(devid);
			break;
		case '2':
			ok = Calibrate400ohm(devid);
			break;
		case 'T':
			ok = ToggleSensorType(devid);
			break;
		case 'S':
			ok = ScanSensors(0, devid);
			break;
		case 'Q':
			return;
		}
	}
}

